/*
 *	linux/arch/i386/kernel/irq.c
 *
 *	Copyright (C) 1992, 1998 Linus Torvalds, Ingo Molnar
 *
 * This file contains the lowest level x86-specific interrupt
 * entry, irq-stacks and irq statistics code. All the remaining
 * irq logic is done by the generic kernel/irq/ code and
 * by the x86-specific irq controller code. (e.g. i8259.c and
 * io_apic.c.)
 */

#include <asm/uaccess.h>
#include <linux/module.h>
#include <linux/seq_file.h>
#include <linux/interrupt.h>
#include <linux/kernel_stat.h>
#include <linux/notifier.h>
#include <linux/cpu.h>
#include <linux/delay.h>

DEFINE_PER_CPU(irq_cpustat_t, irq_stat) ____cacheline_internodealigned_in_smp;
EXPORT_PER_CPU_SYMBOL(irq_stat);

#ifndef CONFIG_X86_LOCAL_APIC
/*
 * 'what should we do if we get a hw irq event on an illegal vector'.
 * each architecture has to answer this themselves.
 */
void ack_bad_irq(unsigned int irq)
{
	printk("unexpected IRQ trap at vector %02x\n", irq);
}
#endif

#ifdef CONFIG_4KSTACKS
/*
 * per-CPU IRQ handling contexts (thread information and stack)
 */
union irq_ctx {
	struct thread_info      tinfo;
	u32                     stack[THREAD_SIZE/sizeof(u32)];
};

static union irq_ctx *hardirq_ctx[NR_CPUS] __read_mostly;
static union irq_ctx *softirq_ctx[NR_CPUS] __read_mostly;
#endif

/*
 * do_IRQ handles all normal device IRQ's (the special
 * SMP cross-CPU interrupts have their own specific
 * handlers).
 */
fastcall unsigned int do_IRQ(struct pt_regs *regs)
{
#ifdef CONFIG_NOOKS

#define struct_cpy(x,y)                         \
({                                              \
        if (sizeof(*(x)) != sizeof(*(y)))       \
                printk ("Fatal:Error.\n");               \
        memcpy(x, y, sizeof(*(x)));             \
})


  nook_task_state_t nook_state;
  int copied_state = 0;

  if ((current->nook_state.curr.current_driver != NULL)  ||
      (current->nook_state.curr.current_nook != NULL))

  {
    struct_cpy(&nook_state, &current->nook_state);
    memset(&current->nook_state.curr, 0, sizeof(nook_current_state_t));
    //
    // Reset the top of the stack
    //
    //    if (regs.nook_reg_esp != 0) {
    //      if ((unsigned long) regs.nook_reg_esp >
    //    (unsigned long) current->nook_state.thread_state.trap_state.nook_stack) {
    //  printk(KERN_ERR "Nooks stack less than last base\n");
    //  BUG();
    //      }
    //      current->nook_state.thread_state.trap_state.nook_stack =
    //  (void *) regs.nook_reg_esp;
    //    }
    copied_state = 1;
  }


        {
#endif // CONFIG_NOOKS
	
	/* high bit used in ret_from_ code */
	int irq = ~regs->orig_eax;
#ifdef CONFIG_4KSTACKS
	union irq_ctx *curctx, *irqctx;
	u32 *isp;
#endif

	if (unlikely((unsigned)irq >= NR_IRQS)) {
		printk(KERN_EMERG "%s: cannot handle IRQ %d\n",
					__FUNCTION__, irq);
		BUG();
	}

	irq_enter();
#ifdef CONFIG_DEBUG_STACKOVERFLOW
	/* Debugging check for stack overflow: is there less than 1KB free? */
	{
		long esp;

		__asm__ __volatile__("andl %%esp,%0" :
					"=r" (esp) : "0" (THREAD_SIZE - 1));
		if (unlikely(esp < (sizeof(struct thread_info) + STACK_WARN))) {
			printk("do_IRQ: stack overflow: %ld\n",
				esp - sizeof(struct thread_info));
			dump_stack();
		}
	}
#endif

#ifdef CONFIG_4KSTACKS

	curctx = (union irq_ctx *) current_thread_info();
	irqctx = hardirq_ctx[smp_processor_id()];

	/*
	 * this is where we switch to the IRQ stack. However, if we are
	 * already using the IRQ stack (because we interrupted a hardirq
	 * handler) we can't do that and just have to keep using the
	 * current stack (which is the irq stack already after all)
	 */
	if (curctx != irqctx) {
		int arg1, arg2, ebx;

		/* build the stack frame on the IRQ stack */
		isp = (u32*) ((char*)irqctx + sizeof(*irqctx));
		irqctx->tinfo.task = curctx->tinfo.task;
		irqctx->tinfo.previous_esp = current_stack_pointer;

		/*
		 * Copy the softirq bits in preempt_count so that the
		 * softirq checks work in the hardirq context.
		 */
		irqctx->tinfo.preempt_count =
			(irqctx->tinfo.preempt_count & ~SOFTIRQ_MASK) |
			(curctx->tinfo.preempt_count & SOFTIRQ_MASK);

		asm volatile(
			"       xchgl   %%ebx,%%esp      \n"
			"       call    __do_IRQ         \n"
			"       movl   %%ebx,%%esp      \n"
			: "=a" (arg1), "=d" (arg2), "=b" (ebx)
			:  "0" (irq),   "1" (regs),  "2" (isp)
			: "memory", "cc", "ecx"
		);
	} else
#endif
		__do_IRQ(irq, regs);

	irq_exit();
#ifdef CONFIG_NOOKS
  }

  //
  // copy back the current nook state from the stack
  //

  if (copied_state) {
    struct_cpy(&current->nook_state, &nook_state);
  }

#endif // CONFIG_NOOKS

	return 1;
}

#ifdef CONFIG_NOOKS
/*
 * do_nooks_IRQ handles polling normal interrupts that were
 * disabled due to a driver crash (nook unload).
 */
asmlinkage unsigned int do_nooks_IRQ(struct pt_regs regs)
{
        int irq = regs.orig_eax & 0xff; /* high bits used in ret_from_ code  */
        irq_desc_t *desc = irq_desc + irq;
        struct irqaction * action;
        unsigned int status;
        unsigned long flags;



        spin_lock_irqsave(&desc->lock,flags);

        //
        // If our is the only lock on disabling the IRQ, then
        // deliver it anyway
        //

        if ((desc->depth == 1) && (desc->action != NULL)) {

          /*
            REPLAY is when Linux resends an IRQ that was dropped earlier
            WAITING is used by probe to mark irqs that are being tested
          */
          status = desc->status & ~(IRQ_REPLAY | IRQ_WAITING);
          status |= IRQ_PENDING; /* we _want_ to handle it */

          /*
           * If the IRQ is disabled for whatever reason, we cannot
           * use the action we have.
           */
          action = NULL;
          if (!(status & (IRQ_INPROGRESS))) {
            action = desc->action;
            status &= ~IRQ_PENDING; /* we commit to handling */
            status |= IRQ_INPROGRESS; /* we are handling it */
          }
          desc->status = status;

          /*
           * If there is no IRQ handler or it was disabled, exit early.
           Since we set PENDING, if another processor is handling
           a different instance of this same irq, the other processor
           will take care of it.
          */
         if (!action)
            goto out;


          /*
           * Edge triggered interrupts need to remember
           * pending events.
           * This applies to any hw interrupts that allow a second
           * instance of the same irq to arrive while we are in do_IRQ
           * or in the handler. But the code here only handles the _second_
           * instance of the irq, not the third or fourth. So it is mostly
           * useful for irq hardware that does not mask cleanly in an
           * SMP environment.
           */
          for (;;) {
            spin_unlock(&desc->lock);
            handle_IRQ_event(irq, &regs, action);

            spin_lock(&desc->lock);

            if (!(desc->status & IRQ_PENDING))
                        break;
            desc->status &= ~IRQ_PENDING;
          }
          desc->status &= ~IRQ_INPROGRESS;
        }

out:

        spin_unlock_irqrestore(&desc->lock,flags);

  return 1;
}
int request_nooks_irq(unsigned int irq,
                      void (*handler)(int, void *, struct pt_regs *),
                      unsigned long irqflags,
                      const char * devname,
                      void *dev_id)
{
        int retval;
        struct irqaction * action;

#if 1
        /*
         * Sanity-check: shared interrupts should REALLY pass in
         * a real dev-ID, otherwise we'll have trouble later trying
         * to figure out which interrupt is which (messes up the
         * interrupt freeing logic etc).
         */
        if (irqflags & SA_SHIRQ) {
                if (!dev_id)
                        printk("Bad boy: %s (at 0x%x) called us without a dev_id!\n", devname, (&irq)[-1]);
        }
#endif

        if (irq >= NR_IRQS)
                return -EINVAL;
        if (!handler)
                return -EINVAL;

        action = (struct irqaction *)
                        kmalloc(sizeof(struct irqaction), GFP_KERNEL);
        if (!action)
                return -ENOMEM;

        action->handler = handler;
        action->flags = irqflags;
        //action->mask = 0;
        action->name = devname;
        action->next = NULL;
        action->dev_id = dev_id;

        retval = setup_nooks_irq(irq, action);
        if (retval)
                kfree(action);
        return retval;
}


struct irqaction *  get_irq_action(unsigned int irq, void *dev_id)
{
  irq_desc_t *desc;
  struct irqaction * action;
  unsigned long flags;

  if (irq >= NR_IRQS)
    return NULL;

  desc = irq_desc + irq;
  spin_lock_irqsave(&desc->lock,flags);
  action = desc->action;
  while (action != NULL) {
    if (action->dev_id == dev_id)
      break;
    action = action->next;
  }
  spin_unlock_irqrestore(&desc->lock,flags);
  return action;
}

int setup_nooks_irq(unsigned int irq, struct irqaction * new)
{
        int shared = 0;
        unsigned long flags;
        struct irqaction *old, **p;
        irq_desc_t *desc = irq_desc + irq;

        /*
         * Some drivers like serial.c use request_irq() heavily,
         * so we have to be careful not to interfere with a
         * running system.
         */
        if (new->flags & SA_SAMPLE_RANDOM) {
                /*
                 * This function might sleep, we want to call it first,
                 * outside of the atomic block.
                 * Yes, this might clear the entropy pool if the wrong
                 * driver is attempted to be loaded, without actually
                 * installing a new handler, but is this really a problem,
                 * only the sysadmin is able to do this.
                 */
                rand_initialize_irq(irq);
        }
       /*
         * The following block of code has to be executed atomically
         */
        spin_lock_irqsave(&desc->lock,flags);
        p = &desc->action;
        if ((old = *p) != NULL) {
                /* Can't share interrupts unless both agree to */
                if (!(old->flags & new->flags & SA_SHIRQ)) {
                        spin_unlock_irqrestore(&desc->lock,flags);
                        return -EBUSY;
                }

                /* add new interrupt at end of irq queue */
                do {
                        p = &old->next;
                        old = *p;
                } while (old);
                shared = 1;
        }
        *p = new;

        if (!shared) {
                desc->depth = 0;
                desc->status &= ~(IRQ_DISABLED | IRQ_AUTODETECT | IRQ_WAITING);
                // desc->handle_irq(irq); FIXME TBD
        }
        //
        // BUGBUG: new line here
        //
        disable_irq_nosync(irq);
        spin_unlock_irqrestore(&desc->lock,flags);

        register_irq_proc(irq);
        return 0;
}



#endif // CONFIG_NOOKS



#ifdef CONFIG_4KSTACKS

/*
 * These should really be __section__(".bss.page_aligned") as well, but
 * gcc's 3.0 and earlier don't handle that correctly.
 */
static char softirq_stack[NR_CPUS * THREAD_SIZE]
		__attribute__((__aligned__(THREAD_SIZE)));

static char hardirq_stack[NR_CPUS * THREAD_SIZE]
		__attribute__((__aligned__(THREAD_SIZE)));

/*
 * allocate per-cpu stacks for hardirq and for softirq processing
 */
void irq_ctx_init(int cpu)
{
	union irq_ctx *irqctx;

	if (hardirq_ctx[cpu])
		return;

	irqctx = (union irq_ctx*) &hardirq_stack[cpu*THREAD_SIZE];
	irqctx->tinfo.task              = NULL;
	irqctx->tinfo.exec_domain       = NULL;
	irqctx->tinfo.cpu               = cpu;
	irqctx->tinfo.preempt_count     = HARDIRQ_OFFSET;
	irqctx->tinfo.addr_limit        = MAKE_MM_SEG(0);

	hardirq_ctx[cpu] = irqctx;

	irqctx = (union irq_ctx*) &softirq_stack[cpu*THREAD_SIZE];
	irqctx->tinfo.task              = NULL;
	irqctx->tinfo.exec_domain       = NULL;
	irqctx->tinfo.cpu               = cpu;
	irqctx->tinfo.preempt_count     = 0;
	irqctx->tinfo.addr_limit        = MAKE_MM_SEG(0);

	softirq_ctx[cpu] = irqctx;

	printk("CPU %u irqstacks, hard=%p soft=%p\n",
		cpu,hardirq_ctx[cpu],softirq_ctx[cpu]);
}

void irq_ctx_exit(int cpu)
{
	hardirq_ctx[cpu] = NULL;
}

extern asmlinkage void __do_softirq(void);

asmlinkage void do_softirq(void)
{
	unsigned long flags;
	struct thread_info *curctx;
	union irq_ctx *irqctx;
	u32 *isp;

	if (in_interrupt())
		return;

	local_irq_save(flags);

	if (local_softirq_pending()) {
		curctx = current_thread_info();
		irqctx = softirq_ctx[smp_processor_id()];
		irqctx->tinfo.task = curctx->task;
		irqctx->tinfo.previous_esp = current_stack_pointer;

		/* build the stack frame on the softirq stack */
		isp = (u32*) ((char*)irqctx + sizeof(*irqctx));

		asm volatile(
			"       xchgl   %%ebx,%%esp     \n"
			"       call    __do_softirq    \n"
			"       movl    %%ebx,%%esp     \n"
			: "=b"(isp)
			: "0"(isp)
			: "memory", "cc", "edx", "ecx", "eax"
		);
		/*
		 * Shouldnt happen, we returned above if in_interrupt():
	 	 */
		WARN_ON_ONCE(softirq_count());
	}

	local_irq_restore(flags);
}

EXPORT_SYMBOL(do_softirq);
#endif

/*
 * Interrupt statistics:
 */

atomic_t irq_err_count;

/*
 * /proc/interrupts printing:
 */

int show_interrupts(struct seq_file *p, void *v)
{
	int i = *(loff_t *) v, j;
	struct irqaction * action;
	unsigned long flags;

	if (i == 0) {
		seq_printf(p, "           ");
		for_each_online_cpu(j)
			seq_printf(p, "CPU%-8d",j);
		seq_putc(p, '\n');
	}

	if (i < NR_IRQS) {
		spin_lock_irqsave(&irq_desc[i].lock, flags);
		action = irq_desc[i].action;
		if (!action)
			goto skip;
		seq_printf(p, "%3d: ",i);
#ifndef CONFIG_SMP
		seq_printf(p, "%10u ", kstat_irqs(i));
#else
		for_each_online_cpu(j)
			seq_printf(p, "%10u ", kstat_cpu(j).irqs[i]);
#endif
		seq_printf(p, " %14s", irq_desc[i].chip->typename);
		seq_printf(p, "  %s", action->name);

		for (action=action->next; action; action = action->next)
			seq_printf(p, ", %s", action->name);

		seq_putc(p, '\n');
skip:
		spin_unlock_irqrestore(&irq_desc[i].lock, flags);
	} else if (i == NR_IRQS) {
		seq_printf(p, "NMI: ");
		for_each_online_cpu(j)
			seq_printf(p, "%10u ", nmi_count(j));
		seq_putc(p, '\n');
#ifdef CONFIG_X86_LOCAL_APIC
		seq_printf(p, "LOC: ");
		for_each_online_cpu(j)
			seq_printf(p, "%10u ",
				per_cpu(irq_stat,j).apic_timer_irqs);
		seq_putc(p, '\n');
#endif
		seq_printf(p, "ERR: %10u\n", atomic_read(&irq_err_count));
#if defined(CONFIG_X86_IO_APIC)
		seq_printf(p, "MIS: %10u\n", atomic_read(&irq_mis_count));
#endif
	}
	return 0;
}

#ifdef CONFIG_HOTPLUG_CPU
#include <mach_apic.h>

void fixup_irqs(cpumask_t map)
{
	unsigned int irq;
	static int warned;

	for (irq = 0; irq < NR_IRQS; irq++) {
		cpumask_t mask;
		if (irq == 2)
			continue;

		cpus_and(mask, irq_desc[irq].affinity, map);
		if (any_online_cpu(mask) == NR_CPUS) {
			printk("Breaking affinity for irq %i\n", irq);
			mask = map;
		}
		if (irq_desc[irq].chip->set_affinity)
			irq_desc[irq].chip->set_affinity(irq, mask);
		else if (irq_desc[irq].action && !(warned++))
			printk("Cannot set affinity for irq %i\n", irq);
	}

#if 0
	barrier();
	/* Ingo Molnar says: "after the IO-APIC masks have been redirected
	   [note the nop - the interrupt-enable boundary on x86 is two
	   instructions from sti] - to flush out pending hardirqs and
	   IPIs. After this point nothing is supposed to reach this CPU." */
	__asm__ __volatile__("sti; nop; cli");
	barrier();
#else
	/* That doesn't seem sufficient.  Give it 1ms. */
	local_irq_enable();
	mdelay(1);
	local_irq_disable();
#endif
}
#endif

