/*
 * nooks-mm.c -- memory management for nooks
 *
 * Major changes for 2.6.18 porting.
 * 2008 - Asim Kadav : Map and pud(ding) fixes.
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 * external calls into here:
 * nooks_mm_update_pagetable
 * nooks_update_pte
 * nooks_update_pte_phys
 * nooks_update_pmd
 */

 
#include "nooks-i.h"
#include <linux/module.h>
#include "nooks-mm.h"
#include "nooks-wrappers.h"
#include "nooks-switch.h"
#include <linux/slab.h>


#ifndef CONFIG_NOOKS
#warning "Nooks not defined"
#endif // CONFIG_NOOKS

#define FIX_LIST(_y_, _x_) do { if (!list_empty(_y_)) BUG(); \
                           INIT_LIST_HEAD(_x_);\
                           if (!list_empty(_y_)) BUG();} while (0);

/*
 * Global variables
 */

irq_cpustat_t nooks_w_irq_stat[1];
#define NOOKS_INIT_MM(name) \
{			 				\
	mm_rb:		RB_ROOT,			\
	pgd:		NULL, 		                \
	mm_users:	ATOMIC_INIT(2), 		\
	mm_count:	ATOMIC_INIT(1), 		\
	mmap_sem:	__RWSEM_INITIALIZER(name.mmap_sem), \
	page_table_lock: SPIN_LOCK_UNLOCKED, 		\
	mmlist:		LIST_HEAD_INIT(name.mmlist),	\
}


struct mm_struct nooks_mm_struct = NOOKS_INIT_MM(nooks_mm_struct); 
spinlock_t nooks_list_lock;



int 
nooks_copy_pte_set (struct mm_struct * mm, pte_t * pte, 
		    pte_t * swapper_pte, 
		    unsigned long address,
		    int wrprotect)
{
  int i;

  for (i = 0; i < PTRS_PER_PTE; i++) {
    if (!pte_none(*swapper_pte)) {

      //
      // Copy the PTE as readonly
      //
      if (wrprotect) {
	*pte = pte_wrprotect(*swapper_pte);
      } else {
	*pte = *swapper_pte;
      }


      //
      // Make it local
      //

      *pte = pte_mklocal(*pte);


    } else {
      pte_clear(mm, address, pte);
    }
    address += PAGE_SIZE;
    pte++;
    swapper_pte++;
  }
  return(0);
}

int 
nooks_create_phys_pte_set (pte_t * pte, 
			   unsigned long paddr, 
			   unsigned long vaddr,
			   int wrprotect)
{
  int i;

  for (i = 0; i < PTRS_PER_PTE; i++) {
    if (wrprotect) {
      *pte = pfn_pte(paddr >> PAGE_SHIFT, NOOKS_PAGE_KERNEL_RO);
    } else {
      *pte = pfn_pte(paddr >> PAGE_SHIFT, NOOKS_PAGE_KERNEL);
    }
      
    vaddr += PAGE_SIZE;
    paddr += PAGE_SIZE;
    pte++;
  }
  return(0);
}

//
// Copy a set of PMD. This requires iterating through the entries in the set 
// of  PMDs, and allocating a new set of PTEs for each PMD entry
//
int 
nooks_copy_pmd_set(struct mm_struct * mm,
		   pmd_t * pmd, 
		   pmd_t * swapper_pmd, 
		   unsigned long address,
		   int wrprotect)
{
  int i;
  int result;
  pte_t * swapper_pte, *pte;

  for (i = 0; i < PTRS_PER_PMD; i++ ) {
    //
    // If the kernel page table has an entry in this pmd, then copy it
    //
    if (!pmd_none(*swapper_pmd)) {

      //
      // Allocate a new PTE
      //

      pte = pte_alloc_map(mm, pmd, address);
      if (pte == NULL) {
	return -ENOMEM;
      }

      //
      // turn off page_user
      //

      pmd->pmd &= ~_PAGE_USER;
      
      //
      // Check if this is a 4mb superpage
      //
      if ((pmd_val(*swapper_pmd) & _PAGE_PSE) == 0) {

	swapper_pte = pte_offset_map(swapper_pmd, address);

	result = nooks_copy_pte_set(mm, pte, swapper_pte, address, wrprotect);
	if (result < 0) { 
	  return result;
	}
      } else {
	//
	// We are working with a superpage
	//

	result = nooks_create_phys_pte_set(pte, 
					   pmd_val(*swapper_pmd) & PMD_MASK,
					   address,
					   wrprotect);
	if (result < 0) { 
	  return result;
	}
      }
    } else {

      //
      // The kernel has no entry for this pmd, so clear our pmd
      //

      pmd_clear(pmd);
    }
    address += PMD_SIZE;
    pmd++;
    swapper_pmd++;
  }
  return 0;
}



void
nooks_update_addr_prot(struct mm_struct * mm,
		       void * pte_addr)
{
  pgd_t * pgd = NULL;
  pmd_t * pmd = NULL;
  pte_t * pte = NULL;
  pud_t * pud = NULL;
  unsigned long vaddr = (unsigned long) pte_addr;

  pgd = mm->pgd + pgd_index(vaddr);
  if (pgd_none(*pgd)) {
    goto Cleanup;
  }
 
  pud = pud_offset(pgd, vaddr);
  pmd = pmd_offset(pud, vaddr);
  if (pmd_none(*pmd)) {
    goto Cleanup;
  }

  pte = pte_offset_map(pmd, vaddr);
  *pte = pte_mkwrite(*pte);
 Cleanup:
  return;
}

void
nooks_update_pmd_set_prot(struct mm_struct * mm,
			  pmd_t * pmd, 
			  unsigned long address)
{
  int i;
  pte_t *pte;


  for (i = 0; i < PTRS_PER_PMD; i++ ) {
    //
    // If the kernel page table has an entry in this pmd, then copy it
    //
    if (!pmd_none(*pmd)) {

      pte = pte_offset_map(pmd, address);
      nooks_update_addr_prot(mm, pte);
    } 
    address += PMD_SIZE;
    pmd++;
  }
}



int 
nooks_copy_page_table(struct mm_struct * mm,
		      struct mm_struct * from_mm,
		      int wrprotect)
{
  int i;
  int result = 0;
  unsigned long pgd_address;
  pgd_t * source = from_mm->pgd;
  pgd_t * source_pgd;
  pud_t * source_pud;
#if 0
  pgd_t * pgd;
#endif // 0
  //
  // first copy the page directory. This routine actually zeroes out the
  // user-space entries and copies the kernel entries from source_pg_dir,
  // but we want to replace the kernel entries to point to our own
  // pages of PTEs. This allocates a page of PGDs
  //

  mm->pgd = pgd_alloc(mm);
  if (mm->pgd == NULL ) {
    return(-ENOMEM);
  }


  spin_lock(&from_mm->page_table_lock);

  //
  // zero out the page directory. The routines for allocating entries
  // test whether the entries are already filled in, and do nothing
  // if they are.
  //
  memset(mm->pgd,0, PTRS_PER_PGD * sizeof(pgd_t));

  pgd_address = PAGE_OFFSET;

  // 
  // now loop through the entries of the directory, allocating a set of PMDs
  // for each PGD.
  //
  for (i = USER_PTRS_PER_PGD; i < PTRS_PER_PGD; i++) {
    pmd_t * pmd = NULL, * source_pmd;
    pud_t * pud = NULL;

    pgd_clear(mm->pgd+i);
    
    //
    // If the pgd entry is clear, just clear the copy and continue
    //
    source_pgd = source+i;
    if (!pgd_none(*source_pgd)) {

      // 
      // The pgd is not clear, so allocate a set of PMDs for the pgd entry
      //

      pud = pud_alloc(mm, mm->pgd + i, pgd_address);
      if (pud)
      	pmd = pmd_alloc(mm, pud, pgd_address);
      if (pmd == NULL) {
	result = -ENOMEM;
	break;
      }

    
      source_pud = pud_offset(source_pgd, pgd_address);
      source_pmd = pmd_offset(source_pud, pgd_address);

      result = nooks_copy_pmd_set(mm, pmd, source_pmd, pgd_address, wrprotect);
      if (result < 0) {
	break;
      }    
    }
    pgd_address += PGDIR_SIZE;
		      
  }
  spin_unlock(&from_mm->page_table_lock);
  spin_lock_init(&mm->page_table_lock);

  //
  // BUGBUG: update protection to make pagetable writable
  //
#if 0
  pgd = mm->pgd + USER_PTRS_PER_PGD;
  pgd_address = PAGE_OFFSET;
  nooks_update_addr_prot(mm, mm->pgd);

  for (i = USER_PTRS_PER_PGD; i < PTRS_PER_PGD; i++) {
    pmd_t * pmd;
    
    //
    // If the pgd entry is clear, just clear the copy and continue
    //
    if (!pgd_none(*pgd)) {

      // 
      // The pgd is not clear, so allocate a set of PMDs for the pgd entry
      //
      pmd = pmd_offset(pgd, pgd_address);

      nooks_update_pmd_set_prot(mm, pmd, pgd_address);
    }
    pgd_address += PGDIR_SIZE;
    pgd++;

  }
#endif
  PDEBUG(" Nooks copied page table\n");
  nooks_compare_mm(mm, from_mm, 0);
  return(result);   
}

pte_t nooks_pmd_to_pte(pmd_t pmd) {
  return *((pte_t *) &pmd);
}

/**********************************************************
 *
 * nooks_mm_init
 *
 * Allocate the shadow pagetable and associated spinlock
 *
 *********************************************************/

int
nooks_mm_init(void)
{
  int result = 0;
  /*result = nooks_copy_page_table(&nooks_mm_struct, 
				 &init_mm,
				 1); // writeprotect
  if (result < 0) {
    goto Cleanup;
  }

  */ 
 Cleanup:
  return(result);
}

/**********************************************************
 *
 * nooks_mm_cleanup
 *
 * Free the shadow pagetable and associated spinlock
 *
 *********************************************************/
void
nooks_mm_cleanup(void)
{
  if (nooks_mm_struct.pgd != NULL) {
    int i,j;
    pte_t * pte;
    pmd_t * pmd;
    pgd_t * pgd;
    pud_t * pud;

    spin_lock(&nooks_mm_struct.page_table_lock);
    pgd = nooks_mm_struct.pgd;
    
    //
    // Loop through the entries in the the PGD and free each PMD
    //
    
    for (i = USER_PTRS_PER_PGD; i < PTRS_PER_PGD; i++) {
      
      //
      // If the pgd entry is non-null, free it
      //
      
      if (!pgd_none(*(pgd+i)) && !pgd_bad(*(pgd+i))) {
        
	pud = pud_offset(pgd+i, 0);
        pmd = pmd_offset(pud,0);

	pgd_clear((pgd+i));
	
	//
	// Loop through the entries in the PMD and free each page of PTEs
	//
	
	for (j = 0; j < PTRS_PER_PMD ; j++) {
	  if (!pmd_none(*(pmd+j)) && !pmd_bad(*(pmd+j))) {
	    pte = pte_offset_map(pmd+j, 0);
	    pmd_clear(pmd+j);
	    pte_unmap(pte);
	  }
	  pmd_free(pmd);
	}
      }

    }
    //
    // Free the PGD itself
    //
    pgd_free(nooks_mm_struct.pgd);
    nooks_mm_struct.pgd = NULL;
    spin_unlock(&nooks_mm_struct.page_table_lock);
  }
  
  //
  // BUGBUG: do nothing
  //
}


/**********************************************************
 *
 * nooks_udpate_area_pte
 *
 * update the protection on the entries of a pte in the shadow pagetable
 *
 *********************************************************/

void 
nooks_update_area_pte(int writeable, 
		      pte_t * pte, 
		      unsigned long address, 
		      unsigned long size)
{
  unsigned long end;
  //  unsigned long vaddr = address;

  if (pte !=NULL) 
  {
  address &= ~PMD_MASK;
  end = address + size;
  if (end > PMD_SIZE) {
    end = PMD_SIZE;
  }
  //
  // Round down the address so that we don't miss the last page
  //
  address = address & PAGE_MASK;
  do {
    //
    // If the page is writable, make it so. Otherwise 
    // writeprotect.
    //
    if (writeable) {
      *pte = pte_mkwrite(*pte);
    } else {
      *pte = pte_wrprotect(*pte);
    }
    address += PAGE_SIZE;
    pte++;
  } while (address < end);
  }

}

#ifdef NOOKS_DEBUG2
/**********************************************************
 * nooks_dbg_kmalloc
 * 
 *
 * wrapper for kmalloc
 *
 *********************************************************/
void * nooks_dbg_kmalloc(size_t size, int flags)
{
  void * result;

  result = kmalloc(size + NOOKS_MM_KMALLOC_OVERHEAD, flags);
  if (result != NULL) {
    NOOKS_STATS_ADD(NOOKS_OPS_KMALLOC_BYTES, size);
    * (size_t *) result = size;
    return( (u8 *) result + NOOKS_MM_KMALLOC_OVERHEAD);
  } else {
    return(result);
  }  
}


/**********************************************************
 *
 * nooks_dbg_kfree
 *
 * wrapper for kfree
 *
 *********************************************************/
void nooks_dbg_kfree(const void * ptr) 
{
  if (ptr != NULL) {
    u8 * ptr_s = (u8*) ptr - NOOKS_MM_KMALLOC_OVERHEAD;
    NOOKS_STATS_ADD(NOOKS_OPS_KMALLOC_BYTES, - (*(size_t *) ptr_s));
    kfree(ptr_s);
  }   
}
#endif

/**********************************************************
 *
 * nooks_mm_init_nook
 *
 * Initialize the slab cache for a nook
 *
 *********************************************************/

int
nooks_mm_init_nook(pnook_t nook) 
{
  int result = 0;
  int i;
  size_t size = NOOKS_CACHE_FIRST_SIZE;


  printk ("Initialzing nooks_mm_init_nook.\n"); 

  NOOKS_ENTER();
  INIT_LIST_HEAD(&nook->resources.page_map);
  result = nooks_create_hash_table(&nook->resources.hash_pool, 
				   &nook->resources.area_map, 
				   32, 
				   0); // not a dual
  
  if (result < 0) {
    goto Cleanup;
  }

  INIT_LIST_HEAD(&nook->resources.cache_list);
  INIT_LIST_HEAD(&nook->locals->cache_list);
  skb_queue_head_init(&nook->resources.skb_pool);
  skb_queue_head_init(&nook->locals->skb_pool);
  skb_queue_head_init(&nook->locals->skb_rx_list);
  INIT_LIST_HEAD(&nook->locals->skb_free_list);

//  PDEBUG(" pagemap empty = %d\n",list_empty(&nook->resources.page_map));
/*
  memset(&nook->resources.cache_map, 0, sizeof(nooks_cache_map_t));
  memset(&nook->locals->cache_map, 0, sizeof(nooks_local_cache_map_t));
*/
#if 0
  //
  // Set up the buckets for the slab allocator
  //

  for (i = 0; i < NOOKS_NUM_CACHE_SIZES - 1; i++) {
    nook->resources.cache_map.caches[i].cs_size = size;
    nook->locals->cache_map.caches[i].cs_size = size;
    size = size << 1;
  }
  printk ("Omitted till for loop.\n");
#endif   
//  nook->resources.cache_map.caches[0].cs_size = size;
//  nook->locals->cache_map.caches[0].cs_size = size;
  

  if (result != 0) {
    goto Cleanup;
  }


 Cleanup:
  
  NOOKS_EXIT();

  return(result);

}

/**********************************************************
 *
 * nooks_mm_cleanup_nook
 *
 * Clean up the memory management state of a nook
 *
 *********************************************************/

void
nooks_mm_cleanup_nook(pnook_t nook)
{
  int i;
  struct list_head * next, *spare;
  kmem_cache_t * temp;


  //
  // Loop through all the local caches and try to free them
  //

  for (i = 0; i < NOOKS_NUM_CACHE_SIZES; i++) {
    if ( nook->resources.cache_map.caches[i].cs_cachep != NULL) {
      kmem_cache_destroy(nook->resources.cache_map.caches[i].cs_cachep);
      nook->resources.cache_map.caches[i].cs_cachep = NULL;
    }
    if ( nook->resources.cache_map.caches[i].cs_dmacachep != NULL) {
      kmem_cache_destroy(nook->resources.cache_map.caches[i].cs_dmacachep);
      nook->resources.cache_map.caches[i].cs_dmacachep = NULL;
    }

  }

  //
  // Loop through the caches and free them
  //
  list_for_each_safe(next, spare, &nook->resources.cache_list) {
    temp = (kmem_cache_t *)list_entry(next,struct kmem_cache, next);
    kmem_cache_destroy(nook->resources.cache_map.caches[i].cs_dmacachep);
  }
  //
  // Loop through the memory areas and free them
  //

/*
  if ((nook != nooks_global_nook) && (nook->resources.mm != NULL)) {
    mmdrop(nook->resources.mm);
  }
*/
}

/*
 * Release an obj back to its cache. If the obj has a constructed
 * state, it should be in this state _before_ it is released.
 * - caller is responsible for the synchronization
 */

#if DEBUG
# define CHECK_NR(pg)						\
	do {							\
		if (!VALID_PAGE(pg)) {				\
			PERROR(KERN_ERR "kfree: out of range ptr %lxh.\n", \
				(unsigned long)objp);		\
			BUG();					\
		} \
	} while (0)
# define CHECK_PAGE(page)					\
	do {							\
		if (!PageSlab(page)) {				\
			PERROR(KERN_ERR "kfree: bad ptr %lxh.\n", \
				(unsigned long)objp);		\
			BUG();					\
		}						\
	} while (0)

#else
# define CHECK_PAGE(pg)	do { } while (0)
#endif
/*

#define	SET_PAGE_CACHE(pg,x)  ((pg)->list.next = (struct list_head *)(x))
#define	GET_PAGE_CACHE(pg)    ((kmem_cache_t *)(pg)->list.next)
#define	SET_PAGE_SLAB(pg,x)   ((pg)->list.prev = (struct list_head *)(x))
#define	GET_PAGE_SLAB(pg)     ((slab_t *)(pg)->list.prev)

*/

/*
 * Functions for storing/retrieving the cachep and or slab from the page
 * allocator.  These are used to find the slab an obj belongs to.  With kfree(),
 * these are used to find the cache which an obj belongs to.
 */
static inline void page_set_cache(struct page *page, struct kmem_cache *cache)
{
        page->lru.next = (struct list_head *)cache;
}



static inline struct kmem_cache *page_get_cache(struct page *page)
{
        if (unlikely(PageCompound(page)))
                page = (struct page *)page_private(page);
        BUG_ON(!PageSlab(page));
        return (struct kmem_cache *)page->lru.next;
}


static inline void page_set_slab(struct page *page, struct slab *slab1)
{
        page->lru.prev = (struct list_head *)slab1;
}

static inline struct slab *page_get_slab(struct page *page)
{
        if (unlikely(PageCompound(page)))
                page = (struct page *)page_private(page);
        BUG_ON(!PageSlab(page));
        return (struct slab *)page->lru.prev;
}




/**********************************************************
 *
 * nooks_mm_kfree
 *
 * Frees kmalloc'd memory within a nook
 * called in nook context
 *
 *********************************************************/

void 
nooks_mm_kfree (pnook_t nook, const void *objp)
{
  kmem_cache_t *c = NULL; // , *cl = NULL;
  //  unsigned long flags;
  // int i;
  struct page * old_mem_cache_page;
  NOOKS_ENTER();
  if (!objp)
    goto Cleanup;
  // local_irq_save(flags);

  old_mem_cache_page = virt_to_page(objp);
  CHECK_PAGE(virt_to_page(objp));
  c = page_get_cache(virt_to_page(objp));
  kmem_cache_free(c, (void*)objp);

 Cleanup:
  NOOKS_EXIT();
}


/**********************************************************
 *
 * nooks_mm_create_cache 
 *
 * create the slab cache for a particular size
 *
 * calling context: nook
 *
 *********************************************************/

int 
nooks_k_mm_create_cache(int index, int flags)
{
  char name[20];
  size_t size;
  int is_dma;
  kmem_cache_t  ** cache = NULL;
  kmem_cache_t * cacheS = NULL; // cacheS always is a shadow of cache
  kmem_cache_t  * local_cache = NULL;
  int cache_flags = SLAB_HWCACHE_ALIGN | SLAB_NOOK;
  int result = 0;
  pnook_t nook = GET_CURRENT_NOOK();
  int list_count = 0;
  kmem_cache_t a;
  local_cache = &a;
  cache = &local_cache;
  cacheS = *cache;
  //*cache = &a;
  //*local_cache = **cache;
  short in_nook = 0;

  if (IN_NOOK()) 
  {
    in_nook = 1;
    EXIT_NOOK();
  } 
 
  NOOKS_KWRAPPER_ENTER();

  //
  // BUGBUG: context switch to kernel
  //

  size = nook->resources.cache_map.caches[index].cs_size;
  if (size == NULL)
  {
     result = -ENOMEM;
     goto Cleanup;
  }
  else
    printk("size is %l.\n",size);
  is_dma = (flags & GFP_DMA) ? 1 : 0;

  if (nook == NULL)
  {
    result = -ENOMEM;
    goto Cleanup;
  }

  if (nook->locals == NULL)
  {
    printk("Nook not initialized correctly");
    result = -ENOMEM;
    goto Cleanup;
  }



  sprintf(name,"size-%Zd:%c",size, is_dma ? 'd' : 'n');

  printk(" Creating cache named %s\n",name);

  cacheS = (kmem_cache_t *) kmem_cache_create(name, 
			     size, 
			     0,     // offset
			     cache_flags, 
			     NULL, // ctor
			     NULL); // dtor

  //cacheS = *cache;
  if (cacheS == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }

  printk("Created cache %p\n",cacheS);

   if (is_dma) {
    nook->resources.cache_map.caches[index].cs_dmacachep = cacheS;
    nook->locals->cache_map.caches[index].cs_dmacache = *cacheS;
    cache_flags |= SLAB_CACHE_DMA;
  } else {
    nook->resources.cache_map.caches[index].cs_cachep = cacheS;
    nook->locals->cache_map.caches[index].cs_cache = *cacheS;
  }

  
  if (cacheS->flags)
    printk("Cache flags = 0x%x\n",(cacheS)->flags);
  

 Cleanup:
  NOOKS_KWRAPPER_EXIT();
  if (in_nook)
	ENTER_NOOK();
  return(result);
}


int 
nooks_mm_create_cache(pnook_t nook, int index, int flags)
{
  int result = 0;

  NOOKS_ENTER();

  result =  NOOKS_KERNEL_CALL2(nooks_k_mm_create_cache,
			       index,
			       flags);

  NOOKS_EXIT();
  return(result);
}


/**********************************************************
 * nooks_mm_kmalloc - allocate memory within a nook
 * @size: how many bytes of memory are required.
 * @flags: the type of memory to allocate.
 *
 * kmalloc is the normal method of allocating memory
 * in the kernel.
 *
 * The @flags argument may be one of:
 *
 * %GFP_USER - Allocate memory on behalf of user.  May sleep.
 *
 * %GFP_KERNEL - Allocate normal kernel ram.  May sleep.
 *
 * %GFP_ATOMIC - Allocation will not sleep.  Use inside interrupt handlers.
 *
 * Additionally, the %GFP_DMA flag may be set to indicate the memory
 * must be suitable for DMA.  This can mean different things on different
 * platforms.  For example, on i386, it means that the memory must come
 * from the first 16MB.
 *
 *
 *********************************************************/


void * 
nooks_mm_kmalloc (pnook_t nook, size_t size, int flags)
{
  int i;
  void * result = NULL;
  nooks_local_cache_sizes_t *csizep = NULL;
  printk("In nooks_mm_kmalloc.\n");
  
  if (nook == NULL)
  {
	printk ("Nook is null.\n");
	goto Exit;
  }
  if (nook->locals == NULL)
  	goto Exit;

  if (nook->locals->cache_map.caches == NULL)
  	goto Exit;

  csizep = nook->locals->cache_map.caches;
  NOOKS_ENTER();

  for (i = 0; csizep[i].cs_size; i++) {
    if (size > csizep[i].cs_size)
      continue;

    //
    // BUGBUG: how do we get the kernel cache structure to be
    // updated in this case?
    //

    if (flags & GFP_DMA) {
      if (nook->resources.cache_map.caches[i].cs_dmacachep == NULL) {
	printk(" Cache size %d doesn't exist - creating \n",
	       csizep[i].cs_size);
	nooks_mm_create_cache(nook, i, flags);
      }
      if (nook->resources.cache_map.caches[i].cs_dmacachep != NULL) {
	result =  kmem_cache_alloc(&csizep[i].cs_dmacache, flags);
	goto Cleanup;
      }
    } else {
      if (nook->resources.cache_map.caches[i].cs_cachep == NULL) {
	nooks_mm_create_cache(nook, i, flags);
      }
      if (nook->resources.cache_map.caches[i].cs_cachep != NULL) {
	result = kmem_cache_alloc(&csizep[i].cs_cache, flags);
	goto Cleanup;
      }
    }
  }
 Cleanup:
  NOOKS_EXIT();
 Exit:
  return result;
}


/**********************************************************
 *
 * nooks_mm_find_map_from_address
 *
 * Find the mm_map structure that contains the specified address
 *
 * calling mode: kernel, holding lock
 *
 *********************************************************/

pnooks_mm_map_t
nooks_mm_find_map_from_address(pnook_t nook,
			       struct list_head * list,
			       unsigned long address)
{
  struct list_head * next;
  pnooks_mm_map_t map, found = NULL;
  //  unsigned long flags;

  // NOOKS_ENTER();
  //  NOOKS_LOCK_RESOURCES(nook, flags);
  list_for_each(next, list) {
    map = list_entry(next,nooks_mm_map_t, next);
    if ((map->address <= address) && 
	(map->address + map->size > address)) {
      found = map;
      break;
    }
  }
  //  NOOKS_UNLOCK_RESOURCES(nook, flags);
  // NOOKS_EXIT();
  return(found);
}

/**********************************************************
 *
 * nooks_mm_find_adjacent_maps
 *
 * Find the mm_map structures that are adjacent to the
 * requested region.
 *
 * calling mode: must hold resources lock
 *
 *********************************************************/

void
nooks_mm_find_adjacent_maps(pnook_t nook,
			    struct list_head * list,
			    unsigned long start_addr,
			    unsigned long end_addr,
			    pnooks_mm_map_t * start_map,
			    pnooks_mm_map_t * end_map)
{
  struct list_head * next;
  pnooks_mm_map_t map;
  //  unsigned long flags;
  int found = 0;

  // NOOKS_ENTER();
  //  NOOKS_LOCK_RESOURCES(nook, flags);
  list_for_each(next, list) {
    map = list_entry(next,nooks_mm_map_t, next);
    if (map->address == end_addr) {
      found |= 1;
      *end_map = map;
    } else if (map->address + map->size == start_addr) {
      found |= 2;
      *start_map = map;
    }
    if (found == 3) {
      break;
    }
  }
	
  // NOOKS_UNLOCK_RESOURCES(nook, flags);
  // NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_mm_unlink_map
 *
 * Delete a map entry from a list
 *
 * calling mode: kernel, holding resource lock
 *
 *********************************************************/
    
void
nooks_mm_unlink_map(pnook_t nook, 
		    struct list_head * list,
		    pnooks_mm_map_t map)
{
  //  unsigned long flags;
  // NOOKS_ENTER();
  //  NOOKS_LOCK_RESOURCES(nook, flags);

  list_del_init(&map->next);
  nook->resources.maps_qlen--;
  
  //  NOOKS_UNLOCK_RESOURCES(nook, flags);
  // NOOKS_EXIT();
}

struct nooks_mm_log_entry {
  struct list_head next;
  unsigned long address;
  unsigned long size;
  void * list;
  int command;
};

LIST_HEAD(nooks_mm_log_list);

void nooks_mm_log_command(unsigned long address, 
			  unsigned long size, 
			  void * list,
			  int command)
{
#if 0
  struct nooks_mm_log_entry * entry;

  entry = nooks_kmalloc(sizeof(struct nooks_mm_log_entry), NOOKS_MEM_FLAGS);
  if (entry != NULL) {
    entry->address = address;
    entry->size = size;
    entry->command = command;
    entry->list = list;
    list_add_tail(&entry->next, &nooks_mm_log_list);
  }
#endif
}

unsigned long nooks_mm_oversize = 0;
unsigned long nooks_mm_duplicates = 0;
pnooks_mm_map_t nooks_mm_extra_map = NULL;


#define NOOKS_MM_MAP_COALESCE 1
#define NOOKS_MM_MAP_RECURSE  2

#ifdef NOOKS_DEBUG2
void
nooks_mm_validate_map(pnook_t nook, struct list_head * list)
{
  pnooks_mm_map_t outer_map;
  pnooks_mm_map_t inner_map;
  struct list_head * outer;
  struct list_head * inner;
  int found;

  list_for_each(outer, list) {
    outer_map = list_entry(outer,nooks_mm_map_t, next);
    found = 0;
    list_for_each(inner, list) {
      inner_map = list_entry(inner,nooks_mm_map_t, next);
      if (outer_map == inner_map) {
	if (!found) {
	  found = 1;
	} else {
	  BUG();
	}
      }
    }
    {
      pnooks_lookaside_elem_t next = nook->resources.map_pool.list;
      while (next != NULL) {
	if (next == (pnooks_lookaside_elem_t) outer_map) {
	  BUG();
	}
	next = next->next;
      }
  }

    
  }
}
#else
#define nooks_mm_validate_map(_x_, _y_)
#endif // NOOKS_DEBUG2
/**********************************************************
 *
 * nooks_mm_get_map 
 *
 * Create a mm_map structure to hold the range of pages
 * assigned to the nook
 *
 * calling mode: kernel, not holding lock if empty_map is null
 *
 *********************************************************/

pnooks_mm_map_t 
nooks_mm_get_map(pnook_t nook,
		 struct list_head * list,
		 pnooks_mm_map_t empty_map,
		 int flags,
		 int prot,
		 unsigned long address,
		 unsigned long size,
		 int mm_flags)
{
  pnooks_mm_map_t map = NULL;
  pnooks_mm_map_t start_map = NULL, end_map = NULL;
  unsigned long irqflags = 0;
  int free_end_map = 0;

  if ((mm_flags & NOOKS_MM_MAP_RECURSE) == 0) {
    PDEBUG("getting address %lx, size %lx\n",address, size);
    nooks_mm_log_command(address, size, list, 0);
    if (size > 0x1001) {
      nooks_mm_oversize++;
    }
  }


  NOOKS_ENTER();
  if (empty_map == NULL) {
    NOOKS_LOCK_RESOURCES(nook, irqflags);
  }

  if (nooks_mm_find_map_from_address(nook, list, address )) {
    nooks_mm_duplicates++;
  }

  //
  // look for adjacent map for coalescing?
  //

  if ((mm_flags & NOOKS_MM_MAP_COALESCE) != 0) {
    nooks_mm_find_adjacent_maps(nook, 
				list, 
				address, 
				address+size, 
				&start_map, 
				&end_map);
  }

  //
  // only allocate a new map entry if we don't have an adjacent one.
  //
  
  if ((start_map != NULL) && (start_map->prot == prot)) {
    
    //
    // If we have both, we can coalesce.
    //

    if ((end_map != NULL) && (end_map->prot == prot)) {
      start_map->size = end_map->address + end_map->size - 
	start_map->address;

      //
      // Free the end map
      //
      nooks_mm_unlink_map(nook, list, end_map);
      free_end_map = 1;
    } else {
      //
      // merge with start map - just update size
      //
      start_map->size += size;
     
    }
    map = start_map;
  } else if ((end_map != NULL) && (end_map->prot == prot)) {
    //
    // Merge with end map - just update address
    //

    end_map->address -= size;
    end_map->size += size;
    map = end_map;
  } 
  if (empty_map == NULL) {
    NOOKS_UNLOCK_RESOURCES(nook, irqflags);
  }

  if (map == NULL) {
    if (empty_map == NULL) {
      map = (pnooks_mm_map_t) nooks_mm_lookaside_alloc(&nook->resources.map_pool);
      if (map == NULL) {
	goto Cleanup;
      }
    } else {
      map = empty_map;
    }
  
    
    map->address = address;
    map->size = size;
    map->prot = prot;
    
    if (empty_map != NULL) {
      list_add(&map->next, list);
      nook->resources.maps_qlen++;
    } else {
      NOOKS_LOCK_RESOURCES(nook, irqflags);
      list_add(&map->next, list);
      nook->resources.maps_qlen++;
      NOOKS_UNLOCK_RESOURCES(nook, irqflags);
    }
  }
  
  if (free_end_map) {

    //
    // BUGBUG: can we do this with interrupts disabled?
    //

    nooks_mm_lookaside_free(&nook->resources.map_pool, end_map);
  }

 Cleanup:

  nooks_mm_validate_map(nook,list);
  NOOKS_EXIT();
  return(map);
}


/**********************************************************
 *
 * nooks_mm_find_and_unlink_map
 *
 * Find the map starting at the address and delete it
 *
 * calling mode: kernel, not holding lock
 *
 *********************************************************/

pnooks_mm_map_t
nooks_mm_find_and_unlink_map(pnook_t nook,
			     struct list_head * list,
			     unsigned long address)
{
  struct list_head * next;
  pnooks_mm_map_t map, found_map = NULL;
  unsigned long flags;
  NOOKS_ENTER();

  NOOKS_LOCK_RESOURCES(nook, flags);
  list_for_each(next, list) {
    map = list_entry(next,nooks_mm_map_t, next);
    if (map->address == address) {
      list_del_init(&map->next);
      found_map = map;
      nook->resources.maps_qlen--;
      break;
    }
  }
  nooks_mm_validate_map(nook,list);
  NOOKS_UNLOCK_RESOURCES(nook, flags);
  NOOKS_EXIT();
  return(found_map);
}	



/**********************************************************
 *
 * nooks_mm_vmalloc
 *
 * mode: kernel
 *
 * allocates a range of virtually mapped pages and makes them
 * accessible to the driver
 *********************************************************/

void *
nooks_mm_vmalloc(pnook_t nook, 
		 unsigned long size, 
		 int gfp_mask, 
		 pgprot_t prot)
{
  void * result = NULL;
  int nooks_prot = (pgprot_val(prot) & _PAGE_RW) == 0 ? 
    NOOKS_READONLY : NOOKS_WRITEABLE;
  NOOKS_ENTER();

  result = __vmalloc(size, gfp_mask, prot);
  if (result == NULL) {
    goto Cleanup;
  }
  // added by doug
  memset (result, 0, size);
  // end doug
  NOOKS_STATS_ADD(NOOKS_OPS_PAGES_ALLOCED, size / PAGE_SIZE);

  //
  // Add this piece of memory to our list of owned memories
  //

  //Update Asim: We only stalk, we do not own.

  if (nooks_add_to_hash(&nook->resources.area_map, 
			     area_map_e,
			     result,
			     (void *) size) < 0) {
    vfree(result);
    result = NULL;
    goto Cleanup;
  }


  //  PDEBUG(" nooks_vmalloc updating pagetable\n");
 Cleanup:
  
  NOOKS_EXIT();
  return(result);
}

/**********************************************************
 *
 * nooks_mm_vfree
 *
 * mode: kernel
 *  
 * frees a range of virtuall mapped pages, based on the ones
 * that were allocated
 *
 *********************************************************/

void
nooks_mm_vfree(pnook_t nook,
	       void * address) 
{
  int result = 0;
  unsigned long size = 0;
  NOOKS_ENTER();
  result =  nooks_del_from_hash(&nook->resources.area_map, 
				area_map_e,
				address,
				(void **) &size );

  if (result < 0) {
    PERROR("Failed to find map for address %p\n", address);
    goto Cleanup;
  }
  
  //
  // Free the range
  //

  vfree(address);
  NOOKS_STATS_ADD(NOOKS_OPS_PAGES_FREED, size / PAGE_SIZE);

 Cleanup:
  NOOKS_EXIT();
    
}

/**********************************************************
 *
 * nooks_mm_add_pages
 *
 * Adds a range of pages to the nooks page table
 * 
 * context: kernel
 *
 *********************************************************/

int
nooks_mm_add_pages(pnook_t nook, 
		   unsigned long address,
		   unsigned int gfp_mask, 
		   unsigned long size)
{
  int result = 0;
  pnooks_mm_map_t map = NULL;
  NOOKS_ENTER();

 
  map = nooks_mm_get_map(nook, 
			 &nook->resources.page_map, 
			 NULL, // no empty map
			 gfp_mask,
			 NOOKS_WRITEABLE,
			 address,
			 size,
			 NOOKS_MM_MAP_COALESCE); // coalesce



  if (map == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }
    
  //nooks_mm_update_pagetable(nook, NOOKS_WRITEABLE, address, size);

 Cleanup:
  NOOKS_EXIT();
  return(result);
}

/**********************************************************
 *
 * nooks_mm_get_free_pages
 *
 * Allocates free pages from the kernel, adds them to the list
 * of nook-accessible pages
 * 
 * context: kernel
 *
 *********************************************************/

unsigned long 
nooks_mm_get_free_pages(pnook_t nook, 
			unsigned int gfp_mask, 
			unsigned int order)
{
  unsigned long address = 0;
  int result = 0;
  unsigned long size;
  NOOKS_ENTER();

  size = PAGE_SIZE << order;
  address = __get_free_pages(gfp_mask, order);
  if (address != 0) {
    result = nooks_mm_add_pages(nook, address, gfp_mask, size);
    if (result < 0) {      
      free_pages(address, order);
      address = 0;
      goto Cleanup;
    }    
  }
  NOOKS_STATS_ADD(NOOKS_OPS_PAGES_ALLOCED, (1 << order));

 Cleanup:
  NOOKS_EXIT();
  return(address);
}


/**********************************************************
 *
 * nooks_get_free_pages
 *
 * allocates pages of non-mapped memory for small object allocation
 * 
 * Context: kernel or nook
 *
 *********************************************************/

/* Helper functions from old wrappers added. */
unsigned long nooks_k___get_free_pages(unsigned int gfp_mask, unsigned int order)
{
  unsigned long result;
  NOOKS_KWRAPPER_ENTER();
  result =nooks_mm_get_free_pages(GET_CURRENT_NOOK(), gfp_mask, order);
  NOOKS_KWRAPPER_EXIT();
  return(result);
}

unsigned long nooks_mm_w_get_free_pages(unsigned int gfp_mask,
                                       unsigned int order)
{
        unsigned long result;

        result =
            NOOKS_KERNEL_CALL2(nooks_k___get_free_pages, gfp_mask, order);

        return (result);
}

unsigned long nooks_w___get_free_pages(unsigned int gfp_mask,
                                       unsigned int order)
{
        unsigned long result;

        if (order == 0) {
          pnook_t nook = GET_CURRENT_NOOK();
          pnooks_lookaside_list_t list = &nook->locals->page_pool;
          result = (unsigned long) nooks_mm_lookaside_alloc(list);
        } else {
          result = nooks_mm_w_get_free_pages(gfp_mask, order);
        }

        return (result);
}



unsigned long nooks_get_free_pages(unsigned int gfp_mask, unsigned int order)
{
  return(nooks_w___get_free_pages(gfp_mask, order));
#if 0
  unsigned long result;
  //
  // BUGBUG: how do I tell if I'm in a nook?
  //
  NOOKS_ENTER();
  result = NOOKS_KERNEL_CALL2(nooks_k_get_free_pages, gfp_mask, order);
  NOOKS_EXIT();
  return(result);
#endif
}

#define PageSetSlab(page)       set_bit(PG_slab, &(page)->flags)

/**********************************************************
 *
 * nooks_kmem_cache_grow
 *
 * BUGBUG we want to update the nooks version of the cache_t, but
 * in kernel mode we can't read from it.
 *
 *********************************************************/

void NOOKS_KWRAPPER(set_page_slab)(struct page * page, int gfp_order, void * cachep, void * slabp)
{
  pnook_t nook = GET_CURRENT_NOOK();
  unsigned long num_pages = 1 << gfp_order;
  pnooks_mm_map_t map = NULL;
  int i;
  NOOKS_KWRAPPER_ENTER();

  //
  // Validate parameters
  //

  if (page < mem_map) {
    PERROR("bad page parameter to set page slab: %p\n", page);
    goto Cleanup;
  }

  if (((unsigned long) page - (unsigned long) mem_map) % sizeof(struct page) != 0) {
    PERROR("page pointer not struct page size aligned: %p\n", page);
    goto Cleanup;
  }

  if (num_pages > max_mapnr) {
    PERROR("Invalid number of pages to set page slab: %ld, but only %ld physical\n",
	   num_pages,
	   max_mapnr);
    goto Cleanup;
  }


  map = nooks_mm_find_map_from_address(nook, 
				       &nook->resources.page_map, 
				       (unsigned long) page->virtual);
  if (map == NULL) {
    PERROR("trying to set slab on non-nooks memory\n");
    goto Cleanup;
  }

  if (((unsigned long) page->virtual) + (num_pages << PAGE_SHIFT) > 
      (map->address + map->size)) {
    PERROR("Trying to set slab on range not in map: %p, %lx\n",
	   page->virtual, num_pages);
    goto Cleanup;
  }

  
 

				       
  //
  // BUGBUG: look up pages in nooks page map, and slabp and cachep
  //
  i = 1 << gfp_order;
  do {
    page_set_cache(page, cachep);
    page_set_slab(page, slabp);
    PageSetSlab(page);
    page++;
  } while (--i);

 Cleanup:
  NOOKS_KWRAPPER_EXIT();
  return;
}




void nooks_set_page_slab (struct page * page, int gfp_order, void * cachep, void * slabp)
{
  NOOKS_ENTER();
  NOOKS_KERNEL_CALL4(nooks_k_set_page_slab, page, gfp_order, cachep, slabp);
  NOOKS_EXIT();
}



/**********************************************************
 *
 * nooks_mm_remove_map
 *
 * Unmap an area range that may span multiple maps, recursively
 *
 *********************************************************/

int
nooks_mm_remove_map(pnook_t nook, 
		    struct list_head * list,
		    unsigned long address, 
		    unsigned long size,
		    int recursion)
{
  int result = 0;
  int free_map = 0;
  pnooks_mm_map_t map = NULL;
  pnooks_mm_map_t new_map = NULL;
  pnooks_mm_map_t empty_map = NULL;
  unsigned long remaining_address = 0;
  unsigned long remaining_size = 0;
  unsigned long flags;


  if (recursion == 0) {
    nooks_mm_log_command(address, size, list, 2);
    PDEBUG("removing address %lx, size %lx\n",address, size);
  }
  NOOKS_ENTER();
  NOOKS_LOCK_RESOURCES(nook, flags);
  if (nooks_mm_extra_map != NULL) {
    empty_map = nooks_mm_extra_map;
    nooks_mm_extra_map = NULL;
  } else {
    NOOKS_UNLOCK_RESOURCES(nook, flags);
    empty_map = nooks_kmalloc(sizeof(nooks_mm_map_t), GFP_ATOMIC);
    if (empty_map == NULL) {
      result = -ENOMEM;
      goto Cleanup;
    }
    NOOKS_LOCK_RESOURCES(nook, flags);    
  }
  
  map = nooks_mm_find_map_from_address(nook, 
				       list, 
				       address);
  if (map == NULL) {
    NOOKS_UNLOCK_RESOURCES(nook, flags);
    nooks_mm_log_command(address, size, list, 3);
    PERROR("Failed to find map for page address %lx\n", address);
    result = -ENOENT;
    goto Cleanup;
  }

  //
  // If the starting addresses are the same, then 
  // If not all pages are being freed, update the map to reflect that
  //
  if (map->address == address) {
    if (size >= map->size) {
      //
      // perfect match - delete the entry
      //

      //
      // Remove the map size from the size, and add the map size
      // to the address
      //
      remaining_size = size - map->size;
      remaining_address = map->address + map->size;

      nooks_mm_unlink_map(nook, list, map);
      free_map = 1;

    } else {
      
      // 
      // Update the address and size to reflect that the first
      // few pages have been deleted
      //
      map->address = map->address + size;
      map->size = map->size - size;
    }
  } else {
    //
    // The starting addresses are different. There are two cases:
    // first case: the region being freed touches the end of
    // the existing region. Second case: the region being freed
    // cuts a hole in the middle of the existing region.

    // 
    // Case 1: the region freed doesn't leave a hole
    // Just update the size.
    //

    if ((map->address + map->size)  <= (address + size) ) {
      //
      // The new size changes by the difference in the starting
      // addresses.
      //

      //
      // Remaining size is the portion beyond the end of the map.
      // This may be zero if both regions end at the same address
      //

      remaining_size = (address + size) - (map->address + map->size);

      //
      // Remaining address is the last address of the map
      //

      remaining_address = map->address + map->size;

      //
      // New map size is the difference between the start of the map and the
      // the address of the freed memory
      //
      map->size = address - map->address;

    } else if ((address + size) < (map->address + map->size)) {
      int map_end = map->address + map->size;
      //
      // The old map now just maps the beginning of the region
      //

      map->size = address - map->address;

      //
      // Second case: the freed size is smaller and the ending addresses
      // and starting addresses are different, so split the map
      //

      new_map = nooks_mm_get_map(nook,
				 list,
				 empty_map,
				 GFP_ATOMIC,
				 map->prot,
				 address + size,
				 map_end - (address+size),
				 NOOKS_MM_MAP_COALESCE | NOOKS_MM_MAP_RECURSE);
      if (new_map == NULL) {
	if (nooks_mm_extra_map == NULL) {
	  nooks_mm_extra_map = empty_map;
	  empty_map = NULL;
	}
	NOOKS_UNLOCK_RESOURCES(nook, flags);
	result = -ENOMEM;
	goto Cleanup;
      }
      if (new_map == empty_map) {
	empty_map = NULL;
      }


    } 
  }

  if ((empty_map != NULL) && (nooks_mm_extra_map == NULL)) {
    nooks_mm_extra_map = empty_map;
    empty_map = NULL;
  }

  NOOKS_UNLOCK_RESOURCES(nook, flags);

  //
  // recursively handle any remainder
  //

  if (free_map) {

    nooks_mm_lookaside_free(&nook->resources.map_pool, map);
  }

  if (remaining_size > 0) {
    result = nooks_mm_remove_map(nook,
				 list,
				 remaining_address,
				 remaining_size,
				 recursion+1);
  }

 Cleanup:
  if (empty_map != NULL) {
    nooks_mm_lookaside_free(&nook->resources.map_pool, empty_map);
  }
  nooks_mm_validate_map(nook,list);

  NOOKS_EXIT();
  return(result);
}


/**********************************************************
 *
 * nooks_mm_releases_pages 
 *
 * Release a range of addresses from the nook and return them
 * to the kernel. This doesn't actually change the kernel's 
 * view of the pages.
 *
 *********************************************************/

int
nooks_mm_release_pages(pnook_t nook, unsigned long addr, unsigned long size)
{
 int result = 0;
    
 NOOKS_ENTER();
 PDEBUG("Releasing pages %lx size %ld\n", addr, size);
 
 result = nooks_mm_remove_map(nook, &nook->resources.page_map, addr, size, 0);
 if (result < 0) {
   PERROR(" failed to remove map %lx(%lx) : %x\n",
	  addr, size, result);
   goto Cleanup;
 }
 //
 // Make the range unreadable
 //
 //nooks_mm_update_pagetable(nook, NOOKS_READONLY, addr, size);
 
 Cleanup:
  return(result);
}

/**********************************************************
 *
 * Release all memory in use by the nook
 *
 * 
 *
 *********************************************************/

int
nooks_mm_free_page_map(pnook_t nook)
{
  struct list_head * next, *spare;
  pnooks_mm_map_t map;
  struct list_head * list = &nook->resources.page_map;
  unsigned long page;

  list_for_each_safe(next, spare, list) {
    map = list_entry(next,nooks_mm_map_t, next);
    for (page = 0; page < map->size; page += PAGE_SIZE) {
      
      free_pages(page + map->address, 0);      
    }
    nooks_mm_lookaside_free(&nook->resources.map_pool, map);

  }
  return(0);
}

/**********************************************************
 *
 * nooks_mm_free_pages
 *
 * Free pages and update nooks pagetable
 *
 *********************************************************/

void 
nooks_mm_free_pages(pnook_t nook, unsigned long addr, unsigned int order)
{
  unsigned long size = PAGE_SIZE << order;
  int result = 0;
  
  NOOKS_ENTER();
  PDEBUG("Freeing pages %lx order %d\n", addr, order);
  
  result = nooks_mm_release_pages(nook, addr, size);
  if (result < 0) {
    PERROR(" failed to release map %lx(%lx) : %x\n",
	   addr, size, result);
    goto Cleanup;
  }

  free_pages(addr, order);
  NOOKS_STATS_ADD(NOOKS_OPS_PAGES_FREED, (1 << order));

Cleanup:

  NOOKS_EXIT();

}

/**********************************************************
 *
 * nooks_free_pages
 *
 * Callable function from kernel or nook to free unmapped
 * pages and update pagetable.
 *
 *********************************************************/

void
nooks_free_pages(unsigned long addr, unsigned int order)
{
#if 0
  nooks_w_free_pages(addr, order);
#endif
  NOOKS_ENTER();
  if (addr != 0) {
    nooks_mm_free_pages(GET_CURRENT_NOOK(), addr, order);
  }
  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_mm_kmem_cache_create 
 *
 * Creates a slab cache locally for a nook.
 *
 * context: kernel (but calls kmalloc in nook?)
 * BUGBUG: touches nook memory and kernel memory - comingles
 *
 *********************************************************/


kmem_cache_t * NOOKS_KWRAPPER(mm_kmem_cache_create)(const char * name, 
			     size_t size, 
			     size_t offset, 
			     unsigned long flags,
			     kmem_cache_t * local_cache,
			     void (*ctor)(void *, kmem_cache_t *, 
					  unsigned long),
			     void (*dtor)(void *, kmem_cache_t *, 
					  unsigned long))
{
  kmem_cache_t * kernel_cache = NULL;
  char * kernel_name = NULL;
  int result = 0;
  pnook_t nook = GET_CURRENT_NOOK();
			   
  NOOKS_KWRAPPER_ENTER();

  //
  // Copy the driver name into the kernel
  //

  result = nooks_get_driver_string(name, &kernel_name);
  if (result < 0) {
    goto Cleanup;
  }
  // 
  // Turn on the nooks flag so the memory is allocated from the right
  // place.
  //

  flags |= SLAB_NOOK;  
  kernel_cache = kmem_cache_create(name, size, offset, flags, ctor, dtor);
  if (kernel_cache == NULL) {
    goto Cleanup;
  }

  //
  // Add the kernel cache to the per-nook list
  //

  list_add(&kernel_cache->next, &nook->resources.cache_list);


  //
  // Copy the cache
  //

  nooks_memcpy_out(local_cache, kernel_cache, sizeof(struct kmem_cache));
  
  //
  // Switch back to the nook for the local work
  //


 Cleanup:
  if (kernel_name == NULL) {
    nooks_free_driver_string(kernel_name);
  }      
  NOOKS_KWRAPPER_EXIT();
  return(kernel_cache);
}

/**********************************************************
 *
 * nooks_mm_kmem_cache_create 
 *
 * Creates a slab cache locally for a nook.
 *
 * context: kernel (but calls kmalloc in nook?)
 * BUGBUG: touches nook memory and kernel memory - comingles
 *
 *********************************************************/

kmem_cache_t *
nooks_mm_kmem_cache_create(const char * name, 
			   size_t size, 
			   size_t offset, 
			   unsigned long flags,
			   void (*ctor)(void *, kmem_cache_t *, 
					unsigned long),
			   void (*dtor)(void *, kmem_cache_t *, 
					unsigned long))
{
  kmem_cache_t * kernel_cache = NULL, *local_cache = NULL;
  pnook_t nook = GET_CURRENT_NOOK();
  int list_count = 0;			   
  NOOKS_ENTER();

  //
  // Now allocate local buffer for the cache
  //

  local_cache = nooks_w_kmalloc( sizeof(struct kmem_cache), 
				 GFP_KERNEL);
  if (local_cache == NULL) {
    goto Cleanup;
  }


  kernel_cache = (kmem_cache_t *) NOOKS_KERNEL_CALL7(nooks_k_mm_kmem_cache_create,
						     name,
						     size,
						     offset,
						     flags,
						     local_cache,
						     ctor,
						     dtor);
  if (kernel_cache == NULL) {
    nooks_w_kfree(local_cache);
    local_cache = NULL;
    goto Cleanup;
  }
  // 
  // update the pointers & add it to the list in the nook
  //

  /* Old fields gone, new fields need to be fixed. */
  FIX_LIST(&(kernel_cache)->next,&local_cache->next);

  for (list_count=0; list_count<MAX_NUMNODES; list_count++)
  {
        FIX_LIST(&(kernel_cache)->nodelists[list_count]->slabs_full,&local_cache->nodelists[list_count]->slabs_full);
        FIX_LIST(&(kernel_cache)->nodelists[list_count]->slabs_partial,&local_cache->nodelists[list_count]->slabs_partial);
        FIX_LIST(&(kernel_cache)->nodelists[list_count]->slabs_free,&local_cache->nodelists[list_count]->slabs_free);
  }

  /*
  FIX_LIST(&kernel_cache->slabs_full,&local_cache->slabs_full);
  FIX_LIST(&kernel_cache->slabs_partial,&local_cache->slabs_partial);
  FIX_LIST(&kernel_cache->slabs_free,&local_cache->slabs_free);

  local_cache->back_pointer = kernel_cache;
  */

  list_add(&local_cache->next, &nook->locals->cache_list);

 Cleanup:

  NOOKS_EXIT();
  return(local_cache);
}

/**********************************************************
 *
 * nooks_k_mm_kmem_cache_destroy
 *
 * destroys a cache from the local list and in the kernel
 *
 * mode: must be called from a nook
 * BUGBUG: touches nook memory and kernel memory - comingles
 *
 *********************************************************/

int NOOKS_KWRAPPER(mm_kmem_cache_destroy)(kmem_cache_t * cache)
{
  pnook_t nook = GET_CURRENT_NOOK();
  int result = 0;
  struct list_head * next;
  kmem_cache_t * kernel_cache = NULL;
  kmem_cache_t * temp;

  NOOKS_KWRAPPER_ENTER();


  //
  // Look for the kernel cache based on the copy of the local cache pointer
  //

  list_for_each(next, &nook->resources.cache_list) {
    temp = (kmem_cache_t *) list_entry(next,struct kmem_cache, next);
    if (temp == cache) {
      kernel_cache = temp;
      break;
    }
  }
  if (kernel_cache == NULL) {
    PERROR(" unable to find kernel cache: %p\n", cache);
    result = -ENOENT;
    goto Cleanup;
  }

  list_del_init(&kernel_cache->next);

  result = kmem_cache_destroy(kernel_cache);


 Cleanup:

  NOOKS_KWRAPPER_EXIT();
  return(result);
  
}

/**********************************************************
 *
 * nooks_mm_kmem_cache_destroy
 *
 * destroys a cache from the local list and in the kernel
 *
 * mode: must be called from a nook
 * BUGBUG: touches nook memory and kernel memory - comingles
 *
 *********************************************************/

int
nooks_mm_kmem_cache_destroy(kmem_cache_t * cache)
{
  pnook_t nook = NULL;
  int result = 0;
  struct list_head * next;
  kmem_cache_t * local_cache = NULL;
  kmem_cache_t * local_kernel_cache = NULL;
  kmem_cache_t * temp;

  NOOKS_ENTER();

  //
  // We must start out in the nook to be safe
  //


  nook = GET_CURRENT_NOOK();

  //
  // Find the cache in the local list
  //

  list_for_each(next, &nook->locals->cache_list) {
    temp = list_entry(next,struct kmem_cache, next);
    if (temp == cache) {
      local_cache = temp;
      break;
    }
  }
  if (local_cache == NULL) {
    PERROR(" unable to find local cache: %p\n", cache);
    result = -ENOENT;
    goto Cleanup;
  }

  list_del_init(&local_cache->next);

  //local_kernel_cache = local_cache->back_pointer;
  //
  // Free the local cache
  //
  nooks_w_kfree(local_cache);

  result = NOOKS_KERNEL_CALL1(nooks_k_mm_kmem_cache_destroy, 
			      local_kernel_cache);
 Cleanup:

  NOOKS_EXIT();
  return(result);
  
}

/**********************************************************
 *
 * Function: nooks_kw_malloc
 * 
 * Description: allocate nook memory from kernel
 * 
 * Notes:
 *
 *********************************************************/

void *
nooks_kw_malloc(size_t size, int flags) 
{
  void * result;
  NOOKS_ENTER_DRIVER();
  result = (void *) NOOKS_PLAIN_DRIVER_CALL2(nooks_w_kmalloc, size, flags);
  NOOKS_LEAVE_DRIVER();
  if (IS_ERR(result)) {
    result = NULL;
  }
  return(result);
}

/**********************************************************
 *
 * Function: nooks_kw_free
 * 
 * Description: frees nook memory from kernel
 * 
 * Notes:
 *
 *********************************************************/

void
nooks_kw_free(void * ptr)
{
  NOOKS_ENTER_DRIVER();
  NOOKS_PLAIN_DRIVER_CALL1(nooks_w_kfree, ptr);
  NOOKS_LEAVE_DRIVER();
}
