/*
 * nooks-data.c -- Routines for managing nook and driver data
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

 
#include "nooks-i.h"
#include "nooks-iwrappers.h"
#include "nooks-wrappers.h"
#include "nooks-switch.h"
#include "nooks-workers.h"
#include "nooks-recover.h"
#include <linux/module.h>

spinlock_t nook_list_lock;


/**********************************************************
 *
 * nooks_find_nook_internal
 *
 * Finds a nook by name. Caller must be holding lock
 *
 **********************************************************/

pnook_t 
nooks_find_nook_internal(char * name)
{
  struct list_head * next;
  pnook_t nook = NULL;

  NOOKS_ENTER();
  
  list_for_each(next, &nook_list) {
    nook = list_entry(next,nook_t, next);
    PDEBUG(" Comparing names against '%s'\n",nook->name);
    if (strcmp(name, nook->name) == 0) {
      PDEBUG(" found nook!\n");
      goto Cleanup;
    }
  }
  nook = NULL;
 Cleanup:
  NOOKS_EXIT();
  return(nook);
}

/**********************************************************
 *
 * nooks_init_nook_resources
 *
 * Initializes per-nook resource lists
 *
 *********************************************************/

int
nooks_init_nook_resources(pnook_t nook)
{
  int result = 0;
  result = nooks_mm_init_nook(nook);
  if (result < 0) {
    PDEBUG(" Failed to init mm for nook: %d\n", result);
    goto Cleanup;
  }
  INIT_LIST_HEAD(&nook->resources.kernel_work_list.list);
  nook->resources.kernel_work_list.flags = 0;
  nook->resources.kernel_work_list.count = 0;

  spin_lock_init(&nook->resources.lock);
  spin_lock_init(&nook->lock);

  spin_lock_init(&nook->locals->lock);
  INIT_LIST_HEAD(&nook->locals->free_work_list);

  SET_CURRENT_NOOK(nook);

  nooks_mm_init_lookaside(&nook->resources.hash_pool,
			  200,
			  sizeof(nooks_hash_entry_t),
			  NOOKS_LOOKASIDE_KERNEL);

  nooks_mm_init_lookaside(&nook->resources.map_pool,
			  60,
			  sizeof(nooks_mm_map_t),
			  NOOKS_LOOKASIDE_KERNEL);

  nooks_mm_init_lookaside(&nook->resources.work_pool,
			  20,
			  sizeof(nooks_work_item_t),
			  NOOKS_LOOKASIDE_KERNEL);

  //
  // This must be initialized after the map pool, because
  // it depends on the map pool
  //

  nooks_mm_init_lookaside(&nook->resources.stack_pool,
			  20,
			  0, // order for NOOKS_STACK_SIZE,
			  NOOKS_LOOKASIDE_PAGE);

  nooks_mm_init_lookaside(&nook->locals->work_pool,
			  64,
			  sizeof(nooks_work_item_t),
			  NOOKS_LOOKASIDE_NOOK);

  nooks_mm_init_lookaside(&nook->locals->elem_pool,
			  20,
			  sizeof(nooks_list_elem_t),
			  NOOKS_LOOKASIDE_NOOK);

  nooks_mm_init_lookaside(&nook->locals->page_pool,
			  20,
			  0, 
			  NOOKS_LOOKASIDE_NOOK_PAGE);

  result = nooks_create_hash_table(&nook->resources.hash_pool,
				   &nook->resources.resource_list.hash, 
				   512,
				   1); // make it a dual
  if (result < 0) {
    PDEBUG("Failed to create hash table: %d\n", result);
    goto Cleanup;
  }
  result = nooks_create_hash_table(&nook->resources.hash_pool,
				   &nook->resources.active_skbs, 
				   512, //Changed from 128 by Asim
				   0); // not a dual
  if (result < 0) {
    PDEBUG("Failed to create hash table: %d\n", result);
    goto Cleanup;
  }


 printk ("Nook successfully initialized ");
 Cleanup:
  SET_CURRENT_NOOK(NULL);
  return(result);
}


/**********************************************************
 *
 * nooks_i_create_nook
 *
 * Initialize a nook data structure and return it with one
 * reference
 *
 **********************************************************/

int 
nooks_i_create_nook(char *name, pnook_t * ret_nook)
{
  pnook_t nook = NULL;
  pnook_t dup_nook = NULL;
  int result = 0;
  NOOKS_ENTER();

  *ret_nook = NULL;

  printk(" Creating nook named %s\n",name);
  //
  // Check for duplicates in advance (avoid doing a lot of work then
  // throw it away
  //


  spin_lock(&nook_list_lock);

  dup_nook = nooks_find_nook_internal(name);
  if (dup_nook != NULL) {
    result = -EEXIST;
  }
    
  
  spin_unlock(&nook_list_lock);
  if  (result < 0) {
    goto Cleanup;
  }

  nook = (pnook_t) nooks_kmalloc(sizeof(nook_t), NOOKS_MEM_FLAGS);
  if (nook == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }

  //
  // Fill in all lists, so cleanup can be sure that they are valid
  //

  memset(nook, 0, sizeof(nook_t));
  INIT_LIST_HEAD(&nook->next);
  INIT_LIST_HEAD(&nook->drivers);
  nook->state = nook_state_init;
  nook->recovery_state.info.recover = NULL;
  nook->recovery_state.info.driver_recovery = nooks_driver_recovery;
  nook->recovery_state.info.resource_recovery = nooks_resource_recovery;
#ifdef CONFIG_NOOKS_PROC
  nook->shadow_driver_name = "no";
  nook->proc_ent  = NULL;
  nook->last_fault_mod  = NULL;
#endif

  //
  // Fill in the fields of the nook
  //

  if (strlen(name) > NOOK_MAX_NAME_LEN) {
    result = -ENAMETOOLONG;
    goto Cleanup;
  }
  nook->name = nooks_strdup(name, GFP_KERNEL);
  if (nook->name == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }

  //
  // Allocate a page of local resources
  //

  nook->locals = (pnook_locals_t) get_zeroed_page(GFP_KERNEL);
  if (nook->locals == 0) {
    result = -ENOMEM;
    goto Cleanup;
  }

  result = nooks_init_nook_resources(nook);
  if (result < 0) {
    goto Cleanup;
  }

  //
  // Add the nook to the list
  //

  spin_lock(&nook_list_lock);

  dup_nook = nooks_find_nook_internal(name);
  if (dup_nook == NULL) {
    nook->refs = 1;
    list_add(&nook->next, &nook_list);
    *ret_nook = nook;
  } else {
    result = -EEXIST;
  }
    
  spin_unlock(&nook_list_lock);

  if (result < 0) {
    goto Cleanup;
  }


  nook->state = nook_state_running;

  //
  // Initialize the lookaside lists for all but the global nook
  //

  if (ret_nook != &nooks_global_nook) {
    SET_CURRENT_NOOK(nook);
    NOOKS_ENTER_DRIVER();
    result = NOOKS_PLAIN_DRIVER_CALL0(nooks_w_init_pools);
    NOOKS_LEAVE_DRIVER();
    SET_CURRENT_NOOK(NULL);
  }
 
 printk ("Nook successfully created.\n");  
 Cleanup:
  if (result < 0) {
    if (nook != NULL) {
      nooks_free_nook(nook);
    }
    *ret_nook = NULL;
  } 
  NOOKS_EXIT();
  return(result);
}


/**********************************************************
 *
 * nooks_free_nook
 *
 * Frees a nook allocate by nooks_i_create_nook
 *
 **********************************************************/

void
nooks_free_nook(pnook_t nook)
{
  NOOKS_ENTER();
  
  //
  // Clean the higher-level resources first
  //

  nooks_clean_nook(nook);

  //
  // Remove from list
  //

  spin_lock(&nook_list_lock);
  list_del_init(&nook->next);
  spin_unlock(&nook_list_lock);
  if (nook->name != NULL) {
    nooks_kfree(nook->name);

  }

  if (nook->locals != NULL) {
    free_page((unsigned long) nook->locals);
  }
  nooks_kfree(nook);
  NOOKS_EXIT();
}

/**********************************************************
 *
 *  nooks_reference_nook 
 *
 * 
 *
 *********************************************************/

void
nooks_reference_nook(pnook_t nook)
{
  NOOKS_REFERENCE(nook);
}

void
nooks_dereference_nook(pnook_t nook)
{
  if (NOOKS_DEREFERENCE(nook) == 0) {
    nooks_free_nook(nook);
  }
}

/**********************************************************
 *
 * nooks_find_nook
 *
 * Finds a nook by name, and references it. Caller must
 * dereference.
 *
 **********************************************************/

pnook_t 
nooks_find_nook(char * name)
{
  pnook_t nook = NULL;

  NOOKS_ENTER();
  
  PDEBUG("Finding nook named '%s'\n", name);

  nook = nooks_find_nook_internal(name);
  if (nook != NULL) {
    nooks_reference_nook(nook);
  }
  spin_unlock(&nook_list_lock);
  NOOKS_EXIT();
  return(nook);
}

/**********************************************************
 *
 * nooks_create_driver
 *
 * Initialize a nook data structure and return it with one
 * reference
 * holds on to a reference to name
 *
 **********************************************************/

pnook_driver_t
nooks_create_driver(pnook_t nook, 
		    nooks_iwrapper_type_t driver_type, 
		    int major, 
		    char * name)
{
  pnook_driver_t driver = NULL;
  nooks_iwrapper_itf itf = NULL;
  NOOKS_ENTER();

  driver = (pnook_driver_t) nooks_kmalloc(sizeof(nook_driver_t), NOOKS_MEM_FLAGS);
  if (driver == NULL) {
    goto Cleanup;
  }

  memset(driver, 0, sizeof(nook_driver_t));

  //
  // Fill in the fields of the nook
  //

  driver->major = major;
  driver->nook = nook;
  driver->name = name;
  driver->driver_type = driver_type;
  driver->parent = GET_CURRENT_NOOK_DRIVER();
  INIT_LIST_HEAD(&driver->next_in_shadow);
  INIT_LIST_HEAD(&driver->next_in_nook);
  INIT_LIST_HEAD(&driver->next);
  
  list_add(&driver->next_in_nook, &nook->drivers);
  list_add(&driver->next, &nook_driver_list);


  itf = nooks_iwrapper_get_itf(driver_type, driver);
  if (itf == NULL) {
    PERROR(" Failed to get itf for type %d\n", driver_type);
    nooks_free_driver(driver);
    driver = NULL;
    goto Cleanup;
  }
    
  driver->driver_type = driver_type;
  driver->iwrapper_operations.dummy_ops = itf;
  /*
  driver->iwrapper_operations.netops = itf;
  if (driver->iwrapper_operations.netops == NULL)
	printk ("Fatal: operations are null.\n");
  //driver->iwrapper_operations.mops = nooks_iwrapper_get_itf(mops_e, driver);
  */ 
  // Add the nook to the list
  //

  driver->refs = 1;

  spin_lock(&nook_list_lock);

  //
  // Preserve reverse order of lists
  //
    spin_unlock(&nook_list_lock);
 
 printk("ADDING DRIVER %s %d\n", driver->name, driver->major);
 
 Cleanup:

  NOOKS_EXIT();
  return(driver);
}

/**********************************************************
 *
 * 
 *
 * 
 *
 *********************************************************/

void
nooks_delayed_free_driver(pnook_driver_t driver)
{
  driver->need_delete = 1;
}

/**********************************************************
 *
 * nooks_free_driver
 *
 * Frees a nook allocate by nooks_create_driver
 *
 **********************************************************/

void
nooks_free_driver(pnook_driver_t driver)
{
  NOOKS_ENTER();
  // Remove from list
  //

  spin_lock(&nook_list_lock);
  list_del_init(&driver->next);
  list_del_init(&driver->next_in_nook);
  spin_unlock(&nook_list_lock);

  // Free any resources in use
  //

  if (driver->name != NULL) {
    nooks_kfree(driver->name);
  }
  // Free any iwrapper operations in use
  //

  if (driver->iwrapper_operations.dummy_ops != NULL) {
    nooks_iwrapper_release_itf(driver->driver_type, 
			       driver->iwrapper_operations.dummy_ops);
  }
  shadow_del_interface(driver);

  if (driver)
	  nooks_kfree(driver);
  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_find_driver_in_nook
 *
 * Finds a nook by major number among the drivers in a nook
 * and references it. Caller must dereference.
 *
 **********************************************************/

pnook_driver_t 
nooks_find_driver_in_nook(int major, pnook_t nook)
{
  struct list_head * next;
  pnook_driver_t driver = NULL, temp;

  NOOKS_ENTER();
  spin_lock(&nook_list_lock);

  list_for_each(next, &nook->drivers) {
    temp = list_entry(next,nook_driver_t, next_in_nook);
    if (temp->major == major) {
      driver = temp;
      break;
    }
  }
  spin_unlock(&nook_list_lock);

  NOOKS_EXIT();
  return(driver);
}

/**********************************************************
 *
 * nooks_find_driver_in_nook
 *
 * Finds a nook by major number among the drivers in a nook
 * and references it. Caller must dereference.
 *
 **********************************************************/

pnook_driver_t 
nooks_find_driver_by_type_in_nook(nooks_iwrapper_type_t driver_type,
				  void * data,
				  pnook_t nook)
{
  struct list_head * next;
  pnook_driver_t driver = NULL, temp;

  NOOKS_ENTER();
  spin_lock(&nook_list_lock);

  list_for_each(next, &nook->drivers) {
    temp = list_entry(next,nook_driver_t, next_in_nook);
    if ((temp->driver_type == driver_type) && 
	((data == NULL) || (temp->data == data))) {
      driver = temp;
      break;
    }
  }
  spin_unlock(&nook_list_lock);

  NOOKS_EXIT();
  return(driver);
}


/**********************************************************
 *
 * nooks_find_driver_in_nook
 *
 * Finds a nook by major number among the drivers in a nook
 * and references it. Caller must dereference.
 *
 **********************************************************/

pnook_driver_t 
nooks_find_driver_by_name_in_nook(char * name, 
				  nooks_iwrapper_type_t driver_type,
				  pnook_t nook)
{
  struct list_head * next;
  pnook_driver_t driver = NULL, temp;

  NOOKS_ENTER();
  spin_lock(&nook_list_lock);

  list_for_each(next, &nook->drivers) {
    temp = list_entry(next,nook_driver_t, next_in_nook);
    if ((temp->name != NULL) && (strcmp(temp->name, name) == 0) &&
	(temp->driver_type == driver_type)) {
      driver = temp;
      break;
    }
  }
  spin_unlock(&nook_list_lock);

  NOOKS_EXIT();
  return(driver);
}

/**********************************************************
 *
 * nooks_find_driver_by_data_in_nook
 *
 * Finds a nook by major number among the drivers in a nook
 * and references it. Caller must dereference.
 *
 **********************************************************/

pnook_driver_t 
nooks_find_driver_by_data_in_nook(void * data, 
				  nooks_iwrapper_type_t driver_type,
				  pnook_t nook)
{
  struct list_head * next;
  pnook_driver_t driver = NULL, temp;

  NOOKS_ENTER();
  spin_lock(&nook_list_lock);

  list_for_each(next, &nook->drivers) {
    temp = list_entry(next,nook_driver_t, next_in_nook);
    if ((temp->data == data) && (temp->driver_type == driver_type)) {
      driver = temp;
      break;
    }
  }
  spin_unlock(&nook_list_lock);

  NOOKS_EXIT();
  return(driver);
}

/**********************************************************
 *
 * nooks_find_driver
 *
 * Finds a nook by major number
 * and references it. Caller must dereference.
 *
 **********************************************************/

pnook_driver_t 
nooks_find_driver(int major)
{
  struct list_head * next;
  pnook_driver_t driver = NULL, temp;
  NOOKS_ENTER();

  spin_lock(&nook_list_lock);

  list_for_each(next, &nook_driver_list) {
    temp = list_entry(next,nook_driver_t, next);
    if (temp->major == major) {
      driver = temp;
      break;
    }
  }
  spin_unlock(&nook_list_lock);

  NOOKS_EXIT();
  return(driver);
}

/**********************************************************
 *
 * nooks_init_nook_module
 *
 * Initialize the information about a nook that relates to a
 * module
 *
 **********************************************************/

int 
nooks_init_nook_module(pnook_t nook, struct module * module)
{
	int result = 0;
	pnook_driver_t driver = NULL;
	char * driver_name = NULL;
	NOOKS_ENTER();

	driver_name = nooks_strdup(module->name, GFP_KERNEL);
	if (driver_name == NULL) {
		result = -ENOMEM;
		goto out;
	}

	driver = nooks_create_driver(nook, mops_e, NOOK_DRIVER_NO_MAJOR, driver_name);
	if (driver == NULL) {
		goto err_driver;
	}

	// Null this out because the driver holds on to it.
	//
	driver_name = NULL;
	if (driver->nook == NULL)
		driver->nook = nook;

	// Now the dangerous part: update the functions in the module
	// Ensure to leave the pointer NULL if it was NULL.
	//

	if (driver->iwrapper_operations.mops == NULL)
                printk("Fatal: mops are NULL in nooks_init_nook_module.\n");


	nooks_iwrapper_copy_mops(module,
			&driver->operations.mops, 
			driver->iwrapper_operations.mops, 
			module);
	
	
	driver->data = module;
#ifdef CONFIG_NOOKS_SHADOW
	result = shadow_add_smops(driver, module);
#endif
	printk ("DEBUG(nook_init_nook_module): init is at %p for module %s.\n", module->init, module->name);
	if (result < 0) {
	  printk ("shadow_add_smops: Result is %d.\n",result);
	  goto err_driver;
	}

	result = shadow_init_shadow_module(module, driver);

	if (result < 0) {
  	  printk ("shadow_init_shadow_module: Result is %d.\n", result);
	  goto err_driver;
	}

out:
	NOOKS_EXIT();
	return result;


err_driver:
	nooks_free_driver(driver);
	nooks_kfree(driver_name);
	NOOKS_EXIT();
	return result;
}

int nooks_unregister_module(pnook_t nook, char * module_name)
{
  int result = 0;
  pnook_driver_t driver = NULL;
  
  // Find the driver to clean it up.
  //

  driver = nooks_find_driver_by_name_in_nook(module_name, mops_e, nook);
  if (driver == NULL) {
    PERROR(" ERROR: no driver %d in nook %s, function %s\n",
	   major, nook->name, __FUNCTION__);
    result = -ENOENT;
  } else {
    nooks_free_driver(driver);
  }

  return(result);  
 
}


void
nooks_validate_nook(pnook_t nook)
{
  PDEBUG(" nook name: %s\n",nook->name);
}


/**********************************************************
 *
 * nooks_dump_nook_mm
 *
 * print the list of memory maps in the nook
 *
 *********************************************************/

void
nooks_dump_nook_mm(pnook_t nook)
{
  struct list_head * next;
  pnooks_mm_map_t map = NULL;

  NOOKS_ENTER();

  PDEBUG(" Area maps:\n");
  //  list_for_each(next, &nook->resources.area_map) {
  //    map = list_entry(next,nooks_mm_map_t, next);
  //    PDEBUG("\t area address 0x%lx, size 0x%lx \n",
  //	   map->address,
  //	   map->size );
  //  }

  PDEBUG(" Page maps:\n");
  list_for_each(next, &nook->resources.page_map) {
    map = list_entry(next,nooks_mm_map_t, next);
    printk("\t page address 0x%lx, size 0x%lx \n",
	   map->address,
	   map->size );
  }

  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_dump_nook_drivers
 *
 * print the list of drivers for a nook
 *
 *********************************************************/

void 
nooks_dump_nook_drivers(pnook_t nook)
{
  struct list_head * next;
  pnook_driver_t driver = NULL;

  NOOKS_ENTER();

  printk(" Drivers:\n");
  list_for_each(next, &nook->drivers) {
    driver = list_entry(next,nook_driver_t, next_in_nook);
    printk("\t driver: %s (%d)(%d)\n",
	   driver->name,
	   driver->major,
	   driver->driver_type);
  }

  NOOKS_EXIT();
}

