 /*
 * nooks-hash.c -- open hash table for nooks
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

#include "nooks-i.h"



int nooks_create_hash_table(pnooks_lookaside_list_t list, 
			    pnooks_hash_table_t table, 
			    int size,
			    int dual)
{
  int i;

  if (table == NULL)
	return (-ENOMEM);

  spin_lock_init(&table->lock);
  table->size = size;
  table->lookaside = list;


  table->mask = size - 1;
  table->size = size;
  table->shift = 0;
  table->dual = dual;
  
  while ((size >>= 1) != 0) {
    table->shift ++;
  }

  //
  // Double the space for dual tables
  //

  if (dual) {
    size = 2 * table->size;
  } else {
    size = table->size;
  }
 
  table->table = (struct list_head *) 
    kzalloc(size * sizeof(struct list_head), NOOKS_MEM_FLAGS);
  if (table->table == NULL) {
    return(-ENOMEM);
  }
  for (i = 0; i < size; i++) {
    INIT_LIST_HEAD(&table->table[i]);
  }
  
  return(0);
}

int nooks_destroy_hash_table(pnooks_hash_table_t table)
{
  int i;
  struct list_head *l;
  pnooks_hash_entry_t entry;

  printk ("***DESTROYING HASH TABLE %p ****.\n" );
  for (i = 0; i < table->size; i++) {
    l = &table->table[i];
    while (!list_empty(l)) {
      entry = list_entry(l->next,nooks_hash_entry_t, next);
      list_del(&entry->next);
      nooks_mm_lookaside_free(table->lookaside, entry);
    }
  }
  nooks_kfree(table->table);
  return 0;
}


int
nooks_add_to_hash(pnooks_hash_table_t table, 
		  nooks_resource_type_t entry_type,
		  void * key, 
		  void * data)
{
  struct list_head *l;
  pnooks_hash_entry_t entry;
  unsigned long flags;


  entry = nooks_mm_lookaside_alloc(table->lookaside);
  if (entry == NULL) {
    return(-ENOMEM);
  }
  //printk ("Add to hash:Table is %p and type is %d.\n", table, entry_type);

  entry->key = key;
  entry->data = data;
  entry->entry_type = entry_type;
  l = nooks_hash(table, key);
  //printk ("In add: List empty returns %d.\n", list_empty(l));
  spin_lock_irqsave(&table->lock, flags);
  list_add(&entry->next, l);
  //printk ("Added to the hash list %p and %d.\n", entry->key, entry_type);
  if (table->dual) {
    l = nooks_hash(table, data) + table->size;
    list_add(&entry->next_back, l);
  }
    
#ifdef NOOKS_KEEP_STATS
  table->hash_size++;
  table->hash_adds++;
#endif // NOOKS_KEEP_STATS
  spin_unlock_irqrestore(&table->lock, flags);
  return(0);  
}



int
nooks_del_from_hash(pnooks_hash_table_t table, 
		    nooks_resource_type_t entry_type,
		    void * key, 
		    void ** data)
{
  struct list_head *l;
  struct list_head * next;
  pnooks_hash_entry_t entry;

  unsigned long flags;
  if (table == NULL)
	return -ENOMEM;
  else 
	PDEBUG ("del from hash:Table in hash is %p for type %d.\n", table, entry_type);

 
  spin_lock_irqsave(&table->lock, flags);

  l = nooks_hash(table, key);
  list_for_each(next, l) {
    entry = list_entry(next, nooks_hash_entry_t, next);
    if ((entry->key == key) && (entry->entry_type == entry_type)) {
      if (data != NULL) {
	*data = entry->data;
      }
      list_del(&entry->next);
      if (table->dual) {
	list_del(&entry->next_back);
      }
#ifdef NOOKS_KEEP_STATS
      table->hash_hits++;
      table->hash_size--;
#endif // NOOKS_KEEP_STATS
      spin_unlock_irqrestore(&table->lock, flags);
      nooks_mm_lookaside_free(table->lookaside, entry);
      return(0);
    }
  }
#ifdef NOOKS_KEEP_STATS
  table->hash_misses++;
#endif // NOOKS_KEEP_STATS
  spin_unlock_irqrestore(&table->lock, flags);
  return(-ENOENT);
}

int 
nooks_find_in_hash(pnooks_hash_table_t table, 
		   nooks_resource_type_t entry_type,
		   void * key, 
		   void ** data)
{
  struct list_head *l;
  struct list_head * next;
  pnooks_hash_entry_t entry;

  unsigned long flags;

  //printk("Searching in hash in nooks_find_in_hash function with entrytype %d and table %p.\n", entry_type, table);
  
  if ((table == NULL)) {
	printk ("Table is NULL.\n");
	return -ENOMEM;
  }
  spin_lock_irqsave(&table->lock, flags); 
  l = nooks_hash(table, key); 
  if (l == NULL)  {
	printk ("List is NULL. \n");
	spin_unlock_irqrestore(&table->lock, flags);
	return -ENOMEM;
  }
  
  list_for_each(next, l) {
  entry = list_entry(next, nooks_hash_entry_t, next);
    if ((entry->key == key) && (entry->entry_type == entry_type)) {
      if (data != NULL) {
	*data = entry->data;
      }
#ifdef NOOKS_KEEP_STATS
      table->hash_hits++;
#endif // NOOKS_KEEP_STATS
      spin_unlock_irqrestore(&table->lock, flags);
      return(0);
    }
  }
#ifdef NOOKS_KEEP_STATS
  table->hash_misses++;
#endif // NOOKS_KEEP_STATS
  spin_unlock_irqrestore(&table->lock, flags);
  return(-ENOENT);
}

int
nooks_del_from_hash_reverse(pnooks_hash_table_t table, 
			    nooks_resource_type_t entry_type,
			    void * data, 
			    void ** key)
{
  struct list_head *l;
  struct list_head * next;
  pnooks_hash_entry_t entry;
  int i;

  unsigned long flags;
  spin_lock_irqsave(&table->lock, flags);

  //
  // If we have a dual table, look in the reverse path
  //
  if (table->dual) {
    l = nooks_hash(table, data) + table->size;
    list_for_each(next, l) {
      entry = list_entry(next, nooks_hash_entry_t, next_back);
      if ((entry->data == data) && (entry->entry_type == entry_type)) {
	if (key != NULL) {
	  *key = entry->key;
	}
	list_del(&entry->next);
	list_del(&entry->next_back);
#ifdef NOOKS_KEEP_STATS
	table->hash_hits++;
	table->hash_size--;
#endif // NOOKS_KEEP_STATS
	spin_unlock_irqrestore(&table->lock, flags);
	nooks_mm_lookaside_free(table->lookaside, entry);
	return(0);
      }
    }
  } else {
    //
    // Slowly search all entries
    //
    for (i = 0; i < table->size; i++) {
      l = &table->table[i];
      list_for_each(next, l) {
	entry = list_entry(next, nooks_hash_entry_t, next);
	if ((entry->data == data) && (entry->entry_type == entry_type)) {
	  if (key != NULL) {
	    *key = entry->key;
	  }
	  list_del(&entry->next);
#ifdef NOOKS_KEEP_STATS
	  table->hash_hits++;
	  table->hash_size--;
#endif // NOOKS_KEEP_STATS
	  spin_unlock_irqrestore(&table->lock, flags);
	  nooks_mm_lookaside_free(table->lookaside, entry);
	  return(0);
	}
      }
    }
  }
#ifdef NOOKS_KEEP_STATS
  table->hash_misses++;
#endif // NOOKS_KEEP_STATS
  spin_unlock_irqrestore(&table->lock, flags);
  return(-ENOENT);
}

int
nooks_find_in_hash_reverse(pnooks_hash_table_t table, 
			   nooks_resource_type_t entry_type,
			   void * data, 
			   void ** key)
{
  struct list_head *l;
  struct list_head * next;
  pnooks_hash_entry_t entry;
  int i;

  unsigned long flags;
  spin_lock_irqsave(&table->lock, flags);

  //
  // If we have a dual table, look in the reverse half
  //
  if (table->dual) {
    l = nooks_hash(table, data) + table->size;
    list_for_each(next, l) {
      entry = list_entry(next, nooks_hash_entry_t, next_back);
      if ((entry->data == data) && (entry->entry_type == entry_type)) {
	if (key != NULL) {
	  *key = entry->key;
	}
#ifdef NOOKS_KEEP_STATS
	table->hash_hits++;
#endif // NOOKS_KEEP_STATS
	spin_unlock_irqrestore(&table->lock, flags);
	return(0);
      }
    }
  } else {
    //
    // Look in the forward path by checking all entries
    //
    for (i = 0; i < table->size; i++) {
      l = &table->table[i];
      list_for_each(next, l) {
	entry = list_entry(next, nooks_hash_entry_t, next);
	if ((entry->data == data) && (entry->entry_type == entry_type)) {
	  if (key != NULL) {
	    *key = entry->key;
	  }
#ifdef NOOKS_KEEP_STATS
	  table->hash_hits++;
#endif // NOOKS_KEEP_STATS
	  spin_unlock_irqrestore(&table->lock, flags);
	  return(0);
	}
      }
    }
  }
#ifdef NOOKS_KEEP_STATS
  table->hash_misses++;
#endif // NOOKS_KEEP_STATS
  spin_unlock_irqrestore(&table->lock, flags);
  return(-ENOENT);
}





int 
nooks_replace_in_hash(pnooks_hash_table_t table, 
		      nooks_resource_type_t entry_type,
		      void * key, 
		      void * data,
		      void ** old_data)
{
  struct list_head *l;
  struct list_head * next;
  pnooks_hash_entry_t entry;
  pnooks_hash_entry_t old_entry = NULL;
  unsigned long flags;

  entry = nooks_mm_lookaside_alloc(table->lookaside);
  if (entry == NULL) {
    return(-ENOMEM);
  }
  entry->key = key;
  entry->data = data;
  entry->entry_type = entry_type;

  l = nooks_hash(table, key);
  spin_lock_irqsave(&table->lock, flags);
    
#ifdef NOOKS_KEEP_STATS
  table->hash_size++;
  table->hash_adds++;
#endif // NOOKS_KEEP_STATS
  spin_unlock_irqrestore(&table->lock, flags);
  return(0);  

  spin_lock_irqsave(&table->lock, flags);

  l = nooks_hash(table, key);
  list_for_each(next, l) {
    old_entry = list_entry(next, nooks_hash_entry_t, next);
    if ((old_entry->key == key) && (old_entry->entry_type == entry_type)) {

      //
      // Delete the old entry, but remember the data
      //

      list_del(&old_entry->next);
      list_del(&old_entry->next_back);
      break;
    }
    old_entry = NULL;
  }
  list_add(&entry->next, l);
  spin_unlock_irqrestore(&table->lock, flags);
  if (old_entry != NULL) {
    if (old_data != NULL) {
      *old_data = old_entry->data;
    }
    nooks_mm_lookaside_free(table->lookaside, old_entry);
  }
  return(0);
}

int
nooks_enumerate_in_hash(pnooks_hash_table_t table,
			pnooks_hash_enum_t context,
			pnooks_hash_entry_t * entry)
{
  int index = 0;
  struct list_head * list;
  unsigned long flags;
  spin_lock_irqsave(&table->lock, flags);


  //
  // Find the entry to return
  //
  index = context->index;
  if (context->next == NULL) {
    for (list = &table->table[index]; 
	 ((index < table->size) && (list_empty(list))); 
	  index++, list = &table->table[index])
      ;

    if (index == table->size) {
      spin_unlock_irqrestore(&table->lock, flags);
      return(-ENOENT);
    }
    
    *entry = list_entry( list->next, nooks_hash_entry_t, next);
  } else {
    *entry = list_entry(context->next, nooks_hash_entry_t, next);
    list = &table->table[context->index];
  }
  
  if ((*entry)->next.next != list) {
    context->index = index;
    context->next = (*entry)->next.next;
    if ((context->next == NULL)  || (context->next->next == NULL) ||
	(context->next->prev == NULL)) {
      BUG();
    }
  } else {
    context->index = index + 1;
    context->next = NULL;
  }
  
  spin_unlock_irqrestore(&table->lock, flags);
  return(0);
}

void 
nooks_delete_entry_from_hash(pnooks_hash_table_t table,			     
			     pnooks_hash_entry_t entry)
{
  unsigned long flags;
  spin_lock_irqsave(&table->lock, flags);

  list_del(&entry->next);
  if (table->dual) {
    list_del(&entry->next_back);
  }
  nooks_mm_lookaside_free(table->lookaside, entry);
  spin_unlock_irqrestore(&table->lock, flags);
  
}


void
nooks_dump_hash(pnooks_hash_table_t table)
{
  nooks_hash_enum_t context = {0};
  pnooks_hash_entry_t entry = NULL;
  int i = 0;

  //
  // Traverse the active skbs and free them now
  //

  while (nooks_enumerate_in_hash(table,
				 &context,
				 &entry) == 0) {
    printk("%d: type %d value %p data %p\n",
	   i, entry->entry_type, entry->key, entry->data);
    i++;
  }
    
}

void
nooks_dump_hash_histogram(pnooks_hash_table_t table)
{
  int histogram[100];
  nooks_hash_enum_t context = {0};
  pnooks_hash_entry_t entry = NULL;
  int i = 0;

  memset(histogram, 0, sizeof(histogram));
  //
  // Traverse the active skbs and free them now
  //

  while (nooks_enumerate_in_hash(table,
				 &context,
				 &entry) == 0) {
    if (entry->entry_type < 100) {
      histogram[entry->entry_type]++;
    } else {
      histogram[0]++;
    }
  }
  for (i = 0; i < 100; i++) {
    printk("type %d: %d\n",i,histogram[i]);
  }
    
}

int
nooks_hash_empty(pnooks_hash_table_t table)
{
  struct list_head *l;
  int i;
  unsigned long flags;
  int empty = 1;

  spin_lock_irqsave(&table->lock, flags);

  for (i = 0; i < table->size; i++) {
    l = &table->table[i];
    if (!list_empty(l)) {
      empty = 0;
      break;
    }
  }
  spin_unlock_irqrestore(&table->lock, flags);
  return(empty);

}
