/*
 * nooks-iwrappers.c -- internal wrappers for driver functions
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  No warranty 
 * is attached; * we cannot take responsibility for errors or 
 * fitness for use.
 *
 * These functions are called directly by the kernel, and don't specify
 * which nook/driver to use. These set the current nook and then call
 * the dwrapper functions, which will change protection domains and 
 * copy parameters to the driver.
 */
#include "nooks-i.h"
#include "nooks-iwrappers.h"
#include <linux/module.h>


/**********************************************************
 *
 * Instance wrapper global variable definitions
 *
 *********************************************************/

spinlock_t nooks_iwrapper_lock;
nooks_iwrapper_itf * nooks_iwrapper_itf_array[NOOKS_IWRAPPER_TYPE_COUNT];
pnook_driver_t * nooks_iwrapper_driver_array[NOOKS_IWRAPPER_TYPE_COUNT];
nook_iwrapper_t nooks_iwrapper_array[NOOKS_IWRAPPER_TYPE_COUNT][MAX_NOOK_DRIVERS];

//#include "nooks-iwrap-c-gen.h"

unsigned long nooks_iwrapper_flags[NOOKS_IWRAPPER_TYPE_COUNT];
char * nooks_iwrapper_names[NOOKS_IWRAPPER_TYPE_COUNT];


/**********************************************************
 *
 *  nooks_iwrapper_debug
 *
 * point for breaking at all iwrapper calls
 *
 *********************************************************/

void nooks_iwrapper_debug(void)
{
  static int calls = 0;

  calls++;
}


/**********************************************************
 *
 * nooks_init_iwrappers
 *
 * Initializes the instance wrappers for calling specific 
 * instances of driver entry interfaces.
 *
 *********************************************************/
int
nooks_init_iwrappers(void)
{
  NOOKS_ENTER();

  // 
  // Initialize the tables that point to the drivers using each
  // instance of an interface
  //

  if (NOOKS_IWRAPPER_TYPE_COUNT != (int) last_ops_e) {
    PDEBUG("\n\n\n\n ****** Nooks Iwrapper mismatch: %d vs %d\n\n\n\n",
	   NOOKS_IWRAPPER_TYPE_COUNT,
	   last_ops_e);
  }
  spin_lock_init(&nooks_iwrapper_lock);

  //
  // Initialize the arrays of pointers to interfaces
  //

#define NOOKS_INIT_OPS_ARRAY(_type_, _flags_, _name_) \
  nooks_iwrapper_flags[_type_ ## _e] = (_flags_); \
  nooks_iwrapper_names[_type_ ## _e] = (_name_);


  NOOKS_INIT_OPS_ARRAY(mops, NOOKS_IWRAPPER_REGISTERED_ITF | 
		       NOOKS_IWRAPPER_NAMED, "module");
  NOOKS_INIT_OPS_ARRAY(fops, NOOKS_IWRAPPER_REGISTERED_ITF |
		       NOOKS_IWRAPPER_NAMED | NOOKS_IWRAPPER_NUMBERED, 
		       "file");
  NOOKS_INIT_OPS_ARRAY(procops, 0, "proc-fs");
  NOOKS_INIT_OPS_ARRAY(netops, NOOKS_IWRAPPER_REGISTERED_ITF |
		       NOOKS_IWRAPPER_NAMED, "network");
  NOOKS_INIT_OPS_ARRAY(pciops, 0, "pci-device");
  NOOKS_INIT_OPS_ARRAY(callbackops, 0, "callback");
  NOOKS_INIT_OPS_ARRAY(timerops, 0, "timer");
  NOOKS_INIT_OPS_ARRAY(notifierops, 0, "notifier");

  NOOKS_EXIT();
  return(0);  
}

/**********************************************************
 * nooks_iwrapper_generic
 *
 * generic iwrapper function. It's gets a pointer to an
 * nooks_iwrapper structure and call the good wrappers
 * - orignal by mikesw - modified by dmg
 **********************************************************/
NOOKS_ICALL int nooks_iwrapper_generic(void)
{
  int result = 0;
  NOOKS_IWRAPPER_ENTER;
#ifdef CONFIG_NOOKS_SHADOW
  //printk ("No. of arguments is %d for driver %s.\n", _iwrap->nb_args, _iwrap->driver->name);
  //printk ("No. of arguments is %d for nook %s.\n", _iwrap->nb_args, _iwrap->driver->nook->name);
  if (__nook->shadow != NULL) {
    if ((__nook->state == nook_state_running) ||
        (__driver->shadow_state == shadow_state_running) ||
        IN_SHADOW_RECOVERY())
      {  //printk ("nooks_iwrapper_generic sees normal shadow conditions.\n");
	 NOOKS_IWRAPPER_CALL_RES(_iwrap->dwrapper, _iwrap->nb_args, result); }
                                                                                
    if (_iwrap->is_void)
      { NOOKS_IWRAPPER_CALL_DRIV(_iwrap->wrapped, _iwrap->nb_args, _iwrap->driver); }
    else
      { 
	//printk ("No. of arguments is %d for driver %s.\n", _iwrap->nb_args, _iwrap->driver->name);
	NOOKS_IWRAPPER_CALL_RES_DRIV(_iwrap->wrapped, _iwrap->nb_args, result, _iwrap->driver); }
  
  } else
#endif // CONFIG_NOOKS_SHADOW
  {
    NOOKS_IWRAPPER_CHECK( _iwrap->errnum );
    NOOKS_IWRAPPER_CALL_RES(_iwrap->dwrapper, _iwrap->nb_args, result);
  }
  if (__driver->need_delete) {
    nooks_free_driver(__driver);
  }
  NOOKS_IWRAPPER_EXIT();
  return(result);
}

/**********************************************************
 * nooks_iwrapper_set
 *
 * set an iwrapper structure
 * - dmg
 **********************************************************/
inline void nooks_iwrapper_set( pnook_iwrapper_t iwrapper,
				pnook_driver_t driver,
				int nb_args,
				int is_void,
				void *dwrapper,
				int errnum )
{
  void * iwrapper_gen = &nooks_iwrapper_generic;
// movl iwrapper, %ecx = b9 iwrapper
#define MOVL_ECX  0xb9
// movl &nooks_iwrapper_generic, %eax = b8 &nooks_iwrapper_generic
#define MOVL_EAX  0xb8
// jmp *%eax = ff e0
#define JMP_EAX   "\xff\xe0"

  iwrapper->nb_args  = nb_args;
  iwrapper->is_void  = is_void;
  iwrapper->dwrapper = dwrapper;
  iwrapper->wrapped  = NULL;
  iwrapper->driver   = driver;
  iwrapper->errnum   = errnum;
  // hard part : generate automatiquely the specific code for this
  // iwrapper
  iwrapper->iwrapper[0] = MOVL_ECX;
  memcpy(iwrapper->iwrapper+1, &iwrapper, 4);
  iwrapper->iwrapper[5] = MOVL_EAX;
  memcpy(iwrapper->iwrapper+6, &iwrapper_gen, 4);
  memcpy(iwrapper->iwrapper+10, JMP_EAX, 2);
  // end of hard part
}

/**********************************************************
 *
 * nooks_iwrapper_get_itf
 *
 * Request an instance of a particular wrapper
 * dmg - it's just a wrapper to the correct get_itf
 *       function now 
 *
 *********************************************************/
// automatically generate the allocation functions for each type
#include "nooks-iwrapper-gen_c.h"
// just a wrapper for the iwrapper allocation function
nooks_iwrapper_itf
    nooks_iwrapper_get_itf(nooks_iwrapper_type_t type,
                           pnook_driver_t driver)
{
  switch(type)
  {
    NOOKS_IWRAPPER_GET_ITF(driver);
  }
  return NULL;
}

/**********************************************************
 *
 * nooks_iwrapper_release_itf
 *
 * Release the interface back to the pool of interfaces for
 * the specified type.
 *
 * itf_type no longer used, keep for backward compatibility
 *    - dmg
 *********************************************************/
void 
nooks_iwrapper_release_itf(nooks_iwrapper_type_t itf_type,
			   nooks_iwrapper_itf itf)
{
  NOOKS_ENTER();
  nooks_kfree(itf);
  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_iwrapper_copy_mops
 *
 * Takes the functions pointers from the input module and 
 * copies them to the wrapper operations and updates the
 * output module with the correct function pointers
 *
 * wrapper operations: the table of functions that will be
 *      passed to kernel functions
 * mops_op: the internal wrapper functions that will be passed
 *      in that table
 * module_in: the module data passed in, with nook-side functions
 * module_out: the module data to be passed to the kernel, with
 *      instance wrapper functions
 *
 *********************************************************/

void
nooks_iwrapper_copy_mops(struct module * module_in,
			 struct module_operations * wrapper_operations, 
			 struct iwrapper_module_operations * iwrapper_operations,
			 struct module * module_out)
{
  
#define IWRAPPER_COPY_MOPS(_min_, _w_, _i_, _mout_, _f_) \
  do { \
  if ((_min_)->_f_ != NULL) { \
    (_w_)->_f_ = (_min_)->_f_; \
    (_mout_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_mout_)->_f_ = NULL; \
  }  } while (0)  
  
  NOOKS_ENTER();
  IWRAPPER_COPY_MOPS(module_in, wrapper_operations, iwrapper_operations, 
		     module_out, init);
  IWRAPPER_COPY_MOPS(module_in, wrapper_operations, iwrapper_operations, 
		     module_out, exit);
/*   IWRAPPER_COPY_MOPS(module_in, wrapper_operations, iwrapper_operations, 
		     module_out, unsafe);
 */ 

  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_iwrapper_copy_fops
 *
 * Copy the file operations from a driver into an operations
 * structure, preserving null pointers
 *
 * driver_operations: the table of operations suplied by a
 *    driver.
 * wrapper_operations: the table of functions in the nook 
 *    that stores a copy of the driver's function pointers
 * iwrapper_operations: the table of instance wrapper operations
 *    for this driver
 * system_operations: the table of operations that is handed to
 *    the operating system
 *
 *********************************************************/

void
nooks_iwrapper_copy_fops(struct file_operations * driver_operations, 
			 struct file_operations * wrapper_operations,
			 struct iwrapper_file_operations * iwrapper_operations,
			 struct file_operations * system_operations)
{

#define IWRAPPER_COPY_FOPS(_d_, _w_,  _i_, _s_, _f_) \
  do { \
  if ((_d_)->_f_ != NULL) { \
    (_w_)->_f_ = (_d_)->_f_; \
    (_s_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_s_)->_f_ = NULL; \
  }  } while (0)  

  NOOKS_ENTER();


//  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
//		     iwrapper_operations, system_operations, owner);
  system_operations->owner = driver_operations->owner;
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, llseek);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, read);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, write);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, readdir);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, poll);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, ioctl);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, mmap);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, open);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, flush);

  //
  // We always want a release function, even if the driver doesn't
  //

  wrapper_operations->release =driver_operations->release;
  system_operations->release = (void*)(iwrapper_operations->release.iwrapper);

  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, fsync);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, fasync);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, lock);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, readv);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, writev);
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, sendpage);  
  IWRAPPER_COPY_FOPS(driver_operations, wrapper_operations, 
		     iwrapper_operations, system_operations, get_unmapped_area);
  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_iwrapper_copy_netops
 *
 * Takes the functions pointers from the input module and 
 * copies them to the wrapper operations and updates the
 * output module with the correct function pointers
 *
 * wrapper operations: the table of functions that will be
 *      passed to kernel functions
 * mops_op: the internal wrapper functions that will be passed
 *      in that table
 * module_in: the module data passed in, with nook-side functions
 * module_out: the module data to be passed to the kernel, with
 *      instance wrapper functions
 *
 * Notes: allow kernel functions to be used directly. 
 * BUGBUG: We should
 * check types (the function is in the right slot)
 *
 *********************************************************/

#include <linux/etherdevice.h>
#include <linux/fddidevice.h>
#include <linux/hippidevice.h>
#include <linux/trdevice.h>
#include <linux/fcdevice.h>

int eth_header_parse(struct sk_buff *skb, unsigned char *haddr);
int eth_mac_addr(struct net_device *dev, void *p);
int eth_change_mtu(struct net_device *dev, int new_mtu);
int fddi_change_mtu(struct net_device *dev, int new_mtu);
int hippi_change_mtu(struct net_device *dev, int new_mtu);
int hippi_mac_addr(struct net_device *dev, void *p);
int hippi_neigh_setup_dev(struct net_device *dev, struct neigh_parms *p);
int ltalk_change_mtu(struct net_device *dev, int mtu);
int ltalk_mac_addr(struct net_device *dev, void *addr);

void * nooks_legal_net_function_ptrs[] = {
  eth_change_mtu,
  eth_header,
  eth_rebuild_header,
  eth_mac_addr,
  eth_header_cache,  
  eth_header_cache_update,
  eth_header_parse,  
#ifdef CONFIG_FIDDI
  fddi_change_mtu,
  fddi_header,
  fddi_rebuild_header,
#endif
#ifdef CONFIG_HIPPI
  hippi_change_mtu,
//  hippi_header,
//  hippi_rebuild_header,
  hippi_mac_addr,
  hippi_neigh_setup_dev, 
#endif
#ifdef CONFIG_ATALK
  ltalk_change_mtu,
  ltalk_mac_addr,
#endif
#ifdef CONFIG_TR
//  tr_header,
//  tr_rebuild_header,
#endif
#ifdef CONFIG_NET_FC
//  fc_header,
//  fc_rebuild_header,
#endif
  NULL };


void
nooks_iwrapper_copy_netops(struct net_device * netdev_in,
			   struct netdev_operations * wrapper_operations, 
			   struct iwrapper_netdev_operations * iwrapper_operations,
			   struct net_device * netdev_out)
{
  
#define IWRAPPER_COPY_NETOPS(_nin_, _w_, _i_, _nout_, _f_) \
  do { int _x_; \
  for (_x_ = 0; nooks_legal_net_function_ptrs[_x_] != NULL; _x_++ ) { \
    if (nooks_legal_net_function_ptrs[_x_] == (void *) (_nin_)->_f_) break; \
  } \
  if (nooks_legal_net_function_ptrs[_x_] != NULL) {\
    (_w_)->_f_ = (_nin_)->_f_; \
    (_nout_)->_f_ = (_nin_)->_f_; \
    break; \
  } \
  if ((_nin_)->_f_ != NULL) { \
    (_w_)->_f_ = (_nin_)->_f_; \
    (_nout_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_nout_)->_f_ = NULL; \
  }  } while (0)  
  
  NOOKS_ENTER();
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, init);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, get_stats);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, get_wireless_stats);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, uninit);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, destructor);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, open);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, stop);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, hard_start_xmit);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, hard_header);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, rebuild_header);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, set_multicast_list);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, set_mac_address);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, do_ioctl);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, set_config);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, hard_header_cache);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, header_cache_update);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, change_mtu);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, tx_timeout);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, hard_header_parse);
  IWRAPPER_COPY_NETOPS(netdev_in, wrapper_operations, iwrapper_operations, 
		     netdev_out, neigh_setup);
  /* FIXME accept_fastpath no longer defined. Are there any new functions -- not above FIXME
   * Need to vlan* functions here -- See netdevice.h.
   */ 
  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_iwrapper_copy_irqops
 *
 * Copy the irq operations from a driver into an operations
 * structure, preserving null pointers
 *
 * driver_operations: the table of operations suplied by a
 *    driver.
 * wrapper_operations: the table of functions in the nook 
 *    that stores a copy of the driver's function pointers
 * iwrapper_operations: the table of instance wrapper operations
 *    for this driver
 * system_operations: the table of operations that is handed to
 *    the operating system
 *
 *********************************************************/

void
nooks_iwrapper_copy_irqops(void (*handler)(int, void *, struct pt_regs *),
			   struct irq_operations * wrapper_operations,
			   struct iwrapper_irq_operations * iwrapper_operations,
			   struct irq_operations * system_operations)
{

#define IWRAPPER_COPY_IRQOPS(_d_, _w_,  _i_, _s_, _f_) \
  do { \
  if ((_d_) != NULL) { \
    (_w_)->_f_ = (_d_); \
    (_s_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_s_)->_f_ = NULL; \
  }  } while (0)  

  NOOKS_ENTER();

  IWRAPPER_COPY_IRQOPS(handler, wrapper_operations, 
		     iwrapper_operations, system_operations, handler);
  

  NOOKS_EXIT();
}


/**********************************************************
 *
 * nooks_iwrapper_copy_callbackops
 *
 * Copy the timer operations from a driver into an operations
 * structure, preserving null pointers
 *
 * driver_operations: the table of operations suplied by a
 *    driver.
 * wrapper_operations: the table of functions in the nook 
 *    that stores a copy of the driver's function pointers
 * iwrapper_operations: the table of instance wrapper operations
 *    for this driver
 * system_operations: the table of operations that is handed to
 *    the operating system
 *
 *********************************************************/

void
nooks_iwrapper_copy_callbackops(callback_handler_t * handler,
				struct callback_operations * wrapper_operations,
				struct iwrapper_callback_operations * iwrapper_operations,
				struct callback_operations * system_operations)
{

#define IWRAPPER_COPY_CALLBACKOPS(_d_, _w_,  _i_, _s_, _f_) \
  do { \
  if ((_d_) != NULL) { \
    (_w_)->_f_ = (_d_); \
    (_s_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_s_)->_f_ = NULL; \
  }  } while (0)  

  NOOKS_ENTER();

  IWRAPPER_COPY_CALLBACKOPS(handler, wrapper_operations, 
			    iwrapper_operations, system_operations, handler);
  

  NOOKS_EXIT();
}

void
nooks_iwrapper_copy_pm_callbackops(pm_callback pm_handler,
				struct callback_operations * wrapper_operations,
				struct iwrapper_callback_operations * iwrapper_operations,
				struct callback_operations * system_operations)
{

#define IWRAPPER_COPY_PM_CALLBACKOPS(_d_, _w_,  _i_, _s_, _f_) \
  do { \
  if ((_d_) != NULL) { \
    (_w_)->_f_ = (_d_); \
    (_s_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_s_)->_f_ = NULL; \
  }  } while (0)  

  NOOKS_ENTER();

  IWRAPPER_COPY_PM_CALLBACKOPS(pm_handler, wrapper_operations, 
			    iwrapper_operations, system_operations, pm_handler);
  

  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_iwrapper_copy_timerops
 *
 * Copy the timer operations from a driver into an operations
 * structure, preserving null pointers
 *
 * driver_operations: the table of operations suplied by a
 *    driver.
 * wrapper_operations: the table of functions in the nook 
 *    that stores a copy of the driver's function pointers
 * iwrapper_operations: the table of instance wrapper operations
 *    for this driver
 * system_operations: the table of operations that is handed to
 *    the operating system
 *
 *********************************************************/

void
nooks_iwrapper_copy_timerops(void (*handler)(unsigned long),
				struct timer_operations * wrapper_operations,
				struct iwrapper_timer_operations * iwrapper_operations,
				struct timer_operations * system_operations)
{

#define IWRAPPER_COPY_TIMEROPS(_d_, _w_,  _i_, _s_, _f_) \
  do { \
  if ((_d_) != NULL) { \
    (_w_)->_f_ = (_d_); \
    (_s_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_s_)->_f_ = NULL; \
  }  } while (0)  

  NOOKS_ENTER();

  IWRAPPER_COPY_TIMEROPS(handler, wrapper_operations, 
			 iwrapper_operations, system_operations, handler);
  

  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_iwrapper_copy_pciops
 *
 * Takes the functions pointers from the input module and 
 * copies them to the wrapper operations and updates the
 * output module with the correct function pointers
 *
 * wrapper operations: the table of functions that will be
 *      passed to kernel functions
 * mops_op: the internal wrapper functions that will be passed
 *      in that table
 * module_in: the module data passed in, with nook-side functions
 * module_out: the module data to be passed to the kernel, with
 *      instance wrapper functions
 *
 *********************************************************/

void
nooks_iwrapper_copy_pciops(struct pci_driver * pcidrv_in,
			   struct pci_operations * wrapper_operations, 
			   struct iwrapper_pci_operations * iwrapper_operations,
			   struct pci_driver* pcidrv_out)
{
  
#define IWRAPPER_COPY_PCIOPS(_pin_, _w_, _i_, _pout_, _f_) \
  do { \
  if ((_pin_)->_f_ != NULL) { \
    (_w_)->_f_ = (_pin_)->_f_; \
    (_pout_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_pout_)->_f_ = NULL; \
  }  } while (0)  
  
  NOOKS_ENTER();
  IWRAPPER_COPY_PCIOPS(pcidrv_in, wrapper_operations, iwrapper_operations, 
		     pcidrv_out, probe);
  IWRAPPER_COPY_PCIOPS(pcidrv_in, wrapper_operations, iwrapper_operations, 
		     pcidrv_out, remove);
  IWRAPPER_COPY_PCIOPS(pcidrv_in, wrapper_operations, iwrapper_operations, 
		     pcidrv_out, suspend);
  IWRAPPER_COPY_PCIOPS(pcidrv_in, wrapper_operations, iwrapper_operations, 
		     pcidrv_out, resume);
  IWRAPPER_COPY_PCIOPS(pcidrv_in, wrapper_operations, iwrapper_operations, 
		     pcidrv_out, enable_wake);
  IWRAPPER_COPY_PCIOPS(pcidrv_in, wrapper_operations, iwrapper_operations,
                     pcidrv_out, shutdown);
  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_iwrapper_copy_pciops
 *
 * Takes the functions pointers from the input module and 
 * copies them to the wrapper operations and updates the
 * output module with the correct function pointers
 *
 * wrapper operations: the table of functions that will be
 *      passed to kernel functions
 * mops_op: the internal wrapper functions that will be passed
 *      in that table
 * module_in: the module data passed in, with nook-side functions
 * module_out: the module data to be passed to the kernel, with
 *      instance wrapper functions
 *
 *********************************************************/

void
nooks_iwrapper_copy_isapnp_pciops(isapnp_probe_fn probe,
				  struct pci_operations * wrapper_operations, 
				  struct iwrapper_pci_operations * iwrapper_operations,
				  struct pci_operations * system_operations)
{
  
#define IWRAPPER_COPY_ISAPNP_PCIOPS(_pin_, _w_, _i_, _pout_, _f_) \
  do { \
  if ((_pin_) != NULL) { \
    (_w_)->_f_ = (_pin_); \
    (_pout_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_pout_)->_f_ = NULL; \
  }  } while (0)  
  
  NOOKS_ENTER();
  IWRAPPER_COPY_ISAPNP_PCIOPS((pci_probe_fn) probe, 
			      wrapper_operations, 
			      iwrapper_operations, 
			      system_operations, 
			      probe);

  NOOKS_EXIT();
}

/**********************************************************
 *
 * nooks_iwrapper_copy_procops
 *
 * Takes the functions pointers from the input module and 
 * copies them to the wrapper operations and updates the
 * output module with the correct function pointers
 *
 * wrapper operations: the table of functions that will be
 *      passed to kernel functions
 * mops_op: the internal wrapper functions that will be passed
 *      in that table
 * module_in: the module data passed in, with nook-side functions
 * module_out: the module data to be passed to the kernel, with
 *      instance wrapper functions
 *
 *********************************************************/

void
nooks_iwrapper_copy_procops(struct proc_dir_entry * pentry_in,
			   struct procfs_operations * wrapper_operations, 
			   struct iwrapper_procfs_operations * iwrapper_operations,
			   struct proc_dir_entry * pentry_out)
{
  
#define IWRAPPER_COPY_PROCOPS(_pin_, _w_, _i_, _pout_, _f_) \
  do { \
  if ((_pin_)->_f_ != NULL) { \
    (_w_)->_f_ = (_pin_)->_f_; \
    (_pout_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_pout_)->_f_ = NULL; \
  }  } while (0)  
  
  NOOKS_ENTER();
  IWRAPPER_COPY_PROCOPS(pentry_in, wrapper_operations, iwrapper_operations, 
		     pentry_out, get_info);
  IWRAPPER_COPY_PROCOPS(pentry_in, wrapper_operations, iwrapper_operations, 
		     pentry_out, read_proc);
  IWRAPPER_COPY_PROCOPS(pentry_in, wrapper_operations, iwrapper_operations, 
		     pentry_out, write_proc);

  NOOKS_EXIT();
}

void * nooks_legal_sysctl_proc_handler_function_ptrs[] = {
  proc_dostring,
  proc_dointvec,
  proc_dointvec_bset,
  proc_dointvec_minmax,
  proc_dointvec_jiffies,
  proc_doulongvec_minmax,
  proc_doulongvec_ms_jiffies_minmax,
  NULL };

void * nooks_legal_sysctl_strategy_function_ptrs[] = {
  sysctl_string,
  sysctl_intvec,
  sysctl_jiffies,
  NULL };

/**********************************************************
 *
 * nooks_iwrapper_copy_sysctlops
 *
 * Copy the sysctl operations from a driver into an operations
 * structure, preserving null pointers
 *
 * driver_operations: the table of operations suplied by a
 *    driver.
 * wrapper_operations: the table of functions in the nook 
 *    that stores a copy of the driver's function pointers
 * iwrapper_operations: the table of instance wrapper operations
 *    for this driver
 * system_operations: the table of operations that is handed to
 *    the operating system
 *
 *********************************************************/

void
nooks_iwrapper_copy_sysctlops(ctl_table * ctl_in,
			   struct sysctl_operations * wrapper_operations, 
			   struct iwrapper_sysctl_operations * iwrapper_operations,
			   ctl_table * ctl_out)
{
#define IWRAPPER_COPY_SYSCTLOPS(_cin_, _w_, _i_, _cout_, _f_) \
  do { \
  if ((_cin_)->_f_ != NULL) { \
    (_w_)->_f_ = (_cin_)->_f_; \
    (_cout_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_cout_)->_f_ = NULL; \
  }  } while (0)  
  
  NOOKS_ENTER();

  IWRAPPER_COPY_SYSCTLOPS(ctl_in, wrapper_operations, iwrapper_operations, 
		     ctl_out, proc_handler);
  IWRAPPER_COPY_SYSCTLOPS(ctl_in, wrapper_operations, iwrapper_operations, 
		     ctl_out, strategy);
  NOOKS_EXIT();
}


/**********************************************************
 *
 * nooks_iwrapper_copy_actorops
 *
 * Copy the actor operations from a driver into an operations
 * structure, preserving null pointers
 *
 * driver_operations: the table of operations suplied by a
 *    driver.
 * wrapper_operations: the table of functions in the nook 
 *    that stores a copy of the driver's function pointers
 * iwrapper_operations: the table of instance wrapper operations
 *    for this driver
 * system_operations: the table of operations that is handed to
 *    the operating system
 *
 *********************************************************/


void
nooks_iwrapper_copy_actorops(read_actor_t actor,
			     struct actor_operations * wrapper_operations,
			     struct iwrapper_actor_operations * iwrapper_operations,
			     struct actor_operations * system_operations)
{

#define IWRAPPER_COPY_ACTOROPS(_d_, _w_,  _i_, _s_, _f_) \
  do { \
  if ((_d_) != NULL) { \
    (_w_)->_f_ = (_d_); \
    (_s_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_s_)->_f_ = NULL; \
  }  } while (0)  

  NOOKS_ENTER();

  IWRAPPER_COPY_ACTOROPS(actor, wrapper_operations, 
			    iwrapper_operations, system_operations, read);
  

  NOOKS_EXIT();
}


/**********************************************************
 *
 * nooks_iwrapper_copy_notifierops
 *
 * Takes the functions pointers from the input module and 
 * copies them to the wrapper operations and updates the
 * output module with the correct function pointers
 *
 * wrapper operations: the table of functions that will be
 *      passed to kernel functions
 * mops_op: the internal wrapper functions that will be passed
 *      in that table
 * module_in: the module data passed in, with nook-side functions
 * module_out: the module data to be passed to the kernel, with
 *      instance wrapper functions
 *
 * Notes: allow kernel functions to be used directly. 
 * BUGBUG: We should
 * check types (the function is in the right slot)
 *
 *********************************************************/


void
nooks_iwrapper_copy_notifierops(struct notifier_block * notifier_in,
				struct notifier_operations * wrapper_operations, 
				struct iwrapper_notifier_operations * iwrapper_operations,
				struct notifier_block * notifier_out)
{
  
#define IWRAPPER_COPY_NOTIFIEROPS(_nin_, _w_, _i_, _nout_, _f_) \
  do {  \
  if ((_nin_)->_f_ != NULL) { \
    (_w_)->_f_ = (_nin_)->_f_; \
    (_nout_)->_f_ = (void*)((_i_)->_f_.iwrapper); \
  } else { \
    (_w_)->_f_ = NULL; \
    (_nout_)->_f_ = NULL; \
  }  } while (0)  
  
  NOOKS_ENTER();
  IWRAPPER_COPY_NOTIFIEROPS(notifier_in, 
			    wrapper_operations, 
			    iwrapper_operations, 
			    notifier_out, 
			    notifier_call);
  NOOKS_EXIT();
}

//#include "nooks-iwrappers-muthu.c"
