/*
 * nooks-mm.c -- memory management for nooks
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

 
#include "nooks-i.h"
#include "nooks-mm.h"
#include <linux/module.h>
#include "nooks-switch.h"
#include "nooks-sys.h"


/**********************************************************
 *
 * Function: nooks_dump_page_table 
 *
 * Prints out the non-null contents of a page table
 *
 *********************************************************/

void
nooks_dump_pte_set (pte_t * pte, unsigned long address)
{
  int i;

  PDEBUG("dumping pte set %p\n", pte);
  for (i = 0; i < PTRS_PER_PTE; i++) {
    if (!pte_none(*pte)) {

      PDEBUG("Pte for addr %lx at %p, value %lx\n",address, pte, pte->pte_low);


    } 
    address += PAGE_SIZE;
    pte++;
  }
}

void
nooks_dump_pmd_set(pmd_t * pmd, unsigned long address)
{
  int i;
  pte_t *pte;

  PDEBUG("dumping pmd set for pmd %lx\n", address);

  for (i = 0; i < PTRS_PER_PMD; i++ ) {
    //
    // If the kernel page table has an entry in this pmd, then copy it
    //
    if (!pmd_none(*pmd)) {

      pte = pte_offset_map(pmd, address);
      nooks_dump_pte_set(pte, address);
    } 
    address += PMD_SIZE;
    pmd++;
  }
}



void 
nooks_dump_mm(struct mm_struct * mm)
{
  int i;
  unsigned long pgd_address;
  pgd_t * pgd;
  pud_t * pud;
  //
  // Bail early if there is no page table to dump
  //

  if (mm->pgd == NULL) {
    return;
  }

  spin_lock(&mm->page_table_lock);

  pgd_address = PAGE_OFFSET;

  // 
  // now loop through the entries of the directory, allocating a set of PMDs
  // for each PGD.
  //
  pgd = nooks_mm_struct.pgd + USER_PTRS_PER_PGD;

  PDEBUG("entering outer loop\n");
  for (i = USER_PTRS_PER_PGD; i < PTRS_PER_PGD; i++) {
    pmd_t * pmd;
    
    //
    // If the pgd entry is clear, just clear the copy and continue
    //
    if (!pgd_none(*pgd)) {

      // 
      // The pgd is not clear, so allocate a set of PMDs for the pgd entry
      //
      pud = pud_offset(pgd, pgd_address);
      pmd = pmd_offset(pud, pgd_address);

      nooks_dump_pmd_set(pmd, pgd_address);
    }
    pgd_address += PGDIR_SIZE;
    pgd++;

  }
  spin_unlock(&mm->page_table_lock);
  return;  
}


/**********************************************************
 *
 * Function: nooks_compare_page_table
 *
 * Compares a page table against the kernel page table
 *
 *********************************************************/

int
nooks_compare_pte_set (pte_t * pte, 
		       pte_t * spte, 
		       unsigned long address, 
		       int flags)
{
  int result = 0;
  int i;


  for (i = 0; i < PTRS_PER_PTE; i++) {

    //
    // If the kernel is zero and we aren't 
    //

    if (spte->pte_low != pte->pte_low) {
      PDEBUG("mismatched PTEs at address %lx: %lx vs %lx\n",
	     address, spte->pte_low, pte->pte_low);
      result = -1;
    } else {
 
    if (pte_pageindex(*spte) == 0) {
      if (pte_pageindex(*pte) != 0) {
	PDEBUG("mismatched PTEs at address %lx: %lx vs %lx\n",
	       address, spte->pte_low, pte->pte_low);
	result = -1;
	// break;

      }

      // 
      // If our page indexes are different
      //
    } else if (pte_pageindex(*spte) != pte_pageindex(*pte)) {
      PDEBUG("PTE index mismatch address %lx: swapper = %lx, nooks = %lx\n",
	     address, pte_pageindex(*spte), pte_pageindex(*pte));
      result = -2;
      // break;
      //
      // If our protections are different
      //
    } else if (flags & NOOKS_IOCTL_COMP_PAGETABLE_PROT) {
      if (pte_pageprot(*spte) != pte_pageprot(*pte)) {
	PDEBUG("PTE prot mismatch address %lx: swapper = %lx, nooks = %lx\n",
	       address, pte_pageprot(*spte), pte_pageprot(*pte));
	result = -3;
	// break;
      }
      
    }
    }
    address += PAGE_SIZE;
    pte++;
    spte++;
  }
  return(result);
}

//
// Compare a set of PMD. This requires iterating through the entries in 
// the set of  PMDs, and comparing to a new set of PTEs for each PMD entry
//

int
nooks_compare_pmd_set(pmd_t * pmd, 
		      pmd_t * spmd, 
		      unsigned long address,
		      int flags)
{
  int result = 0;
  int i;
  pte_t *pte, *spte;

  // PDEBUG("comparing pmd set for pmd %lx\n", address);

  for (i = 0; i < PTRS_PER_PMD; i++ ) {
    //
    // If the kernel page table has an entry in this pmd, then copy it
    //
    if (!pmd_none(*spmd)) {
      if (pmd_none(*pmd)) {
	PDEBUG("mismatched PMDs at address %lx: %lx vs %lx\n",
	       address, spmd->pmd, pmd->pmd);
	result = -1;
	break;
      }

      pte = pte_offset_map(pmd, address);
      spte = pte_offset_map(spmd, address);

      result = nooks_compare_pte_set(pte, spte, address, flags);
      //      if (result < 0) {
      //	break;
      //      }
    } else if (!pmd_none(*pmd)) {
      PDEBUG("mismatched PMDs at address %lx: %lx vs %lx\n",
	     address, spmd->pmd, pmd->pmd);
      result = -1;
      break;
    }
    address += PMD_SIZE;
    pmd++;
    spmd++;
  }
  return(result);
}



int
nooks_compare_mm_dbg(struct mm_struct * mm_one, 
		     struct mm_struct * mm_two, 
		     int flags)
{
  int result = 0;
  int i;
  unsigned long pgd_address;
  pgd_t * pgd;
  pgd_t * spgd;
  pud_t * pud;
  pud_t * spud;
  //
  // Bail early if there is no page table to dump
  //

  if (nooks_mm_struct.pgd == NULL) {
    return(0);
  }


  // pgd_address = PAGE_OFFSET;
  pgd_address = 0;

  // 
  // now loop through the entries of the directory, allocating a set of PMDs
  // for each PGD.
  //
  pgd = mm_one->pgd; //  + USER_PTRS_PER_PGD;
  spgd = mm_two->pgd; // + USER_PTRS_PER_PGD;

  PDEBUG("entering outer loop\n");
  //  for (i = USER_PTRS_PER_PGD; i < PTRS_PER_PGD; i++) {
  for (i = 0; i < PTRS_PER_PGD; i++) {
    pmd_t * pmd, *spmd;
    
    //
    // If the pgd entry is clear, just clear the copy and continue
    //
    if (!pgd_none(*spgd)) {
      if (pgd_none(*pgd)) {
	PDEBUG("mismatched PGDs at address %lx: %lx vs %lx\n",
	       pgd_address, spgd->pgd, pgd->pgd);
	result = -1;
	break;
      }

      // 
      // The pgd is not clear, so allocate a set of PMDs for the pgd entry
      //

      pud = pud_offset(pgd, pgd_address);
      pmd = pmd_offset(pud, pgd_address);
      spud = pud_offset(spgd, pgd_address);
      spmd = pmd_offset(spud, pgd_address);

      result = nooks_compare_pmd_set(pmd, spmd, pgd_address, flags);
      if (result < 0) {
	break;
      }
    }
    pgd_address += PGDIR_SIZE;
    pgd++;
    spgd++;
  }
  return(result);  
}


/**********************************************************
 *
 * Function: nooks_protect_page_table 
 *
 * Prints out the non-null contents of a page table
 *
 *********************************************************/


void
nooks_protect_pte_set (pte_t * pte, unsigned long address, int writeable)
{
  int i;

  PDEBUG("protecting pte set %p\n", pte);
  for (i = 0; i < PTRS_PER_PTE; i++) {
    if (!pte_none(*pte)) {
      if (writeable) {
	*pte = pte_mkwrite(*pte);
      } else {
	*pte = pte_wrprotect(*pte);
      }
    } 
    address += PAGE_SIZE;
    pte++;
  }
}

void
nooks_protect_pmd_set(pmd_t * pmd, unsigned long address, int writeable)
{
  int i;
  pte_t *pte;

  PDEBUG("protecting pmd set for pmd %lx\n", address);

  for (i = 0; i < PTRS_PER_PMD; i++ ) {
    //
    // If the kernel page table has an entry in this pmd, then copy it
    //
    if (!pmd_none(*pmd)) {

      pte = pte_offset_map(pmd, address);
      nooks_protect_pte_set(pte, address, writeable);
    } 
    address += PMD_SIZE;
    pmd++;
  }
}


int nooks_mm_dbg_pt_write = 0;

void 
nooks_protect_mm(struct mm_struct * mm, 
		 int start, 
		 int count)     
{
  int i;
  unsigned long pgd_address;
  pgd_t * pgd;
  pud_t * pud;
  //
  // Indicate that future updates should make pages readable
  //


  spin_lock(&mm->page_table_lock);

  pgd_address = start << PGDIR_SHIFT;

  if (count == 0) {
    count = PTRS_PER_PGD - start;
  }


  // 
  // now loop through the entries of the directory, allocating a set of PMDs
  // for each PGD.
  //
  pgd = mm->pgd + start;

  PDEBUG("entering outer loop\n");
  for (i = start; i < start + count; i++) {
    pmd_t * pmd;
    
    //
    // If the pgd entry is clear, just  continue
    //
    if (!pgd_none(*pgd)) {

      // 
      // The pgd is not clear, so update protection on the pmd
      //
      pud = pud_offset(pgd, pgd_address);
      pmd = pmd_offset(pud, pgd_address);

      nooks_protect_pmd_set(pmd, pgd_address, nooks_mm_dbg_pt_write);
    }
    pgd_address += PGDIR_SIZE;
    pgd++;

  }
  spin_unlock(&mm->page_table_lock);
  return;  
}


void
nooks_protect_pt_mm(struct mm_struct * mm) 
{
  pgd_t * pgd;
  pud_t * pud;
  unsigned long pgd_address;
  int i;

  pgd = mm->pgd + USER_PTRS_PER_PGD;
  pgd_address = PAGE_OFFSET;
  nooks_update_addr_prot(mm, mm->pgd);

  for (i = USER_PTRS_PER_PGD; i < PTRS_PER_PGD; i++) {
    pmd_t * pmd;
    
    //
    // If the pgd entry is clear, just clear the copy and continue
    //
    if (!pgd_none(*pgd)) {

      // 
      // The pgd is not clear, so allocate a set of PMDs for the pgd entry
      //
      pud = pud_offset(pgd, pgd_address);
      pmd = pmd_offset(pud, pgd_address);

      nooks_update_pmd_set_prot(mm, pmd, pgd_address);
    }
    pgd_address += PGDIR_SIZE;
    pgd++;

  }
}

void 
nooks_use_other_pgd(struct mm_struct * mm_dest, 
		    struct mm_struct * mm_src, 
		    int start, 
		    int count)
{
  int i;
  unsigned long pgd_address;
  pgd_t * dest_pgd;
  pgd_t * src_pgd;

  //
  // Set the flag indicating that we are sharing the kernel pt
  //


  pgd_address = start << PGDIR_SHIFT;

  if (count == 0) {
    count = PTRS_PER_PGD - start;
  }

  // 
  // now loop through the entries of the directory, allocating a set of PMDs
  // for each PGD.
  //
  dest_pgd = mm_dest->pgd + start;
  src_pgd = mm_src->pgd + start;

  for (i = start; i < start + count; i++) {
    
    //
    // If the pgd entry is clear, just clear the copy and continue
    //
    dest_pgd->pgd = src_pgd->pgd;

    pgd_address += PGDIR_SIZE;
    dest_pgd++;
    src_pgd++;
  }

}
