/*
 * nooks-pool.c -- lookaside lists for nooks
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

 
#include "nooks-i.h"
#include "nooks-wrappers.h"


#if 0
typedef struct nooks_lookaside_list {
  pnooks_lookaside_elem_t * list;
  int qlen;
  int highwater;
  size_t size;
} nooks_lookaside_list_t, *pnooks_lookaside_list;
#endif

/**********************************************************
 *
 * nooks_mm_init_lookaside 
 *
 * initialize a lookaside list
 *
 *********************************************************/

void
nooks_mm_init_lookaside(pnooks_lookaside_list_t list,
			int highwater,
			int size,
			int flags)
{
  list->list = NULL;
  list->highwater = highwater;
  list->size = size;
  list->qlen = 0;
  list->flags = flags;
#ifdef NOOKS_KEEP_STATS
  list->alloc_heap = 0;
  list->alloc_pool = 0;
  list->free_heap = 0;
  list->free_pool = 0;
#endif
  spin_lock_init(&list->lock);
  
}

/**********************************************************
 *
 * nooks_mm_lookaside_alloc
 *
 * Alloc elemnt off a lookaside list
 *
 *********************************************************/

void *
nooks_mm_lookaside_alloc(pnooks_lookaside_list_t list)
{
  unsigned long flags;
  pnooks_lookaside_elem_t elem;
  spin_lock_irqsave(&list->lock, flags);
  elem = list->list;
  if (elem != NULL) {
    list->list = elem->next;
    list->qlen--;
#ifdef NOOKS_KEEP_STATS
  list->alloc_pool ++;
#endif
    spin_unlock_irqrestore(&list->lock, flags);
    return(elem);
  }
#ifdef NOOKS_KEEP_STATS
  list->alloc_heap ++;
#endif
  spin_unlock_irqrestore(&list->lock, flags);
  if (list->flags & NOOKS_LOOKASIDE_KERNEL) {
    return(nooks_kmalloc(list->size, NOOKS_MEM_FLAGS));
  } else if (list->flags & NOOKS_LOOKASIDE_NOOK) {
    return(nooks_kmalloc(list->size, NOOKS_MEM_FLAGS)); //FIXME Changed from _w_
  } else if (list->flags & NOOKS_LOOKASIDE_PAGE) {
    return((void *) nooks_mm_get_free_pages(GET_CURRENT_NOOK(),
					    NOOKS_MEM_FLAGS,
					    list->size ));
  } else if (list->flags & NOOKS_LOOKASIDE_NOOK_PAGE) {
    return((void *) nooks_mm_w_get_free_pages(NOOKS_MEM_FLAGS,
					      list->size ));
  }

  return(NULL);
}

/**********************************************************
 *
 * nooks_mm_lookaside_free
 *
 * Free an element to a lookaside list
 *
 *********************************************************/

void
nooks_mm_lookaside_free(pnooks_lookaside_list_t list, void * ptr)
{
  unsigned long flags;
  pnooks_lookaside_elem_t elem = (pnooks_lookaside_elem_t) ptr;
  spin_lock_irqsave(&list->lock, flags);

#if NOOKS_DEBUG2
  //
  // check for duplicates
  //
  {
    pnooks_lookaside_elem_t next = list->list;
    while (next != NULL) {
      if (next == ptr) {
	BUG();
      }
      next = next->next;
    }
  }
#endif

  if (in_irq() || (list->qlen < list->highwater)) {
    list->qlen++;
    elem->next = list->list;
    list->list = elem;
#ifdef NOOKS_KEEP_STATS
    list->free_pool ++;
#endif
    spin_unlock_irqrestore(&list->lock, flags);
    return;
  }
#ifdef NOOKS_KEEP_STATS
  list->free_heap ++;
#endif
  spin_unlock_irqrestore(&list->lock, flags);
      
  if (list->flags & NOOKS_LOOKASIDE_KERNEL) {
    return(nooks_kfree(elem));
  } else if (list->flags & NOOKS_LOOKASIDE_NOOK) {
    return(nooks_w_kfree(elem));
  } else if (list->flags & NOOKS_LOOKASIDE_PAGE) {
    return(nooks_mm_free_pages(GET_CURRENT_NOOK(), (unsigned long) elem, list->size));
  } else if (list->flags & NOOKS_LOOKASIDE_NOOK_PAGE) {
    return(nooks_mm_w_free_pages((unsigned long) elem, list->size));
  }

}


void
nooks_mm_free_lookaside(pnook_t nook, pnooks_lookaside_list_t list)
{
   pnooks_lookaside_elem_t next = list->list;
   pnooks_lookaside_elem_t safe;
   while (next != NULL) {
     safe = next->next;
     if (list->flags & NOOKS_LOOKASIDE_KERNEL) {
       nooks_kfree(next);
     } else if (list->flags & NOOKS_LOOKASIDE_PAGE) {
       nooks_mm_free_pages(nook, (unsigned long) next, list->size);
     }
     
     next = safe;
   }
   list->list = NULL;
}
