/*
 * nooks-skb.c -- socket buffer for nooks
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  
 * No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

 
#include "nooks-i.h"
#include "nooks-recover.h"


unsigned long nooks_num_skb_moved = 0;
__u32 nooks_mm_skb_pool_size = NOOKS_MM_SKB_POOL_SIZE;


/**********************************************************
 *
 * nooks_skb_merge_queues
 *
 * Move the SKBs off q2 and into q1 in the nook
 *
 *********************************************************/

void
nooks_skb_merge_queues(pnook_t nook,
		       struct sk_buff_head *q1, 
		       struct sk_buff_head *q2)
{
  struct sk_buff * next;
  int result;

 

  while ((next = skb_dequeue(q2)) != NULL) {
    //
    // Move this SKB into the queue
    //
    result = nooks_mm_move_skb_to_nook(nook, next);
    if (result < 0) {
      //
      // Free it again
      //
      nooks_mm_free_skb(nook, next->nook_info->order, next);
      break;
    }
    skb_queue_head(q1, next);
  }
  
}


/**********************************************************
 *
 * nooks_mm_alloc_skb
 *
 * Alloc an SKB from the pool if it only one page, or
 * from the kernel if it is larger
 *
 * Executes in the kernel
 *
 *********************************************************/

struct sk_buff *
nooks_mm_alloc_skb(unsigned int order, int gfp_mask)
{
  pnook_t nook = GET_CURRENT_NOOK(); 
  struct sk_buff * skb = NULL;
  struct sk_buff_head *list = &nook->resources.skb_pool;
  struct sk_buff_head *nook_list = &nook->locals->skb_pool;
  int result;

  // 
  // If the order is 0 (one page skb) try to find it
  // in the pool
  //

  if (order == 0) {
    if (skb_queue_len(list)) {
      unsigned long flags;

      local_irq_save(flags);
      NOOKS_STATS_ADD(NOOKS_OPS_SKB_MOVED, list->qlen);
      //
      // Move the rest of the entries to the nooks queue
      //
      skb = __skb_dequeue(list);


      nooks_skb_merge_queues(nook, nook_list, list);
      
      local_irq_restore(flags);
      NOOKS_STATS_INC(NOOKS_OPS_ALLOC_SKB_POOL);

      if (nooks_mm_move_skb_to_nook(nook, skb) < 0) {
	nooks_mm_free_pages(nook, (unsigned long) skb, order);
	skb = NULL;
      }
      goto Cleanup;
    }
  }
  
  //
  // If we havent' found an SKB yet, get one from the kernel
  //

  skb = (struct sk_buff *) nooks_mm_get_free_pages(nook, gfp_mask, order);
  if (skb != NULL) {
    result = nooks_add_to_hash(&nook->resources.active_skbs, 
			       skbuff_n_e,
			       skb, 
			       (void *) (PAGE_SIZE << order));
    
    if (result < 0) {
      nooks_mm_free_pages(nook, (unsigned long) skb, order);
      skb = NULL;
    }
  }


 Cleanup:
  return(skb);
}


/**********************************************************
 *
 * nooks_alloc_skb
 *
 * Alloc an SKB from the pool if it only one page, or
 * from the kernel if it is larger
 *
 * Executes wholly within the nook
 *
 * 
 *
 *********************************************************/

struct sk_buff *
nooks_alloc_skb(pnook_t nook, unsigned int order)
{
  struct sk_buff * skb = NULL;
  struct sk_buff_head *list = &nook->locals->skb_pool;

  // 
  // If the order is 0 (one page skb) try to find it
  // in the pool
  //

  if (order == 0) {
    if (skb_queue_len(list)) {
      unsigned long flags;

      local_irq_save(flags);
      skb = __skb_dequeue(list);
      local_irq_restore(flags);
      NOOKS_STATS_INC(NOOKS_OPS_ALLOC_SKB_NOOK_POOL);
    }
  }
  

  return(skb);
}


/**********************************************************
 *
 * nooks_mm_free_skb
 *
 * Free a skb to the pool
 *
 *********************************************************/

void
nooks_mm_free_skb(pnook_t nook, unsigned int order, struct sk_buff * skb)
{
  struct sk_buff_head *list = &nook->resources.skb_pool;
	
  //
  // If this is only one page, and our pool is not too large already
  // add it to the pool
  //

  if ((order == 0) && 
      (skb_queue_len(list) < nooks_mm_skb_pool_size)) {
    unsigned long flags;
    
    local_irq_save(flags);
    __skb_queue_head(list, skb);
    local_irq_restore(flags);
    NOOKS_STATS_INC(NOOKS_OPS_KFREE_SKB_POOL);
  } else {

    //
    // Else, free the pages
    //

    nooks_mm_free_pages(nook, (unsigned long) skb, order);
  }
}



/**********************************************************
 *
 * nooks_mm_move_skb_to_kernel
 *
 * remove access by the nook to an SKB so it can be safely
 * used in the kernel
 *
 *********************************************************/

int
nooks_mm_move_skb_to_kernel(pnook_t nook, struct sk_buff * skb) 
{
  int result;
  unsigned long size = 0;
  void * nook_info_ptr;
  result = copy_from_user(&nook_info_ptr, &skb->nook_info, (unsigned long)sizeof(void *));
  if (result < 0) {
    NOOKS_FAULT(resource);
    printk ("nooks_mm_move_skb_to_kernel: Failed.\n");
    return(result);
  }

  NOOKS_STATS_INC(NOOKS_OPS_SKB_MOVED_TO_KERNEL);
  if (nook_info_ptr != NULL) {
    //
    // Release the memory
    //
    result = nooks_del_from_hash(&nook->resources.active_skbs, 
				 skbuff_n_e,
				 skb,
				 (void **) &size );
    if (result >= 0) {


    } else {
      printk("nooks_mm_move_skb_to_kernel:skb not in hash %p\n",
	     skb);
    }
  } else {
    //
    // Just delete the record of the skb
    //
    result = nooks_del_resource_elem(nook, skbuff_k_e, skb, NULL);
    //if (result < 0)
    //	printk ("nooks_mm_move_skb_to_kernel: Error deleting nook record from skb. skb's may NOT be getting allocated from nook memory.\n");
  }
  return(result);
}

/**********************************************************
 *
 * nooks_mm_move_skb_to_nook
 *
 *
 * Allow a nook to access an SKB
 * 
 *********************************************************/

int
nooks_mm_move_skb_to_nook(pnook_t nook, struct sk_buff * skb)
{
  int result;

  NOOKS_STATS_INC(NOOKS_OPS_SKB_MOVED_TO_NOOK);

  if (skb->nook_info != NULL) {
    //
    // Add the pages and a record of the SKB
    //
    result = nooks_add_to_hash(&nook->resources.active_skbs, 
			       skbuff_n_e,
			       skb, 
			       (void *) (PAGE_SIZE << skb->nook_info->order));
    

  } else {
    //
    // Just add a record of the SKB so we can free it later
    //

    result = nooks_add_resource_elem(nook, skbuff_k_e, skb, NULL);
  }
  return(result);

}

/**********************************************************
 *
 * nooks_mm_free_skb_pool
 *
 * 
 *
 *********************************************************/

void
nooks_mm_free_skb_pool(pnook_t nook)
{
  struct sk_buff_head *list = &nook->resources.skb_pool;
  struct sk_buff * skb;

  // 
  // If the order is 0 (one page skb) try to find it
  // in the pool
  //

  while ((skb = __skb_dequeue(list)) != NULL) {
    nooks_mm_free_pages(nook, (unsigned long) skb, skb->nook_info->order);
  }
}
