/*
 * nooks-switch.c -- code for context switching between nooks 
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

 
#include "nooks-i.h"
#include "nooks-switch.h"
#include "nooks-sys.h"
#include "nooks-work.h"
#include "nooks-recover.h"
#include "nooks-wrappers.h"
#include <asm/mmu_context.h>

_nooks_call0(unsigned long, nooks_dummy_trap1);
_nooks_call0(unsigned long, nooks_dummy_trap2);

unsigned long nooks_illegal_inst_trap(void)
{
  return(0);
}


/**********************************************************
 *
 * nooks_get_stack_space
 *
 * get stack space and update nook thread state
 *
 *********************************************************/

int
nooks_get_stack_space(pnook_t nook, 
		      struct nook_thread_state * nook_thread_state,
		      int callreturn)
{
  void * stack_base;
  void * stack_top = NULL;
  
  

  //
  // If we were previously in this thread, then we have a stack
  // already
  //

  if ((nook == nook_thread_state->nook) && 
      (nook_thread_state->trap_state.nook_stack != NULL) &&
      (nook_thread_state->depth > 0)) {
    
    if (callreturn == SWITCH_IN) {

      //
      // Bump up the recursion count for the stack
      //

      nook_thread_state->depth++;
      stack_top = nook_thread_state->trap_state.nook_stack;    
      NOOKS_STATS_INC(NOOKS_OPS_STACK_NOT_ALLOC);
    } else  {
      
      //
      // We are returning from a call - do nothing
      //

      stack_top = nook_thread_state->trap_state.nook_stack;
      stack_base = nook_thread_state->stack_base;
    } 
  } else {

    //
    // BUGBUG: this doesn't handle back-and-forth calls
    //

    NASSERT(callreturn == SWITCH_IN);
    stack_base = nooks_mm_lookaside_alloc(&nook->resources.stack_pool);
    if (stack_base == NULL) {
      PERROR("Failed to allocate stack!!!\n");
      goto Cleanup;
    }
    stack_top = stack_base + NOOKS_STACK_SIZE;


    nook_thread_state->stack_base = stack_base;
    nook_thread_state->depth = 1;
    nook_thread_state->nook = nook;
    nook_thread_state->nook_frame = NULL;
  }

  if (callreturn == SWITCH_IN) {

    pnooks_stack_frame_t frame;
    
    //
    // Store the frame pointer.
    //

    //
    // BUGBUG: The frame needs to be on the kernel stack, not the nook stack,
    // because it contains privileged (can't be corrupted) data. Except for
    // the work list ... Perhaps two frames...
    //
    
    stack_top = (void *) ((unsigned long) stack_top - sizeof(nooks_stack_frame_t));
    frame = (pnooks_stack_frame_t) stack_top;
    
    memset(&frame->work_list, 0, sizeof(struct list_head));
    
    nook_thread_state->nook_frame = frame;
    
    //
    // We are ready to record that the new stack is ready
    //
	   
    nook_thread_state->trap_state.nook_stack = stack_top;
  }
  //NASSERT((unsigned long) nook_thread_state->trap_state.nook_stack - 
//	  (unsigned long) nook_thread_state->stack_base > 100);
  //NASSERT(nooks_mm_get_pte_flags(nook, stack_top) & _PAGE_RW);

 Cleanup:
  if (stack_top == 0) {
    return(-ENOMEM);
  } else {
    return(0);
  }
}


/**********************************************************
 *
 * nooks_context_switch_in
 *
 * Switch to executing within a nook
 *
 * callreturn - if non-zero, this is a call return so
 * we should use the previous stack as it.
 *
 *********************************************************/

extern unsigned long softirq_call_count;

void *
nooks_context_switch_in(pnook_t nook, int callreturn)
{
  struct nook_thread_state * nook_thread_state = 
    &(GET_NOOK_STATE().thread_state);

  if (callreturn == SWITCH_IN) {
    if (nook->livelock_state.last_clock == softirq_call_count) {
      nook->livelock_state.calls_since_last_clock++;
      if (nook->livelock_state.calls_since_last_clock > 
	  NOOKS_DEADLOCK_CALL_LIMIT) {
	//
	// Trigger a fault
	//
	NOOKS_FAULT(deadlock);
	//nooks_fault_handler(nook);
	return(NULL);
      }
    } else {
      nook->livelock_state.calls_since_last_clock = 0;
      nook->livelock_state.last_clock = softirq_call_count;
    }
  }

  if (nooks_get_stack_space(nook, nook_thread_state, callreturn) < 0) {
    return(NULL);
  }
    
  nook_thread_state->trap_state.pgd = NULL; 
//    nook->resources.mm->pgd;

  return(nook_thread_state->trap_state.nook_stack);
}

/**********************************************************
 *
 * nooks_release_stack_space
 *
 * release the stack back and update nook thread state
 *
 *********************************************************/

void
nooks_release_stack_space(struct nook_thread_state * nook_thread_state, int callout)
{
  pnook_t last = GET_CURRENT_NOOK();

  //
  // If we were previously in this thread, then we have a stack
  // already
  //

  if (callout == SWITCH_OUT) {
    //
    // Reduce the depth count
    //

    //
    // BUGBUG: do we have a race condition here?
    //

    nook_thread_state->depth--;

    //
    // Restore the frame
    //

    //
    // Pop the frame
    //

    //
    // The stack end just after the old frame
    //

    //    frame = nook_thread_state->nook_frame;
    //    nook_thread_state->nook_frame = nook_thread_state->kernel_frame->thread_state.nook_frame;

    //
    // Make sure to only update the nook-stack variable after the frame has
    // been constructed, because an interrupt could occur here and write to
    // memory after the nook_stack pointer.
    //

    //    nook_thread_state->trap_state.nook_stack = 
    //      nook_thread_state->kernel_frame->thread_state.trap_state.nook_stack;

    //
    // If we are at the top of this stack, free it to the pool
    //

    if (nook_thread_state->depth == 0) {
      nooks_mm_lookaside_free(&last->resources.stack_pool, 
			      nook_thread_state->stack_base);
    } else {
      NOOKS_STATS_INC(NOOKS_OPS_STACK_NOT_FREE);
    }
  }
}

/**********************************************************
 *
 * nooks_context_switch_out
 *
 * callout - if non-zero, then this is a call-out context
 * switch
 *
 * switch to executing in the kernel from a nook
 * returns the nook that was executing
 *
 *********************************************************/

pnook_t
nooks_context_switch_out(int callout)
{
  struct nook_thread_state * nook_thread_state = 
    &(GET_NOOK_STATE().thread_state);
  pnook_t last = GET_CURRENT_NOOK();


  
  nook_thread_state->trap_state.pgd = NULL;


  nooks_process_work_list(last, callout);

  nooks_release_stack_space(nook_thread_state, callout);

  return(last);

}


void
nooks_context_switch_set_param(unsigned long param)
{

}

#if 0 
/**********************************************************
 *
 * nooks_driver_call
 *
 * Call into a driver from the kernel, by switching stacks
 * and possibly lowering privilege levels
 *
 *********************************************************/
extern void __struct_cpy_bug (void);

unsigned long 
nooks_driver_call(void * func, int argc, ...)
{
  /*
   * We get link-time errors if the structure sizes do not match.
   * There is no runtime overhead, it's all optimized away at
   * compile time.
   */

#define struct_cpy(x,y)                         \
({                                              \
        if (sizeof(*(x)) != sizeof(*(y)))       \
                __struct_cpy_bug;               \
        memcpy(x, y, sizeof(*(x)));             \
})

  unsigned long result = 0;
  NOOKS_STATS_INC(NOOKS_OPS_DRIVER_CALLS);
  NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_DRIVER_CALL);
  {
  va_list args;
  int local_argc = argc;
  unsigned long arg[NOOKS_DRIVER_CALL_MAX_PARAM];
  pnook_t nook = GET_CURRENT_NOOK();
  struct task_struct * task = get_current();
  pnook_thread_state_t nook_thread_state = &(GET_NOOK_STATE().thread_state);
  void * stack;
  unsigned long flags;
  nooks_kernel_stack_frame_t frame;

  NASSERT(!IN_NOOK());

  struct_cpy(&frame.thread_state, nook_thread_state);
  //FIXME struct_cpy(&nooks_w_irq_stat[0], &irq_stat[0]);
  

  PERROR("Making driver call: %p\n", func);

  va_start(args, argc);
  while (local_argc > 0) {
    arg[local_argc-1] = va_arg(args, unsigned long);
    local_argc--;
  }
  va_end(args);

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_COPIED_ARGS);

  NOOKS_PROCESS_LIST(nook, NOOKS_WORK_FLAG_ENTRY | NOOKS_WORK_FLAG_KERNEL);

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_PROCESSED_LIST1);

  //
  // Set up for the context switch - get the stack
  //

  stack = nooks_context_switch_in(nook, SWITCH_IN);
  if (stack == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_CS_IN);

  local_irq_save(flags);

  //
  // Check to see if the nook is still running - if not, bail out now
  //

  if (!NOOK_OK(nook)) {
    local_irq_restore(flags);
    nooks_context_switch_out(SWITCH_OUT);
    printk ("Nook not running -- bailing out.\n");
    result = -EFAULT;
    goto Cleanup;
  }


  //
  // Swap stacks. First save the old saved stack and frame pointers on the 
  // stack, then save the stack and frame in the task, and load the stack
  // pointer. The frame pointer is still valid, allowing us to read
  // values from this function.
  //


  nook_thread_state->kernel_frame = &frame;
  
  __asm__ __volatile__ 
         ("movl %%esp, %0\n\t"
	  "movl %%ebp, %1\n\t"
	  "movl %2, %%esp\n\t"
	  : "=m" (nook_thread_state->trap_state.kernel_stack),
	    "=m" (frame.frame_register)
	  : "m" (stack)
	  : "esp" );

  //
  // Change page tables
  //

  ENTER_NOOK();

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_SWAP_REGS);
   
  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_SWAP_PT);

  NOOKS_PROCESS_LIST(GET_CURRENT_NOOK(), NOOKS_WORK_FLAG_ENTRY | NOOKS_WORK_FLAG_NOOK);

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_PROCESSED_LIST2);

  //
  // Push the args on the new stack
  //

  for (local_argc = 0; local_argc < argc; local_argc++) {
    __asm__ __volatile__
            ("movl %0, %%eax \n\t"
	     "pushl %%eax" 
	     : : "m" (arg[local_argc]) 
	     : "esp","eax"); 
  }

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_PUSH_ARGS);

  //
  // Call the function, then save the result on the stack.
  // We then get the current task, move it to ebx, and pop the result
  // into eax. Then, we restore the stack and frame pointers, and
  // then pop the old stack and frame pointers into the task structure
  //

  local_irq_restore(flags);

  //
  // BUGBUG: can we do this here?
  //


  //FIXME need to fix both assembly codes below for 2.6 version.
  /*
  __asm__ __volatile__ 
         ("movl aligned_data, %%eax \n\t"
	  "movl " ASM_TASK_NOOK_OFFSET_STR "(%%eax), %%edx \n\t"
	  "pushl %%edx \n\t"
	  "call nooks_free_work_list \n\t"
	  "popl %%eax \n\t"
	  ::  
	  );
   */ 
   NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_DRIVER_CALLOUT);

  __asm__ __volatile__
         ("call *%0 \n\t"
	  "pushl %%eax \n\t"
	  "pushfl \n\t"	  
	  "cli \n\t"
	  : : "m" (func)
	  : "memory" );
  
  NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_DRIVER_CALLRETURN);

  NOOKS_PROCESS_LIST(GET_CURRENT_NOOK(), NOOKS_WORK_FLAG_EXIT | NOOKS_WORK_FLAG_NOOK);

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_PROCESS_LIST3);

  //
  // Change page tables
  //
   
  //
  // Switch from the nook mm to the active mm. FIXME
  //
 
  __asm__ __volatile__ 
    (".globl nooks_driver_call_fault\n\t"
     "nooks_driver_call_fault: \n\t"
//    "movl aligned_data, %0\n\t" : "=c" (task) );
     "movl %0, %0\n\t" : "=c" (task) ); //FIXME
/*  nooks_switch_out_mm(task->nook_state.curr.current_nook->resources.mm,
		      (task->active_mm ? task->active_mm : &init_mm),
		      smp_processor_id());

*/
  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_SWAP_PT2);

  EXIT_NOOK();
  /* FIXME
  __asm__ __volatile__
         ("movl %1, %%ebx\n\t"
	  "popl %%ecx \n\t"
	  "popl %%eax \n\t"
	  "movl " ASM_TASK_KERNEL_FRAME_OFFSET_STR "(%%ebx), %%edx \n\t"
	  "movl " ASM_TASK_KERNEL_STACK_OFFSET_STR "(%%ebx), %%esp \n\t"
	  "movl " ASM_NOOK_KERNEL_FRAME_OFFSET_STR "(%%edx), %%ebp \n\t"
	  "movl %%eax, %0 \n\t"
	  "pushl %%ecx \n\t"
	  "popfl \n\t"
	  : "=&a" (result) 
	  : "m" (* (unsigned long *) aligned_data)
	  : "memory", "ebx", "ecx", "edx");
  */
  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_SWAP_REGS2);

  if (NOOK_OK(GET_CURRENT_NOOK())) {
    nooks_context_switch_out(SWITCH_OUT);

    NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_CS_OUT);

    NOOKS_PROCESS_LIST(GET_CURRENT_NOOK(), 
		       NOOKS_WORK_FLAG_EXIT | NOOKS_WORK_FLAG_KERNEL);

    NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_DRIVER_PROCESSED_LIST4);
  } else {
    //
    // If we were signalled to crash, convert this into a restartsys result. set up
    // a handler that ignores the signal, so we will rerun the call
    //

    if (signal_pending(current) && sigismember(&current->pending.signal, SIGCRASH)) {
      struct k_sigaction *ka;
      ka = &current->sighand->action[SIGCRASH-1];
      ka->sa.sa_handler = SIG_IGN;
      result = -ERESTARTSYS;
    }
  }
  PDEBUG("Returning from driver call: %p\n", func);
 Cleanup:
  struct_cpy(nook_thread_state, &frame.thread_state);
  }
  NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_DRIVER_RETURN);

  return(result);
}

#endif

#if 0
/**********************************************************
 *
 * nooks_kernel_call
 *
 * Call into the kernel from a driver, by switching stacks
 * and possibly raising privilege levels
 *
 * The complication here is that there is work to do when
 * we enter the kernel, so we need to call nooks_context_switch_out
 * before calling our worker function
 *
 *********************************************************/

unsigned long 
nooks_kernel_call(void * func, int argc, ...)
{
  unsigned long result;
  NOOKS_STATS_INC(NOOKS_OPS_KERNEL_CALLS);
  NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_KERNEL_CALL);
  {
  va_list args;
  size_t argsize = argc * sizeof(unsigned long);
  int local_argc = argc;
  unsigned long arg[NOOKS_DRIVER_CALL_MAX_PARAM];
  unsigned long flags;
  struct task_struct * task = get_current();
  pnook_t nook;
  result = 0;

  NASSERT(IN_NOOK());

  va_start(args, argc);
  while (local_argc > 0) {
    arg[local_argc-1] = va_arg(args, unsigned long);
    local_argc--;
  }
  va_end(args);

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_KERNEL_COPIED_ARGS);

  NOOKS_PROCESS_LIST(GET_CURRENT_NOOK(), NOOKS_WORK_FLAG_EXIT | NOOKS_WORK_FLAG_NOOK);

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_KERNEL_PROCESSED_LIST1);

  local_irq_save(flags);
  

  //
  // Change page tables - we have to do this before swapping stack, because 
  // otherwise we won't have the right to write to the new stack
  //
    
  //
  // Switch from the nook MM to the kernel MM
  //
  
  nooks_switch_out_mm(task->nook_state.curr.current_nook->resources.mm,
		      (task->active_mm ? task->active_mm : &init_mm), 
		      smp_processor_id());
  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_KERNEL_SWAP_PT);
  

  //
  // Swap stacks. The nook stack is stored in the stack_top variable
  // the kernel stack is restored from the kernel_stack variable
  //
  
  __asm__ __volatile__
    ("pushl %2\n\t"
     "movl %%esp, %0\n\t"
     "movl %1, %%esp\n\t"
     : "=m" (task->nook_state.thread_state.trap_state.nook_stack)
     : "m" (task->nook_state.thread_state.trap_state.kernel_stack),
     "0" (task->nook_state.thread_state.trap_state.nook_stack)
     : "esp" );
  
  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_KERNEL_SWAP_REGS);
  
  
  //
  // Now we have swapped stack, but our frame pointer is still
  // valid
  //
  
  //
  // Call the nooks context switch routine. We do this on the
  // kernel stack, but it has to save the old frame pointer
  //
  
  //    __asm__ __volatile__
  //  ("pushl %0 \n\t"
  //   "call *%1 \n\t"
  //   "popl %%eax"
  //   : : "i" (SWITCH_CALLOUT), "m" (nooks_context_switch_out)
  //   : "memory" );
  

  EXIT_NOOK();

  __asm__ __volatile__
    ("pushl %0 \n\t"
     "call nooks_context_switch_out \n\t"
     "popl %%eax"
     : : "i" (SWITCH_CALLOUT)
     : "memory" );
  
  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_KERNEL_CS_OUT);
  
  //
  // Copy the arguments from the old  stack (through the old
  // frame pointer) to the kernel stack
  //
      
  local_irq_restore(flags);

  NOOKS_PROCESS_LIST(GET_CURRENT_NOOK(), NOOKS_WORK_FLAG_EXIT | NOOKS_WORK_FLAG_KERNEL);

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_KERNEL_PROCESSED_LIST2);

  for (local_argc = 0; local_argc < argc; local_argc++) {
    __asm__ __volatile__
      ("movl %0, %%eax \n\t"
       "pushl %%eax \n\t" 
       : : "m" (arg[local_argc]) 
       : "esp","eax"); 
  }

  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_KERNEL_PUSH_ARGS);

  //
  // Call the kernel function, saving the result in the nook-stack
  //

  // 
  // BUGBUG: it is unsafe to get argsize from the old stack -
  // it could have been changed by the nook. Since we are returning,
  // we could probably safely not pop the arguments.
  //
  
  //
  // The return value is pushed on the stack, and then popped later
  // into eax after calling context_switch_in
  //

  NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_KERNEL_CALLOUT);

  __asm__ __volatile__
    ("call *%0 \n\t"
     "addl %1, %%esp \n\t"
     "pushl %%eax \n\t"
     "pushfl \n\t"
     : 
     : "m" (func) , "m" (argsize) : "memory" );

  NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_KERNEL_CALLRETURN);

  //
  // Test whether the nook is still functioning. If not, we want
  // to return to our previous recovery point instead of returning to
  // the nook.
  //
/*
  if (!NOOK_OK(GET_CURRENT_NOOK())) {
    __asm__ __volatile__ 
      ("jmp nooks_driver_fault\n\t" );

  }
*/
  //
  // We then get the task structure, and context switch back to the nook.
  // We then replace the stack pointer with the one from the task.
  // We don't copy the stack pointers for the kernel back, because we 
  // don't want to save them
  //
  
  
  //
  // BUGBUG: if nooks_context_switch_in fails, we don't do anything here
  //
  // NEED TO FIXME the below assembly for 2.6 version.
  /*
  __asm__ __volatile__
    ("movl %1, %%ebx \n\t"
     "pushl %3 \n\t"
     "pushl " ASM_TASK_NOOK_OFFSET_STR "(%%ebx) \n\t"
     "call nooks_process_entry_exit_list \n\t"
     "popl %%ecx \n\t"
     "popl %%eax \n\t"
     "pushl %2 \n\t"
     "pushl %%ecx \n\t"
     "call nooks_context_switch_in \n\t"
     "popl %%eax \n\t"
     "popl %%eax \n\t"
     "popl %%ecx \n\t"
     "popl %%eax \n\t"
     "movl " ASM_TASK_NOOK_STACK_OFFSET_STR "(%%ebx), %%esp \n\t" 
     "popl " ASM_TASK_NOOK_STACK_OFFSET_STR "(%%ebx) \n\t"
     "movl %%eax, %0 \n\t"
     "pushl %%ecx \n\t"
     "popfl"
     : "=ma" (result)
     : "m" (* (unsigned long *) aligned_data),
     "i" (SWITCH_CALLRETURN), 
     "i" (NOOKS_WORK_FLAG_ENTRY | NOOKS_WORK_FLAG_KERNEL)
     : "memory", "ebx", "ecx");
  */ 
  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_KERNEL_RETURN_GOOP);
  
  //
  // Change page tables
  //

  ENTER_NOOK();
  
  //
  // Switch from the current PT back to the nook PT
  //
 /* 
  nooks_switch_in_mm((task->active_mm ? task->active_mm : &init_mm),
		     task->nook_state.curr.current_nook->resources.mm, 
		     smp_processor_id());
 */
  
  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_KERNEL_CS_IN);
  nook = GET_CURRENT_NOOK();
  NOOKS_PROCESS_LIST(nook, NOOKS_WORK_FLAG_ENTRY | NOOKS_WORK_FLAG_NOOK);
  NOOKS_STATS_ADD_EVENT2(NOOKS_EVENT_KERNEL_PROCESSED_LIST3);

  nooks_free_work_list(nook);
  }
  NOOKS_STATS_ADD_EVENT(NOOKS_EVENT_KERNEL_RETURN);

  return(result);
}
#endif 

/**********************************************************
 *
 * nooks_dwrapper_stack_alloc
 *
 *
 * Allocate space on the nook stack for holding temporary 
 * parameters to drivers
 * 
 *
 *********************************************************/

#ifndef CONFIG_NOOKS_NOPROT
void * 
nooks_dwrapper_stack_alloc(size_t size)
{
  struct nook_thread_state * nook_thread_state = 
    &(GET_NOOK_STATE().thread_state);
  pnook_t nook = GET_CURRENT_NOOK();
  pnooks_alloc_stack_frame_t frame = NULL;
  nooks_alloc_stack_frame_t old_frame;
  void * result = NULL;

  old_frame.old_stack = nook_thread_state->trap_state.nook_stack;
  old_frame.old_frame = nook_thread_state->nook_frame;

  if (nooks_get_stack_space(nook, nook_thread_state, SWITCH_IN) < 0) {
    goto Cleanup;
  }

  //
  // We store the frame at the top of the stack, and the data goes
  // just after it
  //

  frame = nook_thread_state->trap_state.nook_stack - 
    ROUND_UP_COUNT(size + sizeof(nooks_alloc_stack_frame_t), sizeof(long));
  *frame = old_frame;
  result = (frame + 1);
  nook_thread_state->trap_state.nook_stack = frame;

 Cleanup:
  return(result);
}


/**********************************************************
 *
 * nooks_dwrapper_stack_free
 *
 * Release all space allocated on the nook stack
 *
 *********************************************************/

void
nooks_dwrapper_stack_release(void)
{
  struct nook_thread_state * nook_thread_state = 
    &(GET_NOOK_STATE().thread_state);
  pnooks_alloc_stack_frame_t frame = nook_thread_state->trap_state.nook_stack;
  nooks_release_stack_space(nook_thread_state, SWITCH_OUT);
  nook_thread_state->trap_state.nook_stack = frame->old_stack;
  nook_thread_state->nook_frame = frame->old_frame;
}
#endif // CONFIG_NOOKS_NOPROT
