/*
 * nooks-sys.c -- nooks-related system calls
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

 
#include "nooks-i.h"
#include "nooks-sys.h"
#include "nooks-util.h"
#include "nooks-switch.h"
#include "nooks-recover.h"
#include "nooks-wrappers.h"
#include <../shadow/netshadow.h>
#ifdef CONFIG_NOOKS_PROC
#include <../shadow/shadow-proc.h>
#endif


char *
nooks_get_nook_name_from_user(const char * name_user)
  
{
  char * name = NULL;
  int namelen;
  int result = 0;

  result = nooks_get_user_string(name_user, &name, &namelen);
  if (result < 0) {
    name = NULL;
    goto Cleanup;
    
  }

  if ((name == NULL) || (namelen <= 0) || (namelen > NOOK_MAX_NAME_LEN)) {
    if (name != NULL) {
      nooks_free_user_string(name);
    }
    goto Cleanup;
  }

 Cleanup:
  return(name);
}

/**********************************************************
 *
 * sys_nooks_create_nook
 *
 * Create a nook
 *
 *********************************************************/
asmlinkage long 
sys_nooks_create_nook(char * name_user)
{
  char * name = NULL;
  int result = 0;
  pnook_t nook = NULL;
  NOOKS_ENTER();



  if (!capable(CAP_SYS_MODULE)) {
    PERROR(" caller doesn't have permission to create nooks\n");
    result = -EPERM;
    goto Cleanup;
  }

  //
  // Capture the name of the nook from usermode
  //

  name = nooks_get_nook_name_from_user(name_user);
  if (name == NULL) {
    result = -EINVAL;
    goto Cleanup;
  }

  //
  // Create a nook with this name
  //

  result = nooks_i_create_nook(name, &nook);
  if (result < 0) {
    goto Cleanup;
  }

  nooks_validate_nook(nook);

  //  TEMPORARY NOTE!	
  //  sys_nooks_ioctl(name_user, 10007, NULL);


 Cleanup:
  if (name != NULL) {
    nooks_free_user_string(name);
  }
  if (result < 0) {
    if (nook != NULL) {
      nooks_free_nook(nook);
    }
  } else {
    nooks_validate_nook(nook);
  }

  NOOKS_EXIT();
  return(result);


}


/**********************************************************
 *
 * sys_nooks_delete_nook
 *
 * Delete a nook
 *
 *********************************************************/
asmlinkage long 
sys_nooks_delete_nook(char * name_user)
{
  char * name = NULL;
  int result = 0;
  pnook_t nook = NULL;
  NOOKS_ENTER();

  if (!capable(CAP_SYS_MODULE)) {
    PERROR(" caller doesn't have permission to delete nooks\n");
    result = -EPERM;
    goto Cleanup;
  }


  //
  // Capture the name of the nook from usermode
  //

  name = nooks_get_nook_name_from_user(name_user);
  if (name == NULL) {
    result = -EINVAL;
    goto Cleanup;
  }
  //
  // Find a nook with this name
  //

  nook = nooks_find_nook(name);
  if (nook == NULL) {
    result = -ENOENT;
    goto Cleanup;
  }

#ifdef CONFIG_NOOKS_PROC
  if (nook->proc_ent)
  	shadow_proc_del_entry(nook);
  if (nook->last_fault_mod)
	kfree(nook->last_fault_mod);
#endif
  
  //
  // Dereference out copy of the nook and free it.
  //

  nooks_dereference_nook(nook);

  //
  // Dereference it again to remove it from the list. If the refcount is
  // zero, then we should free it. Otherwise the other user of it will free
  // it.
  //

  nooks_dereference_nook(nook);



 Cleanup:
  if (name != NULL) {
    nooks_free_user_string(name);
  }

  NOOKS_EXIT();
  return(result);
}


/**********************************************************
 *
 * sys_nooks_query_nook
 *
 * Query properties about a nook
 *
 *********************************************************/
asmlinkage long 
sys_nooks_query_nook(const char *name_user, 
		     int which, 
		     char *buf, 
		     size_t bufsize,
		     size_t *ret)
{
  pnook_t nook = NULL;
  int result;
  char *name = NULL;
  NOOKS_ENTER();
  
  //
  // Capture the name of the nook from usermode
  //

  name = nooks_get_nook_name_from_user(name_user);
  if (name == NULL) {
    result = -EINVAL;
    goto Cleanup;
  }
  result = -ENOENT;
  if ((nook = nooks_find_nook(name)) == NULL) {
    result = -ENOENT;
    goto Cleanup;
  }
  switch (which)
    {
    case NOOKS_QUERY_STATE:
      result = (long) nook->state;
      break;
    default:
      result = -EINVAL;
      break;
    }

 Cleanup:
  if (nook != NULL) {
    nooks_dereference_nook(nook);
  }
  if (name != NULL) {
    nooks_free_user_string(name);
  }
  NOOKS_EXIT();
  return result;
}

/**********************************************************
 *
 * sys_nooks_enum_nooks
 *
 * Enumerate the nooks in the system
 *
 *********************************************************/
asmlinkage long 
sys_nooks_enum_nooks(char * user_buf, 
		     size_t bufsize,
		     size_t *ret)
{
  pnook_t nook = NULL;
  struct list_head * next = NULL;
  int result = 0;
  size_t size = 0;
  char * buf = NULL;

  printk ("In sys_nooks_enum_nooks syscall.\n");
  NOOKS_ENTER();


  // 
  // Calculate the size required for the nooks.
  //

    
  spin_lock(&nooks_list_lock);

  //
  // Enumerate the nooks, counting the size of each nook
  //

  list_for_each(next, &nook_list) {
    nook = list_entry(next,nook_t, next);
    //
    // Add space for the name and a trailing null
    //
    size += strlen(nook->name) + 1;
  }
  spin_unlock(&nooks_list_lock);
  
  //
  // Add space for a trailing null for the list
  //
  size += 1;

  if (bufsize < size) {
    result = -ENOSPC;
    goto Cleanup;
  }
  
  //
  // allocate bufsize space, so that we are guaranteed to have enough
  // space even if the list changes
  //

  buf = (char *) nooks_kmalloc(bufsize, GFP_KERNEL);
  if (buf == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }

  //
  // Copy the strings. We need to be careful because the list may have
  // changed, so we may not have space anymore.
  //

  size = 0;
  spin_lock(&nooks_list_lock);

  list_for_each(next, &nook_list) {
    int len;
    nook = list_entry(next,nook_t, next);
    len = strlen(nook->name) + 1;

    //
    // Add space for the name and a trailing null
    //
    if (size + len + 1 > bufsize) {
      result = -ENOSPC;
      goto Cleanup;
    }
    memcpy(buf+size, nook->name, len);
    size += len;
  }
  //
  // Add a second trailing null
  //

  buf[size++] = '\0';

  spin_unlock(&nooks_list_lock);

  if ((result == 0) || (result == -ENOSPC)) {
    if (copy_to_user(user_buf, buf, size)) {
      PERROR(" failed to copy buffer to user\n");
      result = -EFAULT;
    }
  }
      
 Cleanup:
  if (buf != NULL) {
    nooks_kfree(buf);
  }

  // 
  // Copy the size back to the application
  //

  if ((result == 0) || (result == -ENOSPC)) {
    if (copy_to_user(ret, &size, sizeof(size_t))) {
      result = -EFAULT;
      PERROR(" failed to copy size to user: %d\n", result);
    }
  }
  
  
  NOOKS_EXIT();
  return result;
}


/**********************************************************
 *
 * sys_nooks_ioctl
 *
 * General backdoor interface to nooks
 *
 * name_user - name of nook or NULL for no nook
 *
 * cmd: one of NOOKS_IOCTL_XXX
 *
 *********************************************************/

asmlinkage long 
sys_nooks_ioctl(char * name_user, unsigned int cmd, unsigned long arg)
{	
  long result = 0;
  char * name = NULL;
  pnook_t nook = NULL;

  NOOKS_ENTER();

  if (!capable(CAP_SYS_MODULE)) {
    PERROR(" caller doesn't have permission to ioctl nooks\n");
    result = -EPERM; 
   goto Cleanup;
  }

  if (name_user != NULL) {
    //
    // Capture the name of the nook from usermode
    //
    name = nooks_get_nook_name_from_user(name_user);
    if (name == NULL) {
      result = -EINVAL;
      goto Cleanup;
    }
    //
    // Find a nook with this name
    //
    if (cmd < NOOKS_IOCTL_LAST) {
      nook = nooks_find_nook(name);
      if (nook == NULL) {
	result = -ENOENT;
	goto Cleanup;
      }
    }
  }

  //
  // Do the actual work
  //

  switch(cmd) {
  case NOOKS_IOCTL_VALIDATE_NOOKS:
    {
      struct list_head * next = NULL;
      spin_lock(&nooks_list_lock);

      //
      // Enumerate the nooks, counting the size of each nook
      //
      
      list_for_each(next, &nook_list) {
	nook = list_entry(next,nook_t, next);
	nooks_validate_nook(nook);
      }
      spin_unlock(&nooks_list_lock);
      break;
    }
  case NOOKS_IOCTL_DUMP_DATA:
    {
      struct list_head * next = NULL;
      spin_lock(&nooks_list_lock);

      //
      // Enumerate the nooks, counting the size of each nook
      //

      PDEBUG(" Dumping nook %s\n",nook->name);
      list_for_each(next, &nook_list) {
	nook = list_entry(next,nook_t, next);
	if ((arg & NOOKS_IOCTL_DUMP_DATA_MM) != 0) {
	  nooks_dump_nook_mm(nook);
	}
	if ((arg & NOOKS_IOCTL_DUMP_DATA_DRIVERS) != 0) {
	  nooks_dump_nook_drivers(nook);
	}
	if ((arg & NOOKS_IOCTL_DUMP_DATA_RESOURCES) != 0) {
	  nooks_dump_hash(&nook->resources.resource_list.hash);
	}
	if ((arg & NOOKS_IOCTL_DUMP_DATA_RES_HIST) != 0) {
	  nooks_dump_hash_histogram(&nook->resources.resource_list.hash);
	}


      }
      spin_unlock(&nooks_list_lock);
      break;
    }

  case NOOKS_IOCTL_SET_CSWITCH:
    {

      nooks_context_switch_set_param(arg);

      break;
    }
  case NOOKS_IOCTL_SET_GLOBAL_FLAG:
    {

      nooks_global_flag = arg;

      break;
    }
  case NOOKS_IOCTL_SET_DEBUG_FLAG:
   {
#ifdef NOOKS_DEBUG     
      nooks_debug_level = arg;
#endif

      break;
    }

  case NOOKS_IOCTL_TEST_KMALLOC:
    {
      void * buffer = NULL;
      
      SET_CURRENT_NOOK(nook);
      buffer = nooks_mm_kmalloc(nook, arg, 0);
      if (buffer != NULL) {
	PDEBUG(" alloced %ld bytes\n", arg);
	nooks_mm_kfree(nook, buffer);
      } else {
	PERROR(" failed to alloc %ld bytes\n",arg );
      }
      CLEAR_NOOK();
    }

    break;
  case NOOKS_IOCTL_TEST_VMALLOC:
    {
      void * buffer = NULL;
  PDEBUG(" ioctl areamap empty = %d\n",list_empty(&nook->resources.area_map));      
      buffer = nooks_mm_vmalloc(nook, arg, GFP_KERNEL, PAGE_KERNEL);
      if (buffer != NULL) {
	PDEBUG(" vmalloced %ld bytes\n", arg);
	nooks_mm_vfree(nook, buffer);
      } else {
	PERROR(" failed to vmalloc %ld bytes\n",arg );
      }
    }

    break;

#if 0
  case NOOKS_IOCTL_CALL_AGENT:
    {
      if (nook->recovery_state.info.recover == NULL) {

	/*FIX ME -- Task queues not supported.
	INIT_TQUEUE(&nook->recovery_state.recovery_task, 
		    nooks_fault_alert_usermode, 
		    nook);
	schedule_task(&nook->recovery_state.recovery_task);*/
	printk ("BUG BUG -- NEED TO IMPLEMENT TASK QUEUES - nooks-sys.c");
	result = -EINVAL;
	
      } else {
	result = -EINVAL;
      }
      break;
    }
#endif
#if 0
  case NOOKS_IOCTL_DUMMY_CALL:
    {
    SET_CURRENT_NOOK(nook);
    switch(arg) {
    case 0:
      result = nooks_driver_call(nooks_w_dummy_call0, 0);
      break;
    case 1:
      result = nooks_driver_call(nooks_w_dummy_call1, 1, 1);
      break;
    case 2:
      result = nooks_driver_call(nooks_w_dummy_call2, 2, 1, 2);
      break;
    case 3:
      result = nooks_driver_call(nooks_w_dummy_call3, 3, 1, 2, 3);
      break;
    case 4:
      result = nooks_driver_call(nooks_w_dummy_call4, 4, 1, 2, 3, 4);
      break;
    default:
      break;
    }
    SET_CURRENT_NOOK(NULL);
    break;
    }
#endif
  case NOOKS_IOCTL_DUMMY_IWRAPPER_CALL:
    {
      struct list_head * next;
      pnook_driver_t driver = NULL;

      result=  -ENOENT;
      list_for_each(next, &nook->drivers) {
	driver = list_entry(next,nook_driver_t, next_in_nook);
	if (driver->driver_type == fops_e) {
	  result = ((int (*)(struct file *, int, struct file_lock *))(driver->iwrapper_operations.fops->lock.iwrapper))(NULL, 0, NULL);
	  break;
	}	
      }
      break;
    }
  case NOOKS_IOCTL_FAULT:
    {
      nooks_fault_handler(nook);
      break;
    }
  default:
    result = shadow_ioctl(name, cmd, arg);
    break;
  }

 Cleanup:
  if (name != NULL) {
    nooks_free_user_string(name);
  }
  if (nook != NULL) {
    nooks_dereference_nook(nook);
  }
  
  return result;
}


/**********************************************************
 *
 * sys_nooks_unload_nook
 *
 * Unload a nook - release resources to prepare for deleting
 *
 *********************************************************/
asmlinkage long 
sys_nooks_unload_nook(char * name_user)
{
  char * name = NULL;
  int result = 0;
  pnook_t nook = NULL;
  NOOKS_ENTER();

  if (!capable(CAP_SYS_MODULE)) {
    PERROR(" caller doesn't have permission to delete nooks\n");
    result = -EPERM;
    goto Cleanup;
  }

  //
  // Capture the name of the nook from usermode
  //
  name = nooks_get_nook_name_from_user(name_user);
  if (name == NULL) {
    result = -EINVAL;
    goto Cleanup;
  }

  //
  // Find a nook with this name
  //

  nook = nooks_find_nook(name);
  if (nook == NULL) {
    result = -ENOENT;
    goto Cleanup;
  }

  result = nooks_do_unload_nook(nook);

 Cleanup:
  if (name != NULL) {
    nooks_free_user_string(name);
  }
  if (nook != NULL) {
    nooks_dereference_nook(nook);
  }

  NOOKS_EXIT();
  return(result);
}


void 
nooks_print_string(char * s)
{
  PDEBUG2(s);
}
/**********************************************************
 *
 * nooks_callx
 * 
 * Dummy traps for calling from nooks backinto kernel
 *
 *********************************************************/

asmlinkage long
nooks_call1(int param)
{
  NOOKS_STATS_INC(NOOKS_OPS_TRAP1);
  nooks_print_string("<1> Nooks dummy trap 1\n");
  return(0);
}

asmlinkage long
nooks_call2(int param)
{
  NOOKS_STATS_INC(NOOKS_OPS_ILLEGAL_INST);
  nooks_print_string("<1> Nooks dummy trap 2\n");
  return(0);
}

asmlinkage long
nooks_call3(int param)
{
  nooks_print_string("<1> Nooks dummy trap 3\n");
  return(0);
}

