/*
 * nooks-util.c -- utilities for nooks
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

 
#include "nooks-i.h"
#include "nooks-wrappers.h"
#include <linux/netdevice.h>
#include <linux/percpu.h>


/**********************************************************
 *
 * nooks_get_user_string
 *
 * Copy a string from usermode
 *
 **********************************************************/

int
nooks_get_user_string(const char *user_name, char **buf, int * len)
{
  unsigned long page = 0;
  long namelen;
  int result = 0;
  *len = 0;
  NOOKS_ENTER();

  if (user_name == NULL) {
    result = -EINVAL;
    goto Cleanup;
  }
  page = __get_free_page(GFP_KERNEL);
  if (!page) {
    result = -ENOMEM;
    goto Cleanup;
  }
  namelen = strncpy_from_user((char *)page, user_name, PAGE_SIZE);
  if (namelen > 0) {
    if (namelen < PAGE_SIZE) {

      *buf = (char *)page;
      *len = namelen;
      page = 0;
      goto Cleanup;

    }
    result = -ENAMETOOLONG;
  } else if (!namelen) {
    result = -EINVAL;
  }
  

 Cleanup:
  if (page != 0) {
    free_page(page);
  }

  NOOKS_EXIT();
  return result;
}


/**********************************************************
 *
 * nooks_free_user_string
 *
 * Frees a string captured from usermode
 *
 **********************************************************/

void
nooks_free_user_string(char * kernel_name)
{
  free_page((unsigned long) kernel_name);
}


/**********************************************************
 *
 * nooks_get_driver_string
 *
 * Copy a string from a driver into pool memory.
 *
 * Can only copy strings < 128 characters
 *
 * returns 0 on success
 *
 **********************************************************/

int
nooks_get_driver_string(const char *driver_name, char **buf)
{
  
  long namelen;
  int result = 0;
  char buffer[NOOKS_MAX_DRIVER_STRING];
  NOOKS_ENTER();  

  namelen = __strncpy_from_user(buffer, 
				driver_name, 
				NOOKS_MAX_DRIVER_STRING);
  if (namelen > 0) {
    if (namelen < NOOKS_MAX_DRIVER_STRING) {

      *buf = (char *) nooks_kmalloc(namelen + 1, NOOKS_MEM_FLAGS);
      if (*buf == NULL) {
	result = -ENOMEM;
      } else {
	memcpy(*buf, buffer, namelen+1);
      }
      goto Cleanup;

    }
    result = -ENAMETOOLONG;
  } else if (!namelen) {
    result = -EINVAL;
  }
  

 Cleanup:

  NOOKS_EXIT();
  return result;
}


/**********************************************************
 *
 * nooks_free_driver_string
 *
 * Frees a string captured from a driver
 *
 **********************************************************/

void
nooks_free_driver_string(char * kernel_name)
{
  nooks_kfree( kernel_name);
}


/**********************************************************
 *
 * nooks_memcpy
 *
 * Copy memory from nook into kernel safely
 *
 *********************************************************/

int
nooks_memcpy(void * out, void * in, size_t len)
{
  pnook_t nook = GET_CURRENT_NOOK();
  int result = 0;
  NOOKS_ENTER();
  //if ((nooks_mm_get_pte_flags(nook, in) & _PAGE_PRESENT) == 0) {
  //  return(-EFAULT);
  //}

  if (copy_from_user(out, in, (unsigned long) len)) {
    result = -EFAULT;
  }
  NOOKS_EXIT();
  return(result);
}
/**********************************************************
 *
 * nooks_memcpy_out
 *
 * Copy memory from kernel into nook safely (while executing in kernel)
 *
 *********************************************************/

int
nooks_memcpy_out(void * out, void * in, size_t len)
{
  pnook_t nook = GET_CURRENT_NOOK();
  int result = 0;
  NOOKS_ENTER();
  
  printk ("In nooks_memcpy_out.\n");
  //if ((nooks_mm_get_pte_flags(nook, out) & _PAGE_RW) == 0) {
  //  return(-EFAULT);
  //}
  if (copy_to_user(out, in, (unsigned long) len)) {
    result = -EFAULT;
  }
  NOOKS_EXIT();
  return(result);
}

/**********************************************************
 *
 * nooks_strdup
 *
 * Duplicates a string (that had been verified to terminate)
 *
 **********************************************************/

char *
nooks_strdup(const char * kernel_name, int flags)
{
  char * s;
  int len = strlen(kernel_name);
  s = (char *) nooks_kmalloc(len+1, flags);
  if (s != NULL) {
    memcpy(s, kernel_name, len+1);
  }
  return(s);

}

/**********************************************************
 * 
 * nooks_enter_nook
 *
 * used by kernel code to indicate that it is starting to do
 * work on behalf of a particular nook
 *
 *********************************************************/

void 
nooks_enter_nook(struct nook * nook)
{
  SET_CURRENT_NOOK(nook);
}

/**********************************************************
 * 
 * nooks_exit_nook
 *
 * used by kernel code to indicate that it is starting to do
 * work on behalf of a particular nook
 *
 *********************************************************/

void 
nooks_exit_nook(void)
{
  CLEAR_NOOK();
}


/**********************************************************
 *
 * nooks_copy_softnet_in
 *
 * copy softnet data into the kenel
 *
 *********************************************************/

int 
nooks_copy_softnet_in(void)
{
  int result = 0;
  struct softnet_data data;
  int processor_id;
  pnook_t nook;
  struct softnet_data my_softnet_data;
  NOOKS_ENTER();
  nook = GET_CURRENT_NOOK();

  processor_id = smp_processor_id();
  my_softnet_data = per_cpu(softnet_data, processor_id); 
  
  // 
  // if nothing changed, bail out early
  //
  // FIXME -- Changes made here.. softnet_data to nooks_w_softnet_data.
  if (!test_and_clear_bit(0, &my_softnet_data.modified)) {
    goto Cleanup;
  }

  //
  // Copy the softnet data
  //
  result = nooks_memcpy(&data, 
			&nooks_w_softnet_data, 
			sizeof(struct softnet_data));


  //
  // Copy the queue to the kernel
  //
  
  while (data.output_queue != NULL) {
    pnook_driver_t driver = NULL;
    struct net_device * kernel_dev;
    result = nooks_find_resource_elem(nook, netdev_nook_e, data.output_queue, (void **) &driver);
    if (result < 0) {
      PERROR("Failed to find netdevice on softenet queue\n");
      goto Cleanup;
    }
    kernel_dev = (struct net_device *) driver->data;
    result = nooks_memcpy(&kernel_dev->state, 
			  &data.output_queue->state, 
			  sizeof(unsigned long));
    if (result < 0) {
      goto Cleanup;
    }
    
    /*FIXME */kernel_dev->next_sched = my_softnet_data.output_queue;
    my_softnet_data.output_queue = kernel_dev;

    //
    // Get the next element in the list
    //

    result = nooks_memcpy(&data.output_queue, 
			  &data.output_queue->next_sched, 
			  sizeof(struct netdevice *));
    if (result < 0) {
      PERROR("Failed to copy next_sched pointer\n");
      goto Cleanup;
    }
     
  }

  //
  // BUGBUG: need to track safe SKBs - need to copy the queue
  //
  //
  // BUGBUG: need to find end of list.
  //
  {
    struct sk_buff * queue = data.completion_queue;
    if  (queue != NULL) {
      while (queue->next != NULL) {
	queue = queue->next;
      }
      // point the tail of the nooks list at the kernel list
      queue->next = my_softnet_data.completion_queue;
      // point the kernel list at the head of the nook list
      my_softnet_data.completion_queue = data.completion_queue;
    }
  }

  // BUGBUG: need to make safe

  nooks_w_softnet_data->output_queue = NULL;
  nooks_w_softnet_data->completion_queue = NULL;

 Cleanup:
  NOOKS_EXIT();
  return(result);
}




/**********************************************************
 * nooks_copy_softnet_out
 * 
 * Copies softnet data out of the kernel into the nook
 * 
 *
 *********************************************************/

void
nooks_copy_softnet_out(void)
{
  // Do nothing - we do everything in copy_softnet_in
  //  int processor_id;

  //  processor_id = smp_processor_id();
  //  nooks_w_softnet_data[processor_id].output_queue = NULL;
  //  nooks_w_softnet_data[processor_id].completion_queue = NULL;

}

/**********************************************************
 *
 * nooks_log2pages
 *
 * returns ceiling(log2(1 + num/PAGE_SIZE))
 *
 *********************************************************/

unsigned int
nooks_log2pages(unsigned int num)
{
  unsigned int s = PAGE_SIZE, p = 0;
  while (s < num) {
    s <<= 1;
    p++;
  }
  return (p);
}

int 
nooks_list_del(struct list_head * entry)
{
  int result;
  struct list_head elem;
  result = nooks_memcpy(&elem, entry, sizeof(struct list_head));
  if (result >= 0) {
    result = nooks_memcpy_out(&elem.next->prev, 
			      &elem.prev, 
			      sizeof(struct list_head *));
  }
  if (result >= 0) {
    result = nooks_memcpy_out(&elem.prev->next, 
			      &elem.next, 
			      sizeof(struct list_head *));
  }
  return(result);
}

int
nooks_list_add(struct list_head * new,
	       struct list_head * head)
{
  int result;
  struct list_head *next;
  result = nooks_memcpy(&next, &head->next, sizeof(struct list_head *));
  if (result >= 0) {
    result = nooks_memcpy_out(&next->prev, &new, sizeof(struct list_head *));
  }
  if (result >= 0) {
    result = nooks_memcpy_out(&new->next, &next, sizeof(struct list_head *));
  }
  if (result >= 0) {
    result = nooks_memcpy_out(&new->prev, &head, sizeof(struct list_head *));
  }

  if (result >= 0) {
    result = nooks_memcpy_out(&head->next, &new, sizeof(struct list_head *));
  }
  return(result);
}

#ifdef CONFIG_NOOKS_SHADOW
/*
 * nooks_is_shadow : Does the nook has shadow ops ?
 */
int nooks_is_shadow(struct nook *n)
{
  return (n->shadow != NULL);
}
#endif
