/*
 * nooks-work.c -- Routines for managing deferred work lists
 *
 * Copyright (C) 2001 Mike Swift
 *
 * The source code in this file can be freely used, adapted,
 * and redistributed in source or binary form, so long as an
 * acknowledgment appears in derived source files.  The citation
 * should list that the code comes from the book "Linux Device
 * Drivers" by Alessandro Rubini and Jonathan Corbet, published
 * by O'Reilly & Associates.   No warranty is attached;
 * we cannot take responsibility for errors or fitness for use.
 *
 */

 
#include "nooks-i.h"
#include "nooks-iwrappers.h"
#include "nooks-wrappers.h"
#include "nooks-work.h"
#include "nooks-workers.h"
#include "nooks-switch.h"

spinlock_t nook_work_list_lock;

/**********************************************************
 *
 * nooks_process_work_list
 *
 * Execute all the work items on a list for the given
 * execution environment
 *
 * executes in kernel environment, on exit from a nook
 *
 *********************************************************/

int 
nooks_process_work_list(pnook_t nook, int callout)
{
  struct list_head next;
  struct list_head * spare = NULL;
  pnooks_work_item_t item = NULL;
  nooks_work_item_t kern_item;
  int result = 0;
  struct list_head * list;
  struct list_head kern_list;

  NOOKS_ENTER();

  if (!NOOK_OK(nook)) {
    goto Cleanup;
  }

  
  list = &(GET_NOOK_STATE().thread_state.nook_frame->work_list);
  result = nooks_memcpy(&kern_list, list, sizeof(struct list_head));
  
  if (result < 0) {
    goto Cleanup;
  }
  
  if ((kern_list.next != NULL) && (kern_list.next != list)) {
    
    next = kern_list;
    spare = kern_list.next;
    do {
      result = nooks_memcpy(&next, next.next, sizeof(struct list_head *));
      if (result < 0) {
	goto Cleanup;
      }
      item = list_entry(spare, nooks_work_item_t, next);
      result = nooks_memcpy(&kern_item, item, sizeof(nooks_work_item_t));
      if (result < 0) {
	goto Cleanup;
      }
      
      if (((kern_item.flags & NOOKS_WORK_FLAG_RETURN) == 0) || 
	  (callout == SWITCH_OUT)) {
	result = nooks_list_del(&item->next);
	if (result < 0) {
	  goto Cleanup;
	}

	//
	// BUGBUG: verify item handler
	//

	kern_item.result = kern_item.handler(kern_item.data1, 
					     kern_item.data2, 
					     kern_item.data3);

	result = nooks_list_add(&item->next, 
				&nook->locals->free_work_list);
	if (result < 0) {
	  goto Cleanup;
	}
      }
      spare = next.next;
    } while (spare != list);
  }
 Cleanup:
  NOOKS_EXIT();
  return(result);
}

/**********************************************************
 *
 * nooks_process_entry_exit_list
 *
 * Executes handler from kernel routines that need to
 * run in the nooks environment.
 *
 * Runs in NOOK environment
 *
 *********************************************************/

int 
nooks_process_entry_exit_list(pnook_t nook, int flags)
{
  struct list_head * next = NULL;
  struct list_head * spare = NULL;
  pnooks_work_item_t item = NULL;
  int result = 0;
  struct list_head * list;

  NOOKS_ENTER();

  if ((nook->resources.kernel_work_list.flags & flags) != flags) {
    goto Cleanup;
  }

  if (nook->resources.kernel_work_list.count == 0) {
    goto Cleanup;
  }

  if (!NOOK_OK(nook)) {
    goto Cleanup;
  }

  list = &nook->resources.kernel_work_list.list;
  
  list_for_each_safe(next, spare, list) {
    item = list_entry(next,nooks_work_item_t, next);
    if ((item->flags & flags) == flags) {
      result = item->handler(item->data1, item->data2, item->data3);
      if (result < 0) {
	PDEBUG("Failed to execute  item handler %p: %d\n", item->handler, result);
      }
    }
  }

 Cleanup:
  NOOKS_EXIT();
  return(result);
  
}


pnooks_work_item_t 
nooks_add_work_item(pnook_t nook, 
	   	    void * data1, 
	   	    void * data2, 
	   	    void * data3, 
	            nooks_work_handler_t * handler)
{
  pnooks_work_item_t item = NULL;
  struct list_head * list = NULL;
  NOOKS_ENTER();

  list = &(GET_NOOK_STATE().thread_state.nook_frame->work_list);
  if (list->next == NULL) {
    INIT_LIST_HEAD(list);
  }

  item = (pnooks_work_item_t) nooks_mm_lookaside_alloc(&nook->locals->work_pool);
  if (item == NULL) {
    goto Cleanup;
  }
  item->handler = handler;
  item->data1 = data1;
  item->data2 = data2;
  item->data3 = data3;
  item->flags = NOOKS_WORK_FLAG_ONE_SHOT | 
    NOOKS_WORK_FLAG_KERNEL | 
    NOOKS_WORK_FLAG_EXIT;

  // 
  // Add to tail, to preserve ordering
  //

  list_add_tail(&item->next, list);

 Cleanup:
  NOOKS_EXIT();
  return(item);
}

pnooks_work_item_t 
nooks_add_kernel_work_item(pnook_t nook, 
			    void * data1, 
			    void * data2, 
			    void * data3, 
			    nooks_work_handler_t * handler,
			    int flags)
{
  pnooks_work_item_t item = NULL;
  struct list_head * list = NULL;
  NOOKS_ENTER();

  list = &nook->resources.kernel_work_list.list;

  item = (pnooks_work_item_t) nooks_mm_lookaside_alloc(&nook->resources.work_pool);
  if (item == NULL) {
    goto Cleanup;
  }
  item->handler = handler;
  item->data1 = data1;
  item->data2 = data2;
  item->data3 = data3;
  item->flags = flags;

  // 
  // Add to tail, to preserve ordering
  //

  list_add_tail(&item->next, list);
  nook->resources.kernel_work_list.count++;
  nook->resources.kernel_work_list.flags |= flags;

 Cleanup:
  NOOKS_EXIT();
  return(item);
}

void 
nooks_free_work_list(pnook_t nook)
{
  pnooks_work_item_t item = NULL;
  struct list_head * next = NULL;
  struct list_head * spare = NULL;
  NOOKS_ENTER();
  if (!in_irq()) {
    list_for_each_safe(next, spare, &nook->locals->free_work_list) {
      item = list_entry(next,nooks_work_item_t, next);
      
      list_del_init(&item->next);
      
      nooks_mm_lookaside_free(&nook->locals->work_pool, item);
    }
  }
  NOOKS_EXIT();
}

void
nooks_clean_kernel_work_list(pnook_t nook)
{
  pnooks_work_item_t item = NULL;
  struct list_head * next = NULL;
  struct list_head * spare = NULL;
  NOOKS_ENTER();

  list_for_each_safe(next, spare, &nook->resources.kernel_work_list.list) {
    item = list_entry(next,nooks_work_item_t, next);
    if (item->flags & NOOKS_WORK_FLAG_ONE_SHOT) {
      list_del_init(&item->next);
      nook->resources.kernel_work_list.count--;
      nooks_mm_lookaside_free(&nook->resources.work_pool, item);
    }
  }
  NOOKS_EXIT();
}


void 
nooks_delete_kernel_work_item(pnook_t nook, pnooks_work_item_t item)
{
  NOOKS_ENTER();
  list_del_init(&item->next);
  nooks_mm_lookaside_free(&nook->resources.work_pool, item);
  NOOKS_EXIT();
}

void
nooks_delete_work_list(pnook_t nook)
{
  pnooks_work_item_t item = NULL;
  struct list_head * next = NULL;
  struct list_head * spare = NULL;
  NOOKS_ENTER();

  list_for_each_safe(next, spare, &nook->resources.kernel_work_list.list) {
    item = list_entry(next,nooks_work_item_t, next);
    list_del_init(&item->next);
    nook->resources.kernel_work_list.count--;
    nooks_mm_lookaside_free(&nook->resources.work_pool, item);
  }
  NOOKS_EXIT();
}

