/*
 * net-shadow.c --  Shadow driver for network devices
 *
 * Copyright (C) 2003 Mike Swift
 */

#include "netshadow-i.h"
#ifdef CONFIG_NOOKS_PROC
#include "shadow-proc.h"
#endif
#include <../nooks/nooks-wrappers.h>


MODULE_AUTHOR("Mike Swift");
MODULE_LICENSE("GPL");



//MODULE_PARM(debug, "i");

LIST_HEAD(shadow_list);

extern sknetops_operations_t shadow_kernel_ops;

/**********************************************************
 * shadow_log
 * 
 * Log calls into per-shadow log
 * 
 *
 *********************************************************/

unsigned int shadow_max_log_entries = 500;

spinlock_t shadow_log_lock;

int shadow_verbose_log = 0;

#ifdef NETSHADOW_DEBUG
void
shadow_log(pnook_driver_t driver, char * routine, int result)
{
  pshadow_driver_t device = (pshadow_driver_t) driver->nook->shadow;
  pshadow_log_entry_t entry;
  unsigned long flags;

  if (device == NULL) {
    return;
  }

  if (shadow_verbose_log) {
    printk("Shadow log: %s() returns %x\n",
	   routine, result);
  }
  entry = (pshadow_log_entry_t) kmalloc(sizeof(shadow_log_entry_t),
					NOOKS_MEM_FLAGS);
  if (entry != NULL) {
    struct list_head * last = NULL;
    entry->name = routine;
    entry->result = result;
    entry->timestamp = jiffies;
    spin_lock_irqsave(&shadow_log_lock, flags);
    list_add_tail(&entry->next, &device->log.log );
    device->log.entries++;
    if (device->log.entries > shadow_max_log_entries) {
      last = list_pop(&device->log.log);
    }
    spin_unlock_irqrestore(&shadow_log_lock, flags);
    if (last != NULL) {
      entry = list_entry(last, shadow_log_entry_t, next);
      device->log.entries--;
      kfree(last);
    }
  }
}
#endif // netshadow_debug

/**********************************************************
 *
 * shadow_print_log
 *
 * prints entries in log to syslog
 *
 *********************************************************/

void
shadow_print_log(pshadow_driver_t shadow_dev)
{
  struct list_head * next;
  pshadow_log_entry_t entry;

  list_for_each(next, &shadow_dev->log.log) {
    entry = list_entry(next, shadow_log_entry_t, next);

    printk("%ld: %s(%d)\n",entry->timestamp, entry->name, entry->result);
  }  
}


/**********************************************************
 *
 * shadow_free_log
 *
 * frees all entries in log
 *
 *********************************************************/

void
shadow_free_log(pshadow_driver_t shadow_dev)
{
  pshadow_log_entry_t entry;
  struct list_head *next, *spare;
  
  list_for_each_safe(next, spare, &shadow_dev->log.log) {
    entry = list_entry(next, shadow_log_entry_t, next);
    kfree(entry);
  }  

}

/**********************************************************
 *
 * shadow_free_driver
 *
 * delete the shadow driver for a nook
 *
 *********************************************************/

void shadow_free_driver(pshadow_driver_t driver)
{
  if (driver != NULL) {
    list_del(&driver->next);
    nooks_destroy_hash_table(&driver->kernel_objects);
    nooks_destroy_hash_table(&driver->kernel_drivers);
    shadow_free_recovery_block(driver->nook);
    kfree(driver);
  }
}

/**********************************************************
 *
 * shadow_create_driver
 *
 * create a shadow driver for a nook
 *
 *********************************************************/

int  shadow_create_driver(pnook_t nook, nooks_recovery_fn * recovery_fn) 
{
  int result = 0;
  pshadow_driver_t shadow = NULL;

  printk ("Creating shadow driver for nook %s.\n", nook->name);
  shadow = (pshadow_driver_t) kmalloc(sizeof(shadow_driver_t), GFP_KERNEL);
  if (shadow == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }
  memset(shadow, 0, sizeof(shadow_driver_t));
  INIT_LIST_HEAD(&shadow->next);
  INIT_LIST_HEAD(&shadow->drivers);
  INIT_LIST_HEAD(&shadow->log.log);
  init_waitqueue_head(&shadow->wait_for_recovery);
  init_MUTEX(&shadow->recovery_lock);

  result = shadow_build_recovery_block(nook, recovery_fn);
  if (result < 0) {
    goto Cleanup;
  }
  result = nooks_create_hash_table(&nook->resources.hash_pool,
				   &shadow->kernel_objects, 
				   64,
				   0); // not a dual
  
  if (result < 0) {
    goto Cleanup;
  }

  result = nooks_create_hash_table(&nook->resources.hash_pool,
				   &shadow->kernel_drivers, 
				   4,
				   0); // not a dual
  
  if (result < 0) {
    goto Cleanup;
  }
    
  
  //shadow->state = shadow_state_running;
  list_add(&shadow->next, &shadow_list);
  nook->shadow = shadow;
  shadow->nook = nook;

 Cleanup:
  if (result < 0) {
    if (shadow != NULL) {
      shadow_free_driver(shadow);
    }
  }
  return(result);
}


/**********************************************************
 *
 * shadow_del_interface
 *
 * stop interposing on a driver interface
 *
 *********************************************************/


void shadow_del_interface(pnook_driver_t driver)
{
  if (driver->shadow_operations != NULL) {
    kfree(driver->shadow_operations);
  }
  if (driver->shadow_kernel_operations != NULL) {
    kfree(driver->shadow_kernel_operations);
  }
}

/**********************************************************
 *
 * shadow_add_interface
 *
 * start interposing on an interface
 *
 *********************************************************/

int shadow_add_interface(pnook_driver_t driver)
{
  pnook_t current_nook;
  int result = 0;

  current_nook = driver->nook;
  if (current_nook->shadow == NULL) {
    result = -EINVAL;
    goto Cleanup;
  }


  driver->shadow_operations = (pnooks_shadow_operations_t) 
    kmalloc(sizeof(nooks_shadow_operations_t), NOOKS_MEM_FLAGS);
  if (driver->shadow_operations == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }
  driver->shadow_kernel_operations = (pnooks_shadow_kernel_operations_t) 
    kmalloc(sizeof(nooks_shadow_kernel_operations_t), NOOKS_MEM_FLAGS);
  if (driver->shadow_kernel_operations == NULL) {
    result = -ENOMEM;
    goto Cleanup;
  }
    

   driver->shadow_state = shadow_state_running;

 Cleanup:
  if (result < 0) {
    if (driver->shadow_operations != NULL) {
      kfree(driver->shadow_operations);
    }
    if (driver->shadow_kernel_operations != NULL) {
      kfree(driver->shadow_kernel_operations);
    }
  }
  return(result);
}


int
shadow_add_driver(pnook_t nook,
		  void * key, 
		  pnook_driver_t driver )
{
  int result;
  if (nook->shadow == NULL) {
    return(-ENOENT);
  }

  result = nooks_add_to_hash(&nook->shadow->kernel_drivers,
			     (nooks_resource_type_t) driver->driver_type,
			     key,
			     driver );

  printk ("Adding shadow of type %d and result %d and key %d.\n", driver->driver_type, result, (int)key);
  if (result >= 0) {
    printk ("Deleting driver from list of type %d.\n",driver->driver_type);
    list_del(&driver->next_in_nook);
    list_add(&driver->next_in_shadow, &nook->shadow->drivers);
  }
  return(result);
}



/**********************************************************
 *
 * shadow_find_driver
 *
 * Checks to see if a driver for an existing device already
 * exists
 *
 *********************************************************/
pnook_driver_t 
shadow_find_driver(pnook_t nook, 
		   nooks_iwrapper_type_t driver_type,
		   void * key)
{
  int result;
  pnook_driver_t driver = NULL;

  if (nook->shadow == NULL) {
    return(NULL);
  }

  result = nooks_find_in_hash(&nook->shadow->kernel_drivers,
			       (nooks_resource_type_t) driver_type,
			       key,
			       (void **) &driver );

  return(driver);


}

/**********************************************************
 *
 * shadow_find_driver2
 *
 * Checks to see if a driver for an existing device already
 * exists
 *
 *********************************************************/
pnook_driver_t 
shadow_find_driver2(pnook_t nook, 
		   nooks_iwrapper_type_t driver_type,
		   unsigned int key,
		   char *key2)
{
  struct list_head *next;
  pnook_driver_t driver = NULL;

  if (nook->shadow == NULL) {
    return(NULL);
  }
#define SHADOW_FIND_DRIVER(t) \
  list_for_each(next, &(nook->shadow->drivers)) { \
	  driver = list_entry(next, nook_driver_t, next_in_shadow); \
	  if ((t) && (driver->driver_type == driver_type) \
			  && !strcmp(driver->name, key2)) \
		  return driver; \
  } \
  break

  switch(driver_type)
  {
	case mops_e:   SHADOW_FIND_DRIVER(1);
	case bdops_e:
	case fops_e:
	               SHADOW_FIND_DRIVER(((key && (driver->major == key))
					  || (!key && driver->major)));
	case irqops_e: SHADOW_FIND_DRIVER(driver->major == 
				       		NOOKS_IRQ_TO_MAJOR(key));
	case pciops_e:
	case netops_e:
	default: break;
  }
#undef SHADOW_FIND_DRIVER
  return NULL;


}

/**********************************************************
 *
 * shadow_get_driver
 *
 * Checks to see if a driver for an existing device already
 * exists, and remove it from list
 *
 *********************************************************/
pnook_driver_t 
shadow_get_driver(pnook_t nook, 
		   nooks_iwrapper_type_t driver_type,
		   void * key)
{
  int result;
  pnook_driver_t driver = NULL;

  if (nook->shadow == NULL) {
    return(NULL);
  }


  result = nooks_del_from_hash(&nook->shadow->kernel_drivers,
			       (nooks_resource_type_t) driver_type,
			       key,
			       (void **) &driver );
  if (result >= 0) {
    //
    // Move driver back to nook
    //
    //     list_del(&driver->next_in_shadow);
    list_add(&driver->next_in_nook, &nook->drivers);
    
  }
  return(driver);


}

int s_nooks_send_fake_arp(struct net_device *dev);

/**********************************************************
 * 
 * shadow_ioctl
 *
 * control function for shadows
 *
 *********************************************************/

long
shadow_ioctl(char * name, unsigned int cmd, unsigned long arg)
{
  extern int shadow_migrated;
  extern int shadow_inject_fault;
  long result = 0;
  printk("In shadow_ioctl with args %s %d %d.\n", name, cmd, arg);

  /*
  if ((shadow_migrated == 0) && ((cmd == 10000) || (cmd == 13000)))	{
	printk ("No migration happened. returning.");
	return 0;
  }
  */ 
  switch(cmd) {
  case NETSHADOW_IOCTL_DUMP_LOG: {
    struct list_head *next;
    pshadow_driver_t shadow_dev;
    list_for_each(next, &shadow_list) {      
      shadow_dev = list_entry(next, shadow_driver_t, next);
      shadow_print_log(shadow_dev);
    }     

    break;
  }
  case NETSHADOW_IOCTL_CLEAR_LOG: {
    struct list_head *next;
    pshadow_driver_t shadow_dev;
    
    list_for_each(next, &shadow_list) {      
      shadow_dev = list_entry(next, shadow_driver_t, next);
      shadow_free_log(shadow_dev);      
    }    
    break;
  }
  case NETSHADOW_IOCTL_UNLOAD_NOOK: {
    pnook_t nook = NULL;
    nook = nooks_find_nook(name);
    if (nook == NULL) {
      result = -ENOENT;
      break;
    }
    ENTER_SHADOW_RECOVERY();
    result = shadow_unload_nook(nook);
    EXIT_SHADOW_RECOVERY();
    nooks_dereference_nook(nook);
    break;
  }
  case NETSHADOW_IOCTL_START_RECOVERY: {
    pnook_t nook = NULL;
    nook = nooks_find_nook(name);
    if (nook == NULL) {
      result = -ENOENT;
      break;
    }
    printk ("Calling shadow recovery inner function.\n");
    ENTER_SHADOW_RECOVERY();
    result = shadow_start_recovery(nook);
    EXIT_SHADOW_RECOVERY();
    nooks_dereference_nook(nook);
    break;
  }

  case NETSHADOW_IOCTL_FINISH_RECOVERY: {
    pnook_t nook = NULL;
    nook = nooks_find_nook(name);
    if (nook == NULL) {
      result = -ENOENT;
      break;
    }
    ENTER_SHADOW_RECOVERY();
    result = shadow_finish_recovery(nook);
    EXIT_SHADOW_RECOVERY();
    nooks_dereference_nook(nook);
    break;
  }

  case NETSHADOW_IOCTL_NETDEV_RECOVERY: {
    pnook_t nook = NULL;
    nook = nooks_find_nook(name);
    if (nook == NULL) {
      result = -ENOENT;
      break;
    }
    ENTER_SHADOW_RECOVERY();
    result = shadow_netdev_recovery(nook);
    EXIT_SHADOW_RECOVERY();
    nooks_dereference_nook(nook);
    break;
  }

  case NETSHADOW_IOCTL_SOUND_RECOVERY: {
    pnook_t nook = NULL;
    nook = nooks_find_nook(name);
    if (nook == NULL) {
      result = -ENOENT;
      break;
    }
    ENTER_SHADOW_RECOVERY();
    result = shadow_sound_recovery(nook);
    EXIT_SHADOW_RECOVERY();
    nooks_dereference_nook(nook);
    break;
  }

  case NETSHADOW_IOCTL_TRIGGER_RECOVERY: {
    pnook_t nook = NULL;
    printk("TRIGGER RECOVERY\n");
    nook = nooks_find_nook(name);
    if (nook == NULL) {
      printk ("Nook not found.\n");
      result = -ENOENT;
      break;
    }
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
    nook->waiting_time = arg;
#endif
    if (nook->recovery_state.info.recover != NULL) {
      printk ("Triggering recovery.\n");
      result = nook->recovery_state.info.recover(nook);
    } else {
      printk ("Recovery state not set.\n");
      result = -EINVAL;
    }
    nooks_dereference_nook(nook);
    break;
  }
  case NETSHADOW_IOCTL_SET_SOUND_DRIVER: 
  case NETSHADOW_IOCTL_SET_BLOCK_DRIVER:
  case NETSHADOW_IOCTL_SET_NETDEV_DRIVER: {
    pnook_t nook = NULL;
    nooks_recovery_fn * recovery_fn = NULL;
    nook = nooks_find_nook(name);
    if (nook == NULL) {
      result = -ENOENT;
      break;
    }
    switch(cmd) {
    case NETSHADOW_IOCTL_SET_SOUND_DRIVER:
      recovery_fn =shadow_recover_sound_nook;
#ifdef CONFIG_NOOKS_PROC
      nook->shadow_driver_name = "sound";
#endif
      break;
    case NETSHADOW_IOCTL_SET_NETDEV_DRIVER:
      recovery_fn =shadow_recover_netdev_nook;
#ifdef CONFIG_NOOKS_PROC
      nook->shadow_driver_name = "netdev";
#endif
      break;
/*    case NETSHADOW_IOCTL_SET_BLOCK_DRIVER:
     recovery_fn =shadow_recover_blkdev_nook;

#ifdef CONFIG_NOOKS_PROC
      nook->shadow_driver_name = "blkdev";
#endif
      break;
*/
    }
#ifdef CONFIG_NOOKS_PROC
    shadow_proc_add_entry(nook);
#endif
    printk ("Creating shadow driver for nook -> %s for shadow of type -> %s.\n",nook->name, nook->shadow_driver_name);
    result = shadow_create_driver(nook,recovery_fn);
    nooks_dereference_nook(nook);
    break;
  }

  case NETSHADOW_IOCTL_PRINT_DRIVERS: {
    pnook_t nook = NULL;
    struct list_head * next;
    pshadow_driver_t shadow;
    pnook_driver_t driver = NULL;
    nook = nooks_find_nook(name);
    if (nook == NULL) {
      result = -ENOENT;
      break;
    }
    shadow = nook->shadow;
    list_for_each(next, &shadow->drivers) {
      driver = list_entry(next,nook_driver_t, next_in_shadow);
      printk("driver name: %s, type = %d\n", driver->name, driver->driver_type);
    }
    nooks_dereference_nook(nook);
    break;
  }
  case NETSHADOW_IOCTL_DUMP_IOCTLS: {
    pnook_t nook = NULL;
    nook = nooks_find_nook(name);
    if (nook == NULL) {
      result = -ENOENT;
      break;
    }
    result = shadow_dump_ioctls(nook);
    nooks_dereference_nook(nook);
    break;
  }
 
 case NETSHADOW_INJECT_FAULT:   {
    pnook_t nook = NULL;
    nook = nooks_find_nook(name);
    printk ("Activating fault injection");
    shadow_inject_fault = 1;
    break;
 }
 
 case NETSHADOW_SEND_ARP:	{
  struct list_head * lnext;
  struct list_head * lspare;
  pnook_t nook = NULL;
  nook = nooks_find_nook(name);
  pnook_driver_t driver = NULL;
  struct net_device * net_dev = NULL;

  //printk("***Sending netdev ARP***\n");
  list_for_each_safe(lnext, lspare, &nook->drivers) {
    driver = list_entry(lnext,nook_driver_t, next_in_nook);
    if(driver)
    {
        if (driver->driver_type == netops_e)
                if(driver && driver->name && driver->major)	{
                        //printk("Name: %s, Major: %d, Type: %d\n", driver->name,
                        //                                driver->major, driver->driver_type);
			net_dev = (struct net_device *) driver->data;
		}
                else
                        printk("Arbit driver\n");
    }
    else
    {
        printk("NO DRIVER!\n");
    }

   }
  
   if (net_dev != NULL)
   {
     result = s_nooks_send_fake_arp(net_dev);
     printk("ARP sent as %d.\n", result);
   }
   if (result == 0)
	shadow_migrated = 0;

   break; 
  }

  default: 
    result = -EINVAL;
  }

  //shadow_migrated = 0;
  //printk ("Returning with result %d.\n", result);
  return(result);
  
}

#define ARPOP_REPLY 2

int s_nooks_send_fake_arp(struct net_device *dev)
{
	int result = 0; 
#ifdef CONFIG_INET
        struct sk_buff *skb;
        u32             src_ip, dst_ip;

        dst_ip = INADDR_BROADCAST;
        src_ip = inet_select_addr(dev, dst_ip, RT_SCOPE_LINK);

        /* No IP? Then nothing to do. */
        if (src_ip == 0)        {
                return;
        }

       skb = arp_create(ARPOP_REPLY, ETH_P_ARP,
                 dst_ip, dev, src_ip,
                /*dst_hw*/ NULL, /*src_hw*/ NULL,
                /*target_hw*/ dev->dev_addr);
                if (skb == NULL)
                        return;
       result = dev_queue_xmit(skb);
#endif
       return result;
}



int shadow_init(struct net_device *dev)
{
  
  SET_MODULE_OWNER(dev);
  spin_lock_init(&shadow_log_lock);
  return 0;
}


int shadow_init_module(void)
{
  return(0);
}

void shadow_cleanup(void)
{
    printk("<1> calling netshadow_cleanup\n");
    return;
}


module_init(shadow_init_module);
module_exit(shadow_cleanup);

