/*
 * shadow-proc.c - functions for some stats in the procfs
 *
 * 2004 - Damien Martin-Guillerez
 * work under Mike Swift & Hank Levy
 */

#include <../nooks/nooks-i.h>
#include "shadow-proc.h"

struct proc_dir_entry *proc_nooks;
struct proc_dir_entry *proc_nooks_info;
extern struct module kernel_module;

int shadow_proc_init(void)
{
	proc_nooks = proc_mkdir("nooks", 0);
	if (!proc_nooks) {
		printk (KERN_ERR "cannot init /proc/nooks\n");
		return -ENOMEM;
	}
	proc_nooks_info = create_proc_info_entry ("nooks", S_IRUGO, proc_nooks,
			shadow_proc_info);
	if (!proc_nooks_info) {
		printk (KERN_ERR "cannot init /proc/nooks/nooks\n");
		return -ENOMEM;
	}
	return 0;
}

void shadow_proc_clean(void)
{
	pnook_t nook = NULL;
	struct list_head * next = NULL;
	list_for_each(next, &nook_list) {
		nook = list_entry(next, nook_t, next);
		if (nook->proc_ent)
			__remove_proc_entry(nook->proc_ent);
		nook->proc_ent = NULL;
	}
	__remove_proc_entry (proc_nooks);
	__remove_proc_entry (proc_nooks_info);
	proc_nooks = proc_nooks_info = NULL;
}

int shadow_proc_add_entry(pnook_t nook)
{
	if (nook->proc_ent)
		return -EEXIST;
	nook->last_fault_time.tv_sec = 0;
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
	nook->last_replug_time.tv_sec = 0;
#endif
	nook->proc_ent = create_proc_read_entry(nook->name, S_IRUGO, proc_nooks,
						shadow_proc_read_file, 
						(void*)nook);
	if (!(nook->proc_ent)) {
		printk (KERN_ERR "cannot init /proc/nooks/%s\n", nook->name);
		return -ENOMEM;
	}
	return 0;
}

void shadow_proc_del_entry(pnook_t nook)
{
	if (nook->proc_ent) {
		__remove_proc_entry(nook->proc_ent);
		nook->proc_ent = NULL;
	}
}

int shadow_proc_info(char *buf, char **start, off_t offset, int length)
{
	pnook_t nook = NULL;
	struct list_head * next = NULL;
	int s,n;
	int size = 0, len = 0;
	off_t pos = 0;
	char buffer[1024];
	char tabs[8];
	
#define SHADOW_PROC_HEADER	"nook\t\t\tshadowed\tstate\n"
#define shadow_strncpy(dst,src,x,y) (strncpy(dst,src, ((x) < (y)) ? (x) : (y)),  ((x) < (y)) ? (x) : (y))

	len = sprintf (buffer, SHADOW_PROC_HEADER);
	if (offset < len)
		size += shadow_strncpy(buf, buffer+offset, len-offset, length);
  	pos += len;
	offset -= len;
	if (length <= size)
		return size;

	buffer[1023] = 0;
	list_for_each(next, &nook_list) {
		nook = list_entry(next, nook_t, next);
		s = strlen(nook->name);
		n = 4 - (s / 8) ;
		if(n > 0) {
			tabs[--n] = 0;
			while(n--) tabs[n] = '\t';
		}
		else {
			tabs[0] = ' ';
			tabs[1] = 0;
		}
		len = snprintf(buffer, 1023, "%s%s%s\t\t%s\n", 
					nook->name,
					tabs,
					nook->shadow_driver_name,
					nook_state_to_string(nook->state));
		if (offset < len) {
			size += shadow_strncpy(buf+size,
					(offset > 0) ? (buffer+offset) : buffer,
					len - offset, length-size);
			if (offset > 0)
				offset -= len;
			pos += len;
			if (length <= size)
				return size;
		}
  	}
	return size;
}

char *nook_state_to_string(nook_state_t nstate)
{
	switch(nstate) {
		case nook_state_init:    return "init";
		case nook_state_running: return "running";
		case nook_state_faulted: return "faulted";
		case nook_state_cleanup: return "cleanup";
		case nook_state_zombie:  return "zombie";
	}
	return "bug";
}

int shadow_proc_read_file(char *buf, char **start, off_t offset, int length,
		int *eof, void *data)
{
	pnook_t nook = (pnook_t)data;
	struct list_head * next = NULL;
	pnook_driver_t driver;
	int size = 0, len = 0;
	off_t pos = 0;
	char buffer[1024];
	struct timeval tv;
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
	char tabs[8];
	int n, s;
#endif

	buffer[1023] = 0;
	*eof = 0;
	
	do_gettimeofday(&tv);
#define DIFFTIME(y,x)	\
	(int)((x.tv_sec - y.tv_sec) + ((x.tv_usec - y.tv_usec)/10000000)) 
	len = snprintf(buffer, 1023, "name:\t\t\t%s\n" "type:\t\t\t%s\n",
				     nook->name, nook->shadow_driver_name);

	if(nook->last_fault_time.tv_sec)
		len += snprintf(buffer+len, 1023-len,
				"last fault:\t\t%d s\n"
				"\treason:\t\t%s\n"
				"\teip:\t\t%p\n"
				"\tmodule:\t\t%s\n",
				DIFFTIME(nook->last_fault_time, tv),
				nook->last_fault_reason,
				(void*)(nook->last_fault_eip),
				nook->last_fault_mod);
	else
		len += snprintf(buffer+len, 1023-len, "no last fault\n");
				     
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
	if(nook->last_replug_time.tv_sec)
		len += snprintf(buffer+len, 1023-len,
				"last replugging:\t%d s\n",
				DIFFTIME(nook->last_replug_time, tv));
	else len += snprintf(buffer+len, 1023-len, "no last replugging\n");
#endif
	len += snprintf(buffer+len, 1023-len, "\nmodule"
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
			"\t\treplugged\tlast replug state"
#endif
			"\n");

	if (offset < len) {
		size += shadow_strncpy(buf+size,
				(offset > 0) ? (buffer+offset) : buffer,
				len - offset, length-size);
		if (offset > 0)
			offset -= len;
		pos += len;
		if (length <= size)
			return size;
	}

	list_for_each(next, &nook->drivers) {
		driver = list_entry(next, nook_driver_t, next_in_nook);
		if (driver->driver_type == mops_e) {
			len = snprintf (buffer, 1022, "%s", driver->name);
#ifdef CONFIG_NOOKS_SHADOW_REPLUG
			n = strlen(driver->name);
			s = 3 - n/8;
			if (s > 0) {
				tabs[--s] = 0;
				while(s--) tabs[s] = '\t';
			}
			else {
				tabs[0] = ' ';
				tabs[1] = 0;
			}
			len += snprintf (buffer+len, 
				1022-len, 
				"%s%s\t%s",
				tabs,
				((struct module*)driver->data)->mod_nook_replug_flag 
					& MOD_NOOK_REPLUG_OLD 
				? "   old   " :
				((struct module*)driver->data)->new_version
				? "   yes   " : "   no    ",
				((struct module*)driver->data)->last_replug 
				? ((struct module*)driver->data)->last_replug 
				: "");
#endif
			buffer[len++] = '\n';
			buffer[len] = 0;
			if (offset < len) {
				size += shadow_strncpy(buf+size,
					(offset > 0) ? (buffer+offset) : buffer,
					len - offset, length-size);
				if (offset > 0)
					offset -= len;
				pos += len;
				if (length <= size)
					return size;
			}
		}
  	}
	*eof = 1;
	return size;
}

char *shadow_fault_type_to_string(int type)
{
	switch(type) {
		case FAULT_PARAMETER: return "parameter error";
		case FAULT_RESOURCE:  return "resource error";
		case FAULT_ENOMEM:    return "insufficient memory";
		case FAULT_STACK:     return "stack corruption";
		default:              return "default fault";
	}
}

char * shadow_proc_get_module_name(unsigned addr)
{
	struct module *mod;
	
	if (addr >= (unsigned long) &_stext && addr <= (unsigned long) &_etext)
		return "kernel";

        mod = find_module_by_addr(addr);
	if (mod != NULL)
		return (char*) (mod->name);

	return "??";
}
